/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_PLFT_BASED_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include "osm_ar_plft_based_algorithm.hpp"
#include "osm_armgr.hpp"

void PlftBasedArAlgorithm::SetPlftConfiguration()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int unsupported_num = 0;
	int rc = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (sw_node_it->second.m_support[m_algorithm_feature_] ==
		    NOT_SUPPORTED) {
			continue;
		}

		rc = SetRequiredPLFTInfo(sw_node_it->second);
		if (rc) {
			unsupported_num++;
			/*TODO consider removing support_errs_t ARSWDataBaseEntry::m_error - not used*/
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
							    DF_PLFT_CAP_LESS_THAN_MIN);
		}
	}

	if (unsupported_num) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
			   "Failed to set PlftConfiguration for %u switches.\n",
			   unsupported_num);
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int PlftBasedArAlgorithm::SetRequiredPLFTInfo(ARSWDataBaseEntry &db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_advanced_routing_data_t *p_ar_data =
	&db_entry.m_general_sw_info.m_p_osm_sw->ar_data;
	ib_plft_info_t *p_plft_info = &p_ar_data->plft_info;

	u_int8_t ModeCap = p_plft_info->mode_cap;
	if (ModeCap == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9101: Switch GUID 0x%" PRIx64 ", LID %u, "
			   "Do not have sufficient PrivateLFT cap: (num tables , size).\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
		db_entry.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_ZERO);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, 1);
	}

	ib_plft_bank_entry_t *p_block_element = NULL;
	u_int8_t Mode = 0;

	uint16_t plft_max_supported_lid_number = 0;
	uint8_t min_lftable_size = (m_ar_mgr_.m_p_osm_subn->max_lid / 1024) + 1;
	uint8_t max_lftable_size = 0;
	uint8_t lftable_size;
	uint8_t plfts_per_table;
	uint8_t plfts_number = GetPlftsNumberForSwitch(db_entry);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "SetRequiredPLFTInfo: Switch GUID: 0x%" PRIx64 ", LID: %u, "
		   "ModeCap: %u, max_lid: %u, min_lftable_size: %u plfts_number: %u.\n",
		   db_entry.m_general_sw_info.m_guid,
		   db_entry.m_general_sw_info.m_lid,
		   ModeCap,
		   m_ar_mgr_.m_p_osm_subn->max_lid,
		   min_lftable_size,
		   plfts_number);

	for (int i = 0 ; i < ModeCap ; i++) {
		p_block_element = &p_plft_info->modes[i];

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
			   "ModeCap: %u, Mode: %u, size: %u, num_banks: %u.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   ModeCap, i,
			   p_block_element->size,
			   p_block_element->num_banks);

		if ((p_block_element->size / min_lftable_size) *
		    p_block_element->num_banks < plfts_number) {
		    continue;
		}

		if (p_block_element->num_banks < plfts_number) {
			plfts_per_table = ((plfts_number - 1) /
					   p_block_element->num_banks) + 1;
		} else {
			plfts_per_table = 1;
		}

		lftable_size = p_block_element->size / plfts_per_table;

		if (lftable_size > max_lftable_size) {
			max_lftable_size  = lftable_size;
			Mode = i;
			if (max_lftable_size > min_lftable_size) {
				break; /* It is enough to have extra of 1K lids */
			}
		}
	}

	if (max_lftable_size == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9102: Switch GUID: 0x%" PRIx64 ", LID: %u, "
			   "Do not have sufficient PrivateLFT cap:"
			   "(num tables < %u or max_lid_number < %u). \n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   plfts_number,
			   m_ar_mgr_.m_p_osm_subn->max_lid);

		db_entry.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_LESS_THAN_MIN);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	osm_ar_set_required_plft_active_mode(p_ar_data, Mode + 1);
	osm_ar_set_required_plft_number(p_ar_data, plfts_number);

	/* max lid number is restricted by plft table size */
	plft_max_supported_lid_number = max_lftable_size * 1024 - 1;

	int rc = osm_ar_prepare_sw_tables(m_ar_mgr_.m_p_osm_subn,
					  db_entry.m_general_sw_info.m_p_osm_sw,
					  m_ar_mgr_.m_p_osm_subn->max_lid + 1);

	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9103: Failed to allocate for advanced "
			   "routing capabilities for switch GUID: 0x%" PRIx64 ", LID: %u\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "PLFT Info Switch GUID: 0x%" PRIx64 ", LID: %u"
		   " desired mode:%d Active_Mode:%d plft_max_supported_lid_number: %u.\n",
		   db_entry.m_general_sw_info.m_guid,
		   db_entry.m_general_sw_info.m_lid,
		   Mode + 1,
		   p_plft_info->active_mode,
		   plft_max_supported_lid_number);

	rc = SetPlftDef(db_entry,
			p_plft_info->modes[Mode].size,
			p_plft_info->modes[Mode].num_banks,
			max_lftable_size, plfts_number);

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int PlftBasedArAlgorithm::SetPlftDef(ARSWDataBaseEntry &db_entry,
				     uint8_t bank_size_kb,
				     uint8_t banks_number,
				     uint8_t plft_size_kb,
				     uint8_t plfts_number)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_advanced_routing_data_t *p_ar_data =
		&db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	uint8_t offset = 0;
	uint8_t bank = 0;
	uint8_t plft_id = 0;
	for (; plft_id < plfts_number; ++plft_id) {
		if ((offset + plft_size_kb) > bank_size_kb) {
			bank++;
			offset = 0;
		}
		if (bank >= banks_number) {
			break;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetPlftDef GUID: 0x%" PRIx64 ", LID: %u "
			   "plft_id: %u ofset: %u plft_size_kb: %u bank: %u.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   plft_id, offset, plft_size_kb, bank);

		osm_ar_plft_def_set(p_ar_data, plft_id, offset, plft_size_kb, bank, 0);

		offset += plft_size_kb;
	}

	if (plft_id != plfts_number) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9104: SetPlftDef failed on GUID: 0x%" PRIx64 ", LID: %u. "
			   "Invalid Arguments -  "
			   "bank_size_kb: %u banks_number: %u plft_size_kb: %u "
			   "plfts_number: %u plft_id: %u\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   bank_size_kb, banks_number, plft_size_kb, plfts_number, plft_id);
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int PlftBasedArAlgorithm::GetSL2VL(osm_physp_t *osm_phys_port,
				   uint8_t iport_num, uint8_t oport_num,
				   uint8_t (*sl2vl)[IB_NUMBER_OF_SLS])
{
	int rc = 0;

	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_node_t *p_node = osm_physp_get_node_ptr(osm_phys_port);
	uint64_t guid = osm_node_get_node_guid_ho(p_node);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetSLVL GUID: 0x%" PRIx64 " iport: %u oport: %u\n",
		   guid, iport_num, oport_num);

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {

		rc = GetSwitchSL2VL(osm_phys_port, iport_num, oport_num, p_node,
				    sl2vl);

		if (OSM_LOG_DEBUG & m_p_osm_log->level) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "GetSLVL switch GUID: 0x%" PRIx64
				   " iport: %u oport: %u rc: %d sl2vl: %s\n",
				   guid, iport_num, oport_num, rc,
				   SLToVLMappingToStr(*sl2vl).c_str());
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	uint8_t op_vls = ib_port_info_get_op_vls(&osm_phys_port->port_info);
	if (m_ar_mgr_.m_p_osm_subn->opt.enable_vl_packing) {
		memcpy(sl2vl, m_compact_sl2vl_per_op_vls_[op_vls], sizeof(*sl2vl));
	} else {
		memcpy(sl2vl, m_sl2vl_per_op_vls_[op_vls], sizeof(*sl2vl));
	}
	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "GetSLVL GUID: 0x%" PRIx64
			   " iport: %u oport: %u op_vls: %u sl2vl: %s\n",
			   guid, iport_num, oport_num, op_vls,
			   SLToVLMappingToStr(*sl2vl).c_str());
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int PlftBasedArAlgorithm::GetSwitchSL2VL(osm_physp_t *osm_phys_port,
					 uint8_t iport_num, uint8_t oport_num,
					 osm_node_t *p_node,
					 uint8_t (*sl2vl)[IB_NUMBER_OF_SLS]) const
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	uint8_t op_vls = 0;
	bool enable_vl_packing = m_ar_mgr_.m_p_osm_subn->opt.enable_vl_packing;
	const SwitchConnection *from_connection;
	const SwitchConnection *to_connection;
	const SwitchConnection port0_connection;
	ArAlgorithmSwData *p_algorithm_sw_data;
	bool handle_port0 = false;

	if (!p_node->sw->priv) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	/*
	 * sw->priv holds routing engine's information for current switch.
	 * This information was allocated by the routing engine upon routing.
	 */
	ARSWDataBaseEntry &sw_db_entry = *(ARSWDataBaseEntry *)p_node->sw->priv;

	p_algorithm_sw_data = GetArAlgorithmSwData(sw_db_entry);
	if (iport_num == 0) {
		from_connection = &port0_connection;
		handle_port0 = true;
	} else {
		from_connection = p_algorithm_sw_data->GetPortConnection(iport_num);
	}

	if (oport_num == 0) {
		if (!ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info)) {
			return -1;
		}
		to_connection = &port0_connection;
		handle_port0 = true;
		/* for enhanced out port0 leave op vls number 0 */
	} else {
		rc = m_ar_mgr_.GetOpVlForVL2VL(sw_db_entry, osm_phys_port, oport_num,
					       GetMinimumVlCap(), op_vls);
		if (rc)
			OSM_AR_LOG_RETURN(m_p_osm_log, rc);

		to_connection = p_algorithm_sw_data->GetPortConnection(oport_num);
	}

	if (from_connection == NULL || to_connection == NULL) {
		/* From inactive port or to inactive port */
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	const sl2vl_t *p_vl2vl;
	if ((handle_port0 || enable_vl_packing) &&
	    ((from_connection->m_remote_switch == NULL) ||
	     (to_connection->m_remote_switch == NULL))) {

		if ((from_connection->m_remote_switch == NULL) &&
		    (to_connection->m_remote_switch == NULL)) {
			/* If iport_num == 0 we configure SL2VL and not VL2VL */
			p_vl2vl = &m_vl2vl_per_op_vls_[op_vls];
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Calculate Vl2Vl Mapping from "
				   "end port to end port\n");
		} else if (from_connection->m_remote_switch == NULL) {
			if (iport_num == 0) {
				/* If iport_num == 0 we configure SL2VL and not VL2VL */
				p_vl2vl = &m_sl2vl_per_op_vls_[op_vls];
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate Sl2Vl Mapping from "
					   "port0 to switch\n");
			} else {
				p_vl2vl = &m_end_point2sw_compact_vl2vl_per_op_vls_[op_vls];
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate Vl2Vl Mapping from "
					   "end_point to switch\n");
			}
		} else {
			p_vl2vl = &m_sw2end_point_compact_vl2vl_per_op_vls_[op_vls];
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Calculate Vl2Vl Mapping from "
				   "switch to end port\n");
		}
	} else {
		ar_algorithm_turn_t turn_type = GetTurnType(*from_connection,
							    *to_connection,
							    true); /* flip from.dim_sign_t */

		p_vl2vl = &m_turn_type_to_vl2vl_per_op_vls_[turn_type][op_vls];

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Calculate Vl2Vl Mapping on "
				   "switch GUID: 0x%" PRIx64 ", LID: %u "
				   "from port %u to port: %u turn_type: %u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   iport_num, oport_num, turn_type);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u turn type %u vl to vl mapping %s\n",
				   op_vls, turn_type,
				   SLToVLMappingToStr(*p_vl2vl).c_str());
		}
	}

	memcpy(sl2vl, p_vl2vl, sizeof(*sl2vl));
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void PlftBasedArAlgorithm::BuildSl2VlPerOpVl(uint16_t en_sl_mask)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* Currently support only one vl per plane, therefore: */
	uint8_t vls_per_stream = m_planes_number_;

	/* don't skip the first time init, even if sl mask is 0 */
	if (en_sl_mask != 0 && en_sl_mask == m_en_sl_mask_) {
		return;
	}
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildSl2VlPerOpVl.\n");

	m_en_sl_mask_ = en_sl_mask;
	m_update_sl2vl_ = true;
	u_int8_t num_vls;
	u_int8_t ar_enabled_vl;
	u_int8_t ar_first_enabled_vl;
	u_int8_t ar_compact_enabled_vl;
	u_int8_t ar_compact_first_enabled_vl;
	u_int8_t max_ar_enabled_vl;

	memset(&m_sl2vl_per_op_vls_, 0, sizeof(m_sl2vl_per_op_vls_));
	 /*
	 * For compact VL use only one VL per QoS
	 * on link between end port and switch.
	 */
	memset(&m_compact_sl2vl_per_op_vls_, 0, sizeof(m_compact_sl2vl_per_op_vls_));

	for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));

		if (num_vls < vls_per_stream) { /* map all SLs to VL0 */
			continue;
		} else {
			max_ar_enabled_vl = num_vls - vls_per_stream;
			if (m_en_sl_mask_ == OSM_ALL_SL_VL_AR_ENABLED) {
				/*
				 * SLs will be mapped (round robin) to:
				 * VL(0) = 0, VL(n) = VL(n-1) + vls_per_stream, .. max op vl
				 * compact mode SLs will be mapped (round robin) to:
				 * VL(0) = 0, VL(n) = VL(n-1) + 1, .. max op vl
				 */
				ar_enabled_vl = 0;
				ar_first_enabled_vl = 0;
				ar_compact_enabled_vl = 0;
				ar_compact_first_enabled_vl = 0;

			} else {
				/*
				 * Disabled SLs will be mapped to VL 0
				 * Enabled SLs will be mapped (round robin) to
				 * vls_per_stream, .. max_ar_enabled_vl
				 */
				if (num_vls == vls_per_stream) {
					/* for non compact mode map all SLs to VL0 */
					ar_enabled_vl = 0;
					ar_first_enabled_vl = 0;
				} else {
					ar_enabled_vl = vls_per_stream;
					ar_first_enabled_vl = vls_per_stream;
				}
				ar_compact_enabled_vl = 1;
				ar_compact_first_enabled_vl = 1;

			}

			for (u_int8_t sl = 0; sl < IB_NUMBER_OF_SLS; ++sl) {
				if ((1 << sl) & m_en_sl_mask_) {

					m_sl2vl_per_op_vls_[op_vls][sl] = ar_enabled_vl;
					m_compact_sl2vl_per_op_vls_[op_vls][sl] =
						ar_compact_enabled_vl;
					ar_enabled_vl += vls_per_stream;
					ar_compact_enabled_vl += 1;
					if (ar_enabled_vl > max_ar_enabled_vl) {
						ar_enabled_vl = ar_first_enabled_vl;
					}
					if (ar_compact_enabled_vl >= num_vls) {
						ar_compact_enabled_vl =
							ar_compact_first_enabled_vl;
					}
				} else {
					m_sl2vl_per_op_vls_[op_vls][sl] = 0;
					m_compact_sl2vl_per_op_vls_[op_vls][sl] = 0;
				}
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
					   "slvl_mapping[%u][%u]=%u\n",
					   sl, op_vls, m_sl2vl_per_op_vls_[op_vls][sl]);
			}
		}
	} /* op_vls loop */

	/* fix m_compact_sl2vl_per_op_vls_[MAX_OP_VL_CODE] */
	memcpy(m_compact_sl2vl_per_op_vls_[MAX_OP_VL_CODE],
	       m_compact_sl2vl_per_op_vls_[MAX_OP_VL_CODE-1],
	       sizeof(m_compact_sl2vl_per_op_vls_[MAX_OP_VL_CODE]));

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated sl to vl mapping "
			   "vls_per_stream: %u en_sl_mask: 0x%x\n",
			   vls_per_stream, m_en_sl_mask_);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u slvl_mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_sl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u compact slvl_mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_compact_sl2vl_per_op_vls_[op_vls]).c_str());
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void PlftBasedArAlgorithm::BuildVl2VlPerOpVl()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* there is one vl per plane, therefore: */
	uint8_t vls_per_stream = m_planes_number_;

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildVl2VlPerOpVl.\n");
	u_int8_t num_vls;
	u_int8_t max_vl;
	u_int8_t qos_max_vl; /* max vl for the current qos */

	memset(m_vl2vl_per_op_vls_, 0, sizeof(m_vl2vl_per_op_vls_));
	memset(m_inc_vl2vl_per_op_vls_, 0, sizeof(m_inc_vl2vl_per_op_vls_));
	memset(m_sw2end_point_compact_vl2vl_per_op_vls_, 0,
	       sizeof(m_sw2end_point_compact_vl2vl_per_op_vls_));
	memset(m_end_point2sw_compact_vl2vl_per_op_vls_, 0,
	       sizeof(m_end_point2sw_compact_vl2vl_per_op_vls_));

	for (uint8_t op_vls = 2; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));
		max_vl = min((uint8_t)(num_vls - 1), (uint8_t)(IB_DROP_VL - 1));

		if (num_vls < vls_per_stream)
			continue; /* Not Supported */

		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			/* vl < num_vls */
			u_int8_t vl = in_vl % num_vls;

			/*
			 * The qos groups number of given vl [1..last] is:
			 * (vl / vls_per_stream) + 1
			 *
			 * In each qos group there are vls_per_stream vls so for vl in specific
			 * qos group the max vl is:
			 */
			qos_max_vl = ((vl / vls_per_stream) + 1 ) * vls_per_stream - 1;
			qos_max_vl = min(qos_max_vl, max_vl);

			m_vl2vl_per_op_vls_[op_vls][in_vl] = min(qos_max_vl, vl);
			m_inc_vl2vl_per_op_vls_[op_vls][in_vl] = min(qos_max_vl, (uint8_t)(vl + 1));

			/* qos_id = in_vl / vls_per_stream */
			m_sw2end_point_compact_vl2vl_per_op_vls_[op_vls][in_vl] =
				(in_vl / vls_per_stream) % num_vls;
			m_end_point2sw_compact_vl2vl_per_op_vls_[op_vls][in_vl] =
				(in_vl * vls_per_stream) % num_vls;
		}
	} /* op_vls loop */

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated vl to vl mapping vls_per_stream: %u\n",
			   vls_per_stream);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u increase vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_inc_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u switch to end_point "
				   "compact vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_sw2end_point_compact_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u end_point to switch "
				   "compact vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_end_point2sw_compact_vl2vl_per_op_vls_[op_vls]).c_str());
		}

	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/* TODO remove doplicated code: AdaptiveRoutingManager::ARDragonFlyCycleEnd() */
void PlftBasedArAlgorithm::CycleEnd(int rc)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "CycleEnd.\n");

	m_update_sl2vl_ = false;

	if (m_ar_mgr_.m_is_permanent_error)
		m_ar_mgr_.m_is_temporary_error = false;
	else if (rc && !m_ar_mgr_.m_is_temporary_error)
		m_ar_mgr_.m_is_permanent_error = true;

	if (m_ar_mgr_.m_is_permanent_error) {
		ClearAlgorithmConfiguration();
	} else {
		map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
		for (sw_node_it = m_sw_map_.begin();
			sw_node_it != m_sw_map_.end(); ++sw_node_it) {
			if (IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
				sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;
			}
		}
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

string PlftBasedArAlgorithm::SLToVLMappingToStr(const sl2vl_t &sl2vl) const
{
	string str = "";
	char buff[1024];
	sprintf(buff,"%u, %u, %u, %u, %u, %u, %u, %u, "
		"%u, %u, %u, %u, %u, %u, %u, %u",
		sl2vl[0],
		sl2vl[1],
		sl2vl[2],
		sl2vl[3],
		sl2vl[4],
		sl2vl[5],
		sl2vl[6],
		sl2vl[7],
		sl2vl[8],
		sl2vl[9],
		sl2vl[10],
		sl2vl[11],
		sl2vl[12],
		sl2vl[13],
		sl2vl[14],
		sl2vl[15]);
	str += buff;
	return(str);
}

/*
 * Perform BFS for each destination switch.
 * Collect static information of routes with minimal VL increase
 */
int PlftBasedArAlgorithm::BuildStaticRouteInfo()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	m_max_vlinc_ = 0;
	int rc = m_ar_mgr_.m_port_groups_calculator.CalculateStaticRouteInfo(this, false);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	/* If build non minimal static routes supported*/
	if (m_support_non_minimal_hops_ && BuildNonMinimalStaticRouteInfo()) {
		/*
		 * If there is a route that exeedes vl budget, it can
		 * potantially create a credit loop.
		 * Recalculate StaticRouteInfo based on new
		 * non minimal routes, that might solve the problem.
		 */
		if (m_max_vlinc_ >= m_planes_number_) {
			m_max_vlinc_ = 0;
			rc = m_ar_mgr_.m_port_groups_calculator.
				CalculateStaticRouteInfo(this, true); /* recalculate */
			if (rc) {
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
		}
	}

	if (m_max_vlinc_ >= m_planes_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
			   "Max required VLs per SL: %u "
			   "is higher than VL budget per SL: %u.\n",
			   m_max_vlinc_ + 1, m_planes_number_);
	} else {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Max required VL increase: %u, "
			   "VL budget per SL: %u.\n",
			   m_max_vlinc_, m_planes_number_);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int PlftBasedArAlgorithm::BuildStaticRouteInfoToSwitch(
	ARSWDataBaseEntry &dst_db_entry,
	bool rebuild_route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Build static routing info for dest switch GUID: 0x%" PRIx64
		   ", LID: %u\n",
		   dst_db_entry.m_general_sw_info.m_guid,
		   dst_db_entry.m_general_sw_info.m_lid);

	ARSWDataBaseEntry *p_sw_db_entry;
	ArAlgorithmBfsData bfs_data((uint32_t)m_sw_map_.size());
	SwDbEntryPrtList& bfs_q = bfs_data.GetQueue();

	bfs_q.push_back(&dst_db_entry);
	while (!bfs_q.empty()) {
		p_sw_db_entry = bfs_q.front();
		bfs_q.pop_front();

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Popped from BFS Q: Switch GUID: 0x%" PRIx64 ", LID: %u \n",
			   p_sw_db_entry->m_general_sw_info.m_guid,
			   p_sw_db_entry->m_general_sw_info.m_lid);
		if (ProcessNeighborsBfs(
			bfs_data, p_sw_db_entry, &dst_db_entry, rebuild_route_info)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
		bfs_data.GetData(*GetArAlgorithmSwData(*p_sw_db_entry)).SetRouteStatus(
			ArAlgorithmBfsEntry::BFS_DONE);
	}

	/* Update m_max_vlinc_ should be thread safe */
	const uint8_t &max_vlinc_to_switch = bfs_data.GetMaxVlIncreaseToSwitch();
	if (m_max_vlinc_ < max_vlinc_to_switch) {
		pthread_mutex_lock(&m_ar_mgr_.m_sw_db.m_db_lock);
		if (m_max_vlinc_ < max_vlinc_to_switch) {
			m_max_vlinc_ = max_vlinc_to_switch;
		}
		pthread_mutex_unlock(&m_ar_mgr_.m_sw_db.m_db_lock);
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Adds unvisited neighbors
 */
int PlftBasedArAlgorithm::ProcessNeighborsBfs(ArAlgorithmBfsData &bfs_data,
					      ARSWDataBaseEntry *p_sw_db_entry,
					      ARSWDataBaseEntry *p_dst_db_entry,
					      bool rebuild_route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	RouteInfo minimal_route_info;
	RouteInfo route_info;
	bool calculate = true;
	bool set_route_info = false;
	SwDbEntryPrtList& bfs_q = bfs_data.GetQueue();

	ArAlgorithmSwData *p_local_switch_data =
		GetArAlgorithmSwData(*p_sw_db_entry);
	ArAlgorithmSwData *p_dst_switch_data =
		GetArAlgorithmSwData(*p_dst_db_entry);

	if (rebuild_route_info) {
		minimal_route_info = p_local_switch_data->
			GetRouteInfo(p_dst_switch_data);
		if (minimal_route_info.m_vl_inc < m_planes_number_) {
			/* No need to recalculate for valid VL increase */
			calculate = false;
		}
	}

	ConnectionVec &connections = p_local_switch_data->GetConnectionsVec();
	int connections_number = p_local_switch_data->GetConnectionsNumber();
	uint16_t sw_distance = bfs_data.GetData(*p_local_switch_data).GetRouteDistance();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "switch GUID: 0x%" PRIx64 ", LID %u: connections num :%u\n",
		   p_sw_db_entry->m_general_sw_info.m_guid,
		   p_sw_db_entry->m_general_sw_info.m_lid,
		   (int)connections.size());

	for (int connection_idx = 0; connection_idx < connections_number;
	     ++connection_idx) {

		SwitchConnection &connection = connections[connection_idx];
		if (connection.m_remote_switch == NULL) {
			continue; /* CAs connection */
		}
		ARSWDataBaseEntry *p_remote_db_entry = connection.GetRemoteSwitchDbEntry();
		ArAlgorithmSwData *p_remote_switch_data =
			GetArAlgorithmSwData(*p_remote_db_entry);
		ArAlgorithmBfsEntry &remote_switch_bfs_entry =
			bfs_data.GetData(*p_remote_switch_data);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Processing neighbor Switch GUID: 0x%" PRIx64 ", LID: %u \n",
			   p_remote_db_entry->m_general_sw_info.m_guid,
			   p_remote_db_entry->m_general_sw_info.m_lid);

		if (remote_switch_bfs_entry.GetRouteStatus() ==
		    ArAlgorithmBfsEntry::BFS_INIT) {
			bfs_q.push_back(p_remote_db_entry);
			remote_switch_bfs_entry.SetRouteStatus(
				ArAlgorithmBfsEntry::BFS_QUEUED);
			remote_switch_bfs_entry.SetRouteDistance(sw_distance + 1);
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Neighbor in BFS_INIT, adding to queue, distance: %d\n",
				   remote_switch_bfs_entry.GetRouteDistance());
		} else if ((remote_switch_bfs_entry.GetRouteStatus() ==
			    ArAlgorithmBfsEntry::BFS_DONE) &&
				   (remote_switch_bfs_entry.GetRouteDistance() + 1 == sw_distance)) {
			if (!calculate) {
				continue;
			}

			/* Considering shortest path candidate only */
			CalculateRouteInfo(connection,
					   p_remote_switch_data,
					   p_dst_switch_data,
					   route_info);

			if (rebuild_route_info) {
				if (route_info.m_vl_inc < m_planes_number_) {
					calculate = false;
					set_route_info = true;
					minimal_route_info = route_info;

					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Recalculate none shortest path "
						   "to destination switch, "
						   "vl_inc: %u turn_type: %u\n",
						   route_info.m_vl_inc,
						   route_info.m_turn_type);
					continue;

				}

				CalculateNonMinimalRouteInfo(connection,
							     *p_sw_db_entry,
							     *p_dst_db_entry,
							     route_info);
				if (route_info.m_vl_inc < m_planes_number_) {
					calculate = false;
					set_route_info = true;
					minimal_route_info = route_info;

					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Recalculate none shortest path "
						   "to destination switch, "
						   "vl_inc: %u turn_type: %u\n",
						   route_info.m_vl_inc,
						   route_info.m_turn_type);
					continue;

				}
			} else {
				set_route_info = true;
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Neighbor in BFS_DONE, "
					   "on shortest path to destination switch, "
					   "vl_inc: %u turn_type: %u\n",
					   route_info.m_vl_inc, route_info.m_turn_type);

				if (minimal_route_info.m_connection == NULL) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "No best route so far, take this neighbor as best\n");
					minimal_route_info = route_info;
				} else if (route_info < minimal_route_info) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "This route to neighbor is better\n");
					minimal_route_info = route_info;
				} else if (route_info == minimal_route_info) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "This route to neighbor equals to the best route. "
						   "Set m_alternative_connection flag.\n");

					minimal_route_info.m_alternative_connection = true;
				}
			}
		} else {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Neighbor in BFS_QUEUED, do nothing\n");
		}
	}

	if (minimal_route_info.m_connection != NULL) {
		SetRouteInfoToSwitch(bfs_data,
				     p_sw_db_entry, p_local_switch_data,
				     p_dst_db_entry, p_dst_switch_data,
				     rebuild_route_info,
				     set_route_info, /* Found the best neighbor for routing */
				     minimal_route_info);
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}


void PlftBasedArAlgorithm::SetRouteInfoToSwitch(ArAlgorithmBfsData &bfs_data,
						ARSWDataBaseEntry *p_sw_db_entry,
						ArAlgorithmSwData *p_local_switch_data,
						ARSWDataBaseEntry *p_dst_db_entry,
						ArAlgorithmSwData *p_dst_switch_data,
						bool rebuild_route_info,
						bool set_route_info,
						RouteInfo &minimal_route_info)
{
	uint8_t &max_vlinc_to_switch = bfs_data.GetMaxVlIncreaseToSwitch();

	if (set_route_info) {
		RouteInfo &sw_route_info = p_local_switch_data->
			GetRouteInfo(p_dst_switch_data);
		sw_route_info = minimal_route_info;

		uint16_t next_hop_lid = 0;
		if ((sw_route_info.m_connection != NULL) &&
			(sw_route_info.m_connection->m_remote_switch != NULL)) {

			next_hop_lid = osm_node_get_base_lid_ho(
			    sw_route_info.m_connection->m_remote_switch->p_node, 0);
		}

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Set RouteInfo from "
			   "Switch GUID: 0x%" PRIx64 ", LID: %u "
			   "to switch GUID: 0x%" PRIx64 ", LID: %u next_hop_lid: %u\n",
			   p_sw_db_entry->m_general_sw_info.m_guid,
			   p_sw_db_entry->m_general_sw_info.m_lid,
			   p_dst_db_entry->m_general_sw_info.m_guid,
			   p_dst_db_entry->m_general_sw_info.m_lid,
			   next_hop_lid);

		if (max_vlinc_to_switch < minimal_route_info.m_vl_inc) {
			max_vlinc_to_switch = minimal_route_info.m_vl_inc;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Max vlinc of %d for path from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u \n",
				   max_vlinc_to_switch,
				   p_sw_db_entry->m_general_sw_info.m_guid,
				   p_sw_db_entry->m_general_sw_info.m_lid,
				   p_dst_db_entry->m_general_sw_info.m_guid,
				   p_dst_db_entry->m_general_sw_info.m_lid);
		}
	}

	if (minimal_route_info.m_vl_inc >= m_planes_number_) {
		if (m_support_non_minimal_hops_ == 0 || rebuild_route_info) {
			/* This is the final result */
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Max required VL increase: %u, for path from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u "
				   "is greater than allowed: %u.\n",
				   minimal_route_info.m_vl_inc,
				   p_sw_db_entry->m_general_sw_info.m_guid,
				   p_sw_db_entry->m_general_sw_info.m_lid,
				   p_dst_db_entry->m_general_sw_info.m_guid,
				   p_dst_db_entry->m_general_sw_info.m_lid,
				   m_planes_number_-1);
		} else {
			/* This is temporary result without non minhop routes */
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Max required VL increase: %u "
				   "for shortest path from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u "
				   "is greater than allowed: %u.\n",
				   minimal_route_info.m_vl_inc,
				   p_sw_db_entry->m_general_sw_info.m_guid,
				   p_sw_db_entry->m_general_sw_info.m_lid,
				   p_dst_db_entry->m_general_sw_info.m_guid,
				   p_dst_db_entry->m_general_sw_info.m_lid,
				   m_planes_number_-1);
		}
	}
}

void PlftBasedArAlgorithm::SetRoutingNotification(ARSWDataBaseEntry &sw_db_entry,
						  uint8_t consume_plft_id)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetRoutingNotification for Switch GUID 0x%" PRIx64
			   ", LID %u ARN: %u FRN: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   m_ar_mgr_.IsGenerateArn(sw_db_entry),
			   m_ar_mgr_.IsGenerateFrn(sw_db_entry));
	}

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;

	/* RNGenStringTable */
	osm_ar_rn_direction_string_set(&p_sw->ar_data, 0, 0);

	/* RNGenBySubGroupPriority */
	uint8_t flags;
	flags = (m_ar_mgr_.IsGenerateArn(sw_db_entry) ? 0x1 : 0) |
		(m_ar_mgr_.IsGenerateFrn(sw_db_entry) ? 0x2 : 0);

	uint8_t max_sub_group_id =
		ib_ar_info_get_sub_grps_active(&sw_db_entry.m_required_ar_info);
	for (uint8_t sub_group_id = 0; sub_group_id <= max_sub_group_id;
	     ++sub_group_id) {
		osm_ar_subgroup_pri_rn_flags_set(&p_sw->ar_data, sub_group_id, flags);
	}

	UpdateRNRcvString(sw_db_entry, consume_plft_id);
	UpdateRNXmitPortMask(sw_db_entry);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void PlftBasedArAlgorithm::UpdateRNRcvString(ARSWDataBaseEntry &sw_db_entry,
					     uint8_t consume_plft_id)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_rn_string_handling_clear(p_ar_data);

	uint8_t consume_decision =
		(m_ar_mgr_.IsGenerateFrn(sw_db_entry) ?
		 RN_RCV_STRING_CONSUME_ALL : RN_RCV_STRING_CONSUME_ARN);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "UpdateRNRcvString Switch GUID 0x%" PRIx64 ", LID %u "
		   " string: 0 consume_decision:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   consume_decision);

	/* consume all string 0 on plft_id: consume_plft_id; */
	osm_ar_rn_string_handling_set(p_ar_data, 0, consume_decision,
				      0, consume_plft_id);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void PlftBasedArAlgorithm::UpdateRNXmitPortMask(ARSWDataBaseEntry &sw_db_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	ArAlgorithmSwData *p_switch_data = GetArAlgorithmSwData(sw_db_entry);
	const SwitchConnection * p_connection;
	u_int8_t calc_mask = 0;

	if(m_ar_mgr_.IsGenerateArn(sw_db_entry)) {
		calc_mask |= RN_XMIT_PORT_MASK_GENERATE_ARN;
	}

	if(m_ar_mgr_.IsGenerateFrn(sw_db_entry)) {
		calc_mask |= RN_XMIT_PORT_MASK_GENERATE_FRN;
	}

	calc_mask |= RN_XMIT_PORT_MASK_PASS_ON_RN;

	for (u_int8_t port_num = 1;
	     port_num <= sw_db_entry.m_general_sw_info.m_num_ports;
	     ++port_num) {

		p_connection = p_switch_data->GetPortConnection(port_num);

		if (p_connection == NULL ||
		    p_connection->m_remote_switch == NULL) {
			/* Do not send RN to CA ports or to disconnect port */
			osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, 0);
		} else {
			osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num,
							  calc_mask);
		}
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}
