%{
/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CC_PARSER_Y_Y
#include <opensm/osm_opensm.h>

#define CC_MAX_MODE_LEN 1
#define PROFILE_PERCENT_MODE 1
#define SWITCH_AQS_TIME_MAX 24
#define YYSTYPE char * 

#define HCA_BITS_NUM_ENABLE_NOTIFY			1
#define HCA_BITS_NUM_ENABLE_REACT			1

#define HCA_RP_BITS_NUM_CLAMP_TGT_RATE			1
#define HCA_RP_BITS_NUM_CLAMP_TGT_RATE_AFTER_TIME_INC	1
#define HCA_RP_BITS_NUM_RPG_TIME_RESET			17
#define HCA_RP_BITS_NUM_RPG_BYTE_RESET			19
#define HCA_RP_BITS_NUM_RPG_THRESHOLD			5
#define HCA_RP_BITS_NUM_RPG_MAX_RATE			19
#define HCA_RP_BITS_NUM_RPG_AI_RATE			16
#define HCA_RP_BITS_NUM_RPG_HAI_RATE			16
#define HCA_RP_BITS_NUM_RPG_GD				4
#define HCA_RP_BITS_NUM_RPG_MIN_DEC_FAC			7
#define HCA_RP_BITS_NUM_RPG_MIN_RATE			19
#define HCA_RP_BITS_NUM_RATE_TO_SET_ON_FIRST_CNP	32
#define HCA_RP_BITS_NUM_DCE_TCP_G			10
#define HCA_RP_BITS_NUM_DCE_TCP_RTT			17
#define HCA_RP_BITS_NUM_RATE_REDUCE_MONITOR_PERIOD 	32
#define HCA_RP_BITS_NUM_INITIAL_ALPHA_VALUE		10

#define HCA_NP_BITS_NUM_MIN_TIME_BETWEEN_CNPS 		16
#define HCA_NP_BITS_NUM_CNP_SL_MODE			1
#define HCA_NP_BITS_NUM_CNP_SL				4
#define HCA_BITS_NUM_RTT_SL_MODE			1
#define HCA_BITS_NUM_RTT_RESP_SL			4

#define PROFILE_BITS_NUM_MIN				20
#define PROFILE_BITS_NUM_MAX				20
#define PROFILE_BITS_NUM_PERCENT			7

#define SWITCH_BITS_NUM_ENABLE				1
#define SWITCH_BITS_NUM_AQS_WEIGHT			4
#define SWITCH_BITS_NUM_AQS_TIME			8

typedef struct cc_tmp_parser_struct_ {
	cl_list_t  str_list;
	cl_list_t  num_list;
	cl_list_t  entry_list;
	osm_ppcc_dictionary_entry_t entry;
} cc_tmp_parser_struct_t;

typedef enum ca_conf_id {
	CA_CONFIG_GROUP,
	CA_CONFIG_ENABLE_NOTIFY,
	CA_CONFIG_ENABLE_REACT,
	CA_CONFIG_ENABLE_CLAMP_TGT_RATE,
	CA_CONFIG_ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC,
	CA_CONFIG_RPG_TIME_RESET,
	CA_CONFIG_RPG_BYTE_RESET,
	CA_CONFIG_RPG_THRESHOLD,
	CA_CONFIG_RPG_MAX_RATE,
	CA_CONFIG_RPG_AI_RATE,
	CA_CONFIG_RPG_HAI_RATE,
	CA_CONFIG_RPG_GD,
	CA_CONFIG_RPG_MIN_DEC_FAC,
	CA_CONFIG_RPG_MIN_RATE,
	CA_CONFIG_RATE_TO_SET_ON_FIRST_CNP,
	CA_CONFIG_DCE_TCP_G,
	CA_CONFIG_DCE_TCP_RTT,
	CA_CONFIG_RATE_REDUCE_MONITOR_PERIOD,
	CA_CONFIG_INITIAL_ALPHA_VALUE,
	CA_CONFIG_MIN_TIME_BETWEEN_CNPS,
	CA_CONFIG_CNP_SL_MODE,
	CA_CONFIG_CNP_SL,
	CA_CONFIG_LAST,
} ca_conf_id_t;

typedef enum ppcc_ca_conf_id {
	PPCC_CA_CONFIG_GROUP,
	PPCC_CA_CONFIG_ENABLE_NOTIFY,
	PPCC_CA_CONFIG_ENABLE_REACT,
	PPCC_CA_CONFIG_ALGO_SL_MAP,
	PPCC_CA_CONFIG_MIN_TIME_BETWEEN_CNPS,
	PPCC_CA_CONFIG_CNP_SL_MODE,
	PPCC_CA_CONFIG_CNP_SL,
	PPCC_CA_CONFIG_RTT_SL_MODE,
	PPCC_CA_CONFIG_RTT_RESP_SL,
	PPCC_CA_CONFIG_LAST,
} ppcc_ca_conf_id_t;

typedef enum ca_algo_import_conf_id {
	CA_ALGO_IMPORT_CONFIG_ALGO,
	CA_ALGO_IMPORT_CONFIG_LAST,
} ca_algo_import_conf_id_t;

typedef enum algo_conf_id {
	CA_ALGO_CONFIG_ALGO_ID,
	CA_ALGO_CONFIG_FILE_NAME,
	CA_ALGO_CONFIG_PARAMETERS,
	CA_ALGO_CONFIG_LAST,
} algo_conf_id_t;

typedef enum switch_conf_id {
	SWITCH_CONFIG_GROUP,
	SWITCH_CONFIG_ENABLE,
	SWITCH_CONFIG_AQS_WEIGHT,
	SWITCH_CONFIG_AQS_TIME,
	SWITCH_CONFIG_LAST,
} switch_conf_id_t;

typedef enum vl_conf_id {
	VL_CONFIG_GROUP,
	VL_CONFIG_PROFILES,
	VL_CONFIG_LAST,
} vl_conf_id_t;

typedef enum sl_conf_id {
	SL_CONFIG_GROUP,
	SL_CONFIG_PROFILE_NUMS,
	SL_CONFIG_LAST,
} sl_conf_id_t;

typedef enum profiles_id {
	PROFILES_NAME,
	PROFILES_MODE,
	PROFILES_PROFILE,
	PROFILES_LAST,
} profiles_conf_id_t;

typedef enum profile_id {
	PROFILE_MIN,
	PROFILE_MAX,
	PROFILE_PERCENT,
	PROFILE_LAST,
} profile_conf_id_t;

typedef enum conf_kind {
	CONF_CA_ALGO_IMPORT,
	CONF_CA_ALGO,
	CONF_CA,
	CONF_PPCC_CA,
	CONF_SW,
	CONF_SL,
	CONF_VL,
	CONF_PROFILE,
	CONF_PROFILES
} conf_kind_t;

extern FILE * __cc_parser_in;
extern void __cc_parser_restart(FILE *input_file);
static struct osm_opensm *p_osm;
boolean_t ca_algo_import_conf_tags_parsed[CA_ALGO_IMPORT_CONFIG_LAST];
boolean_t algo_conf_tags_parsed[CA_ALGO_CONFIG_LAST];
boolean_t ca_conf_tags_parsed[CA_CONFIG_LAST];
boolean_t ppcc_ca_conf_tags_parsed[PPCC_CA_CONFIG_LAST];
boolean_t sw_conf_tags_parsed[SWITCH_CONFIG_LAST];
boolean_t sl_conf_tags_parsed[SL_CONFIG_LAST];
boolean_t vl_conf_tags_parsed[VL_CONFIG_LAST];
boolean_t profiles_conf_tags_parsed[PROFILES_LAST];
boolean_t profile_conf_tags_parsed[PROFILE_LAST];
boolean_t is_ca_algo_import_parsed = FALSE;

char *__cc_parser_yylval = NULL;

cc_tmp_parser_struct_t cc_tmp_parser_struct;

int cc_column_num;
int cc_line_num;

osm_cc_profiles_conf_t *p_current_profiles_conf;
osm_cc_profile_t *p_current_profile_conf;
osm_cc_sl_conf_t *p_current_sl_conf;
osm_cc_vl_conf_t *p_current_vl_conf;
osm_cc_sw_conf_t *p_current_sw_conf;
osm_cc_ca_conf_t *p_current_ca_conf;
osm_ppcc_ca_algos_import_conf_t *p_current_ca_algos_import_conf;
osm_ppcc_ca_algo_t *p_current_ppcc_ca_algo_conf;
int current_profile_index = 0;

extern char *trim_whitespace(char *str);

static void parser_tmp_struct_init()
{
	is_ca_algo_import_parsed = FALSE;
	cl_list_construct(&cc_tmp_parser_struct.str_list);
	cl_list_init(&cc_tmp_parser_struct.str_list, 10);
	cl_list_construct(&cc_tmp_parser_struct.num_list);
	cl_list_init(&cc_tmp_parser_struct.num_list, 10);
	cl_list_construct(&cc_tmp_parser_struct.entry_list);
	cl_list_init(&cc_tmp_parser_struct.entry_list, 10);
}

/*
 * Do NOT free objects from the temp struct.
 * Either they are inserted into the parse tree data
 * structure, or they are already freed when copying
 * their values to the parse tree data structure.
 */
static void parser_tmp_struct_reset()
{
	cl_list_remove_all(&cc_tmp_parser_struct.str_list);
	cl_list_remove_all(&cc_tmp_parser_struct.num_list);
	cl_list_remove_all(&cc_tmp_parser_struct.entry_list);
}

static void parser_tmp_struct_destroy()
{
	parser_tmp_struct_reset();
	cl_list_destroy(&cc_tmp_parser_struct.str_list);
	cl_list_destroy(&cc_tmp_parser_struct.num_list);
	cl_list_destroy(&cc_tmp_parser_struct.entry_list);
}

void destroy_tmp_num_list()
{
	cl_list_iterator_t list_iterator;
	uint64_t *p_num;

	list_iterator = cl_list_head(&cc_tmp_parser_struct.num_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.num_list)) {
		p_num = (uint64_t*)cl_list_obj(list_iterator);
		free(p_num);
		list_iterator = cl_list_next(list_iterator);
	}
}

void destroy_tmp_str_list()
{
	cl_list_iterator_t list_iterator;
	char *tmp_str;

	list_iterator = cl_list_head(&cc_tmp_parser_struct.str_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.str_list)) {
		tmp_str = (char *)cl_list_obj(list_iterator);
		free(tmp_str);
		list_iterator = cl_list_next(list_iterator);
	}
}

void destroy_tmp_entry_list()
{
	cl_list_iterator_t list_iterator;
	osm_ppcc_dictionary_entry_t *tmp_entry;

	list_iterator = cl_list_head(&cc_tmp_parser_struct.entry_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.entry_list)) {
		tmp_entry = (osm_ppcc_dictionary_entry_t *)cl_list_obj(list_iterator);
		free(tmp_entry);
		list_iterator = cl_list_next(list_iterator);
	}
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void yyerror(const char *format, ...)
{
    char s[256];
    va_list pvar;

    OSM_LOG_ENTER(&p_osm->log);

    va_start(pvar, format);
    vsnprintf(s, sizeof(s), format, pvar);
    va_end(pvar);

    OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR AC05: "
            "Syntax error (line %d:%d): %s\n",
            cc_line_num, cc_column_num, s);
    fprintf(stderr, "Error in CC Policy File (line %d:%d): %s.\n",
            cc_line_num, cc_column_num, s);
    OSM_LOG_EXIT(&p_osm->log);
}

static const char *get_ca_algo_import_param_name_by_id(int param_id)
{
	switch (param_id) {
	case CA_ALGO_IMPORT_CONFIG_ALGO:
		return "ca_algo_import_config_algo";
	default:
		break;
	}

	return "unknown";
}

static const char *get_algo_param_name_by_id(int param_id)
{
	switch (param_id) {
	case CA_ALGO_CONFIG_ALGO_ID:
		return "ca_algo_config_algo_id";
	case CA_ALGO_CONFIG_FILE_NAME:
		return "ca_algo_config_file_name";
	case CA_ALGO_CONFIG_PARAMETERS:
		return "ca_algo_config_parameters";
	default:
		break;
	}

	return "unknown";
}

const char *get_ca_param_name_by_id(int param_id)
{
	switch (param_id) {
	case CA_CONFIG_GROUP:
		return "ca_config_group";
	case CA_CONFIG_ENABLE_NOTIFY:
		return "ca_config_enable_notify";
	case CA_CONFIG_ENABLE_REACT:
		return "ca_config_enable_react";
	case CA_CONFIG_ENABLE_CLAMP_TGT_RATE:
		return "ca_config_enable_clamp_tgt_rate";
	case CA_CONFIG_ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC:
		return "ca_config_enable_clamp_tgt_rate_after_time_inc";
	case CA_CONFIG_RPG_TIME_RESET:
		return "ca_config_rpg_time_reset";
	case CA_CONFIG_RPG_BYTE_RESET:
		return "ca_config_rpg_byte_reset";
	case CA_CONFIG_RPG_THRESHOLD:
		return "ca_config_rpg_threshold";
	case CA_CONFIG_RPG_MAX_RATE:
		return "ca_config_rpg_max_rate";
	case CA_CONFIG_RPG_AI_RATE:
		return "ca_config_rpg_ai_rate";
	case CA_CONFIG_RPG_HAI_RATE:
		return "ca_config_rpg_hai_rate";
	case CA_CONFIG_RPG_GD:
		return "ca_config_rpg_gd";
	case CA_CONFIG_RPG_MIN_DEC_FAC:
		return "ca_config_rpg_min_dec_fac";
	case CA_CONFIG_RPG_MIN_RATE:
		return "ca_config_rpg_min_rate";
	case CA_CONFIG_RATE_TO_SET_ON_FIRST_CNP:
		return "ca_config_rate_to_set_on_first_cnp";
	case CA_CONFIG_DCE_TCP_G:
		return "ca_config_dce_tcp_g";
	case CA_CONFIG_DCE_TCP_RTT:
		return "ca_config_dce_tcp_rtt";
	case CA_CONFIG_RATE_REDUCE_MONITOR_PERIOD:
		return "ca_config_rate_reduce_monitor_period";
	case CA_CONFIG_INITIAL_ALPHA_VALUE:
		return "ca_config_initial_alpha_value";
	case CA_CONFIG_MIN_TIME_BETWEEN_CNPS:
		return "ca_config_min_time_between_cnps";
	case CA_CONFIG_CNP_SL_MODE:
		return "ca_config_cnp_sl_mode";
	case CA_CONFIG_CNP_SL:
		return "ca_config_cnp_sl";
	default:
		break;
	}

	return "unknown";
}

const char *get_ppcc_ca_param_name_by_id(int param_id)
{
	switch (param_id) {
	case PPCC_CA_CONFIG_GROUP:
		return "ppcc_ca_config_group";
	case PPCC_CA_CONFIG_ENABLE_NOTIFY:
		return "ppcc_ca_config_enable_notify";
	case PPCC_CA_CONFIG_ENABLE_REACT:
		return "ppcc_ca_config_enable_react";
	case PPCC_CA_CONFIG_ALGO_SL_MAP:
		return "ppcc_ca_config_algo_sl_map";
	case PPCC_CA_CONFIG_MIN_TIME_BETWEEN_CNPS:
		return "ppcc_ca_config_min_time_between_cnps";
	case PPCC_CA_CONFIG_CNP_SL_MODE:
		return "ppcc_ca_config_cnp_sl_mode";
	case PPCC_CA_CONFIG_CNP_SL:
		return "ppcc_ca_config_cnp_sl";
	case PPCC_CA_CONFIG_RTT_SL_MODE:
		return "ppcc_ca_config_rtt_sl_mode";
	case PPCC_CA_CONFIG_RTT_RESP_SL:
		return "ppcc_ca_config_rtt_resp_sl";
	default:
		break;
	}

	return "unknown";
}

const char *get_sw_param_name_by_id(int param_id)
{
	switch (param_id) {
	case SWITCH_CONFIG_GROUP:
		return "switch_config_group";
	case SWITCH_CONFIG_ENABLE:
		return "switch_config_enable";
	case SWITCH_CONFIG_AQS_WEIGHT:
		return "switch_config_aqs_weight";
	case SWITCH_CONFIG_AQS_TIME:
		return "switch_config_aqs_time";
	default:
		break;
	}

	return "unknown";
}

const char *get_sl_param_name_by_id(int param_id)
{
	switch (param_id) {
	case SL_CONFIG_GROUP:
		return "sl_config_group";
	case SL_CONFIG_PROFILE_NUMS:
		return "sl_config_profile_nums";
	default:
		break;
	}

	return "unknown";
}

const char *get_vl_param_name_by_id(int param_id)
{
	switch (param_id) {
	case VL_CONFIG_GROUP:
		return "vl_config_group";
	case VL_CONFIG_PROFILES:
		return "vl_config_profiles";
	default:
		break;
	}

	return "unknown";
}

const char *get_profile_param_name_by_id(int param_id)
{
	switch (param_id) {
	case PROFILE_MIN:
		return "profile_min";
	case PROFILE_MAX:
		return "profile_max";
	case PROFILE_PERCENT:
		return "profile_percent";
	default:
		break;
	}

	return "unknown";
}


const char *get_profiles_param_name_by_id(int param_id)
{
	switch (param_id) {
	case PROFILES_NAME:
		return "profiles_name";
	case PROFILES_MODE:
		return "profiles_mode";
	case PROFILES_PROFILE:
		return "profiles_profile";
	default:
		break;
	}

	return "unknown";
}

const char *get_param_name_by_id(conf_kind_t conf_kind, int param_id)
{
	switch (conf_kind) {
	case CONF_CA_ALGO_IMPORT:
		return get_ca_algo_import_param_name_by_id(param_id);
	case CONF_CA_ALGO:
		return get_algo_param_name_by_id(param_id);
	case CONF_CA:
		return get_ca_param_name_by_id(param_id);
	case CONF_PPCC_CA:
		return get_ppcc_ca_param_name_by_id(param_id);
	case CONF_SW:
		return get_sw_param_name_by_id(param_id);
	case CONF_SL:
		return get_sl_param_name_by_id(param_id);
	case CONF_VL:
		return get_vl_param_name_by_id(param_id);
	case CONF_PROFILE:
		return get_profile_param_name_by_id(param_id);
	case CONF_PROFILES:
		return get_profiles_param_name_by_id(param_id);
	default:
		break;
	}

	return "unknown";
}

static int are_all_params_defined(conf_kind_t conf_kind, boolean_t *params_def, int array_size)
{
	int param_id = 0;

	for (; param_id < array_size; param_id++) {
		if (!params_def[param_id]) {
			yyerror("Param %s is missing",
				get_param_name_by_id(conf_kind, param_id));
			return 0;
		}
	}

	return 1;
}

static void clear_params_defined(boolean_t *params_def, int array_size)
{
	int param_id = 0;

	for (; param_id < array_size; param_id++) {
		params_def[param_id] = FALSE;
	}
}

static void mark_tag_parsed(boolean_t *params_id_array, int array_size,
			    int param_id, const char *param_name)
{
	/* Ignore wrong param ids */
	if (param_id >= array_size)
		return;

	params_id_array[param_id] = TRUE;
}

static inline void mark_ca_tag_parsed(int param_id) {
	mark_tag_parsed(ca_conf_tags_parsed, CA_CONFIG_LAST, param_id,
		        get_ca_param_name_by_id(param_id));
}

static inline void mark_ppcc_ca_tag_parsed(int param_id) {
	mark_tag_parsed(ppcc_ca_conf_tags_parsed, PPCC_CA_CONFIG_LAST, param_id,
		        get_ppcc_ca_param_name_by_id(param_id));
}

static inline void mark_sw_tag_parsed(int param_id) {
	mark_tag_parsed(sw_conf_tags_parsed, SWITCH_CONFIG_LAST, param_id,
		        get_sw_param_name_by_id(param_id));
}

static inline void mark_sl_tag_parsed(int param_id) {
	mark_tag_parsed(sl_conf_tags_parsed, SL_CONFIG_LAST, param_id,
		        get_sl_param_name_by_id(param_id));
}

static inline void mark_vl_tag_parsed(int param_id) {
	mark_tag_parsed(vl_conf_tags_parsed, VL_CONFIG_LAST, param_id,
		        get_vl_param_name_by_id(param_id));
}

static inline void mark_profile_tag_parsed(int param_id) {
	mark_tag_parsed(profile_conf_tags_parsed, PROFILE_LAST, param_id,
		        get_profile_param_name_by_id(param_id));
}

static inline void mark_profiles_tag_parsed(int param_id) {
	mark_tag_parsed(profiles_conf_tags_parsed, PROFILES_LAST, param_id,
		        get_profiles_param_name_by_id(param_id));
}

static inline void mark_ca_algo_import_tag_parsed(int param_id) {
	mark_tag_parsed(ca_algo_import_conf_tags_parsed, CA_ALGO_IMPORT_CONFIG_LAST, param_id,
		        get_ca_algo_import_param_name_by_id(param_id));
}

static inline void mark_ca_algo_tag_parsed(int param_id) {
	mark_tag_parsed(algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST, param_id,
		        get_algo_param_name_by_id(param_id));
}

static int is_tag_parsed(boolean_t *params_id_array, int array_size,
			 int param_id, const char *param_name)
{
	if (param_id >= array_size) {
		yyerror("The param %s has a bigger id %d than the array size",
			param_name, param_id);
		return 1;
	}

	if (params_id_array[param_id]) {
		yyerror("The param %s is already defined. Ignore current parsing "
			"of the param", param_name);
		return 1;
	}

	return 0;

}

static inline int is_ca_algo_import_tag_parsed(int param_id)
{
	return is_tag_parsed(ca_algo_import_conf_tags_parsed, CA_ALGO_IMPORT_CONFIG_LAST,
			     param_id, get_ca_algo_import_param_name_by_id(param_id));
}

static inline int is_ca_algo_tag_parsed(int param_id)
{
	return is_tag_parsed(algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST,
			     param_id, get_ca_algo_import_param_name_by_id(param_id));
}

static inline int is_ca_tag_parsed(int param_id)
{
	return is_tag_parsed(ca_conf_tags_parsed, CA_CONFIG_LAST,
			     param_id, get_ca_param_name_by_id(param_id));
}

static inline int is_ppcc_ca_tag_parsed(int param_id)
{
	return is_tag_parsed(ppcc_ca_conf_tags_parsed, PPCC_CA_CONFIG_LAST,
			     param_id, get_ppcc_ca_param_name_by_id(param_id));
}

static inline int is_sw_tag_parsed(int param_id)
{
	return is_tag_parsed(sw_conf_tags_parsed, SWITCH_CONFIG_LAST,
			     param_id, get_sw_param_name_by_id(param_id));
}

static inline int is_sl_tag_parsed(int param_id)
{
	return is_tag_parsed(sl_conf_tags_parsed, SL_CONFIG_LAST,
			     param_id, get_sw_param_name_by_id(param_id));
}

static inline int is_vl_tag_parsed(int param_id)
{
	return is_tag_parsed(vl_conf_tags_parsed, VL_CONFIG_LAST,
			     param_id, get_sw_param_name_by_id(param_id));
}

static inline int is_profile_tag_parsed(int param_id)
{
	return is_tag_parsed(profile_conf_tags_parsed, PROFILE_LAST,
			     param_id, get_profile_param_name_by_id(param_id));
}

static inline int is_profiles_tag_parsed(int param_id)
{
	return is_tag_parsed(profiles_conf_tags_parsed, PROFILES_LAST,
			     param_id, get_profiles_param_name_by_id(param_id));
}

static osm_cc_profiles_conf_t* create_profiles_conf()
{
	osm_cc_profiles_conf_t *p_profiles_conf = NULL;
	current_profile_index = 0;

	p_profiles_conf = calloc(1, sizeof(osm_cc_profiles_conf_t));
	if (!p_profiles_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E06: "
			"Failed allocating memory for new profiles conf\n");
		goto Exit;
	}

Exit:
	return p_profiles_conf;
}

static void destroy_profiles_conf(osm_cc_profiles_conf_t **p_profiles_conf)
{
	free(*p_profiles_conf);
	*p_profiles_conf = NULL;
}

static osm_cc_ca_conf_t *create_ca_conf(osm_cc_type_enum cc_type)
{
	osm_cc_ca_conf_t *p_ca_conf = NULL;

	p_ca_conf = calloc(1, sizeof(osm_cc_ca_conf_t));
	if (!p_ca_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E06: "
			"Failed allocating memory for new CA conf\n");
		goto Exit;
	}

	p_ca_conf->cc_type = cc_type;
	cl_qlist_insert_tail(&p_osm->sm.cc_mgr.ca_conf, &p_ca_conf->list_item);
Exit:
	return p_ca_conf;
}

static int validate_boolean(uint64_t param, const char *name)
{
	if (param != FALSE && param != TRUE) {
		yyerror("%s has invalid value: %llu for a boolean", name, param);
		return 1;
	}

	return 0;
}


static int validate_param_size_by_bits_num(uint64_t param_value,
					   uint8_t bits_num, const char *name)
{
	uint64_t max_size = ((uint64_t)1 << bits_num) - 1;

	if (bits_num == 1)
		return validate_boolean(param_value, name);

	if (param_value > max_size) {
		yyerror("The value: %llu of %s is too big, max size is: %llu",
			param_value, name, max_size);
		return 1;
	}

	return 0;
}


static int set_ca_num_param(osm_cc_ca_conf_t *p_ca_conf, char *str_param_value,
			    int param_id, uint8_t bits_num, char *param_name)
{
	uint64_t tmp_value = strtoull(str_param_value, NULL, 0);
	if (validate_param_size_by_bits_num(tmp_value, bits_num, param_name))
		return 1;

	switch (param_id) {
	case CA_CONFIG_GROUP:
		break;
	case CA_CONFIG_ENABLE_NOTIFY:
		p_ca_conf->enable_notify = tmp_value;
		break;
	case CA_CONFIG_ENABLE_REACT:
		p_ca_conf->enable_react = tmp_value;
		break;
	case CA_CONFIG_ENABLE_CLAMP_TGT_RATE:
		p_ca_conf->hca_rp_params.enable_clamp_tgt_rate = tmp_value;
		break;
	case CA_CONFIG_ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC:
		p_ca_conf->hca_rp_params.enable_clamp_tgt_rate_after_time_inc = tmp_value;
		break;
	case CA_CONFIG_RPG_TIME_RESET:
		p_ca_conf->hca_rp_params.rpg_time_reset = tmp_value;
		break;
	case CA_CONFIG_RPG_BYTE_RESET:
		p_ca_conf->hca_rp_params.rpg_byte_reset = tmp_value;
		break;
	case CA_CONFIG_RPG_THRESHOLD:
		p_ca_conf->hca_rp_params.rpg_threshold = tmp_value;
		break;
	case CA_CONFIG_RPG_MAX_RATE:
		p_ca_conf->hca_rp_params.rpg_max_rate = tmp_value;
		break;
	case CA_CONFIG_RPG_AI_RATE:
		p_ca_conf->hca_rp_params.rpg_ai_rate = tmp_value;
		break;
	case CA_CONFIG_RPG_HAI_RATE:
		p_ca_conf->hca_rp_params.rpg_hai_rate = tmp_value;
		break;
	case CA_CONFIG_RPG_GD:
		p_ca_conf->hca_rp_params.rpg_gd = tmp_value;
		break;
	case CA_CONFIG_RPG_MIN_DEC_FAC:
		p_ca_conf->hca_rp_params.rpg_min_dec_fac = tmp_value;
		break;
	case CA_CONFIG_RPG_MIN_RATE:
		p_ca_conf->hca_rp_params.rpg_min_rate = tmp_value;
		break;
	case CA_CONFIG_RATE_TO_SET_ON_FIRST_CNP:
		p_ca_conf->hca_rp_params.rate_to_set_on_first_cnp = tmp_value;
		break;
	case CA_CONFIG_DCE_TCP_G:
		p_ca_conf->hca_rp_params.dce_tcp_g = tmp_value;
		break;
	case CA_CONFIG_DCE_TCP_RTT:
		p_ca_conf->hca_rp_params.dce_tcp_rtt = tmp_value;
		break;
	case CA_CONFIG_RATE_REDUCE_MONITOR_PERIOD:
		p_ca_conf->hca_rp_params.rate_reduce_monitor_period = tmp_value;
		break;
	case CA_CONFIG_INITIAL_ALPHA_VALUE:
		p_ca_conf->hca_rp_params.initial_alpha_value = tmp_value;
		break;
	case CA_CONFIG_MIN_TIME_BETWEEN_CNPS:
		p_ca_conf->hca_np_params.min_time_between_cnps = tmp_value;
		break;
	case CA_CONFIG_CNP_SL_MODE:
		p_ca_conf->hca_np_params.cnp_sl_mode = tmp_value;
		break;
	case CA_CONFIG_CNP_SL:
		p_ca_conf->hca_np_params.cnp_sl = tmp_value;
		break;
	default:
		break;
	}

	return 0;
}

static int set_group_type(IN char *str_group, OUT osm_cc_group_type_enum *p_group_type)
{
	osm_cc_group_type_enum group = osm_cc_get_group_type(str_group);

	if (group == OSM_CC_GROUP_NOT_DEFINED) {
		yyerror("The group %s is not a valid group definition", str_group);
		return 1;
	}

	*p_group_type = group;
	return 0;
}

static int set_ca_port_group(osm_cc_ca_conf_t *p_ca_conf, char *group_name)
{
	osm_port_group_t *p_ca_port_group = osm_get_port_grp(p_osm, __cc_parser_yylval);
	int ret = 0;

	if (!p_ca_port_group && (p_ca_conf->group_type == OSM_CC_GROUP_PORT_GROUP)) {
		yyerror("CC port group %.200s does not exist!\n", group_name);
		ret = 1;
		goto Exit;
	}

	p_ca_conf->port_group = p_ca_port_group;

Exit:
	return ret;
}

static int parse_ca_param(osm_cc_ca_conf_t *p_ca_conf, char *str_param_value,
			  int param_id, uint8_t bits_num, char *param_name)
{
	int res = 0;
	if (is_ca_tag_parsed(param_id))
		goto Exit;

	if (param_id == CA_CONFIG_GROUP) {
		if (set_group_type(str_param_value, &p_ca_conf->group_type)) {
			res = 1;
			goto Exit;
		}

		if (set_ca_port_group(p_ca_conf, str_param_value)) {
			res = 1;
			goto Exit;
		}
	} else {
		if (set_ca_num_param(p_ca_conf, str_param_value, param_id, bits_num, param_name)) {
			res = 1;
			goto Exit;
		}
	}

	mark_ca_tag_parsed(param_id);
Exit:
	return res;
}

static int parse_ppcc_ca_num_param(osm_cc_ca_conf_t *p_cc_ca_conf, char *str_param_value,
				   int param_id, uint8_t bits_num)
{
	int res = 0;
	uint64_t tmp_value;

	if (is_ppcc_ca_tag_parsed(param_id))
		goto Exit;

	tmp_value = strtoull(str_param_value, NULL, 0);
	if (validate_param_size_by_bits_num(tmp_value, bits_num,
					    get_ppcc_ca_param_name_by_id(param_id))) {
		res = 1;	
		goto Exit;
	}

	switch (param_id) {
	case PPCC_CA_CONFIG_ENABLE_NOTIFY:
		p_cc_ca_conf->enable_notify = tmp_value;
		break;
	case PPCC_CA_CONFIG_ENABLE_REACT:
		p_cc_ca_conf->enable_react = tmp_value;
		break;
	case PPCC_CA_CONFIG_MIN_TIME_BETWEEN_CNPS:
		p_cc_ca_conf->hca_np_params.min_time_between_cnps = tmp_value;
		break;
	case PPCC_CA_CONFIG_CNP_SL_MODE:
		p_cc_ca_conf->hca_np_params.cnp_sl_mode = tmp_value;
		break;
	case PPCC_CA_CONFIG_CNP_SL:
		p_cc_ca_conf->hca_np_params.cnp_sl = tmp_value;
		break;
	case PPCC_CA_CONFIG_RTT_SL_MODE:
		p_cc_ca_conf->hca_np_params.rtt_sl_mode = tmp_value;
		break;
	case PPCC_CA_CONFIG_RTT_RESP_SL:
		p_cc_ca_conf->hca_np_params.rtt_resp_sl = tmp_value;
		break;
	default:
		break;
	}

	mark_ppcc_ca_tag_parsed(param_id);
Exit:
	return res;
}

static int parse_ppcc_ca_sl_map(osm_ppcc_ca_conf_t *p_ppcc_ca_conf, int param_id)
{
	uint32_t elements_num = cl_list_count(&cc_tmp_parser_struct.num_list);
	cl_list_iterator_t list_iterator;
	uint64_t *p_num;
	uint64_t num = 0;
	int i = 0;
	int status = 0;

	if (is_ppcc_ca_tag_parsed(param_id))
		goto Exit;

	if (elements_num != IB_NUMBER_OF_SLS) {
		yyerror("Number of SLs(%u) is incorrect, supposed to be %u",
			elements_num, IB_NUMBER_OF_SLS);
		status = 1;
		goto Exit;
	}

	list_iterator = cl_list_head(&cc_tmp_parser_struct.num_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.num_list)) {
		p_num = (uint64_t*)cl_list_obj(list_iterator);
		num = *p_num;

		if (param_id == PPCC_CA_CONFIG_ALGO_SL_MAP) {
			if (num > PPCC_MAX_CA_ALGOS) {
				yyerror("Wrong algo id value: %u, "
					"value should be less the max algorithm id: %d",
					num, PPCC_MAX_CA_ALGOS);
				status = 1;
				goto Exit;
			}
			p_ppcc_ca_conf->algo_sl_map[i++] = (uint8_t) num;
		}

		list_iterator = cl_list_next(list_iterator);
	}

	mark_ppcc_ca_tag_parsed(param_id);

Exit:
	destroy_tmp_num_list();
	cl_list_remove_all(&cc_tmp_parser_struct.num_list);
	return status;
}

static int parse_profile_nums(osm_cc_sl_conf_t *p_sl_conf)
{
	uint32_t elements_num = cl_list_count(&cc_tmp_parser_struct.num_list);
	cl_list_iterator_t list_iterator;
	uint64_t *p_num;
	uint64_t num = 0;
	int i = 0;
	int status = 0;

	if (is_sl_tag_parsed(SL_CONFIG_PROFILE_NUMS))
		goto Exit;

	if ((elements_num != 1) && (elements_num != IB_NUMBER_OF_SLS)) {
		yyerror("Number of SLs(%u) is incorrect, supposed to be %u",
			elements_num, IB_NUMBER_OF_SLS);
		status = 1;
		goto Exit;
	}

	list_iterator = cl_list_head(&cc_tmp_parser_struct.num_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.num_list)) {
		p_num = (uint64_t*)cl_list_obj(list_iterator);
		num = *p_num;
		if (num > OSM_CC_MAX_PROFILE_NUM) {
			yyerror("Wrong profile number %u, "
				"value should be between 1 to %u",
				num, OSM_CC_MAX_PROFILE_NUM);
			status = 1;
			goto Exit;
		}

		p_sl_conf->sl2profile_num[i++] = (uint8_t)num;
		list_iterator = cl_list_next(list_iterator);
	}

	if (elements_num == 1) {
		for (; i < IB_NUMBER_OF_SLS; i++)
			p_current_sl_conf->sl2profile_num[i] = (uint8_t)num;
	}
	mark_sl_tag_parsed(SL_CONFIG_PROFILE_NUMS);

Exit:
	destroy_tmp_num_list();
	cl_list_remove_all(&cc_tmp_parser_struct.num_list);
	return status;
}

static int parse_vl_profiles(osm_cc_vl_conf_t *p_vl_conf)
{
	uint32_t elements_num = cl_list_count(&cc_tmp_parser_struct.str_list);
	cl_list_iterator_t list_iterator;
	osm_cc_profiles_conf_t *p_profiles_conf = NULL;
	cl_fmap_item_t *p_item;
	char *tmp_str;
	int i = 0;
	int status = 0;

	if (is_vl_tag_parsed(VL_CONFIG_PROFILES))
		goto Exit;

	if ((elements_num != 1) && (elements_num != IB_DROP_VL) &&
	    (elements_num != IB_MAX_NUM_VLS)) {
		yyerror("Number of VLs(%u) is incorrect, supposed to be %u",
			elements_num, IB_DROP_VL);
		status = 1;
		goto Exit;
	}

	list_iterator = cl_list_head(&cc_tmp_parser_struct.str_list);
	while (list_iterator != cl_list_end(&cc_tmp_parser_struct.str_list)) {
		tmp_str = (char *)cl_list_obj(list_iterator);
		p_item = cl_fmap_get(&p_osm->sm.cc_mgr.cc_profiles_map, tmp_str);

		if (p_item == cl_fmap_end(&p_osm->sm.cc_mgr.cc_profiles_map)) {
			yyerror("Profiles triplet with the name: %s is not defined", tmp_str);
			status = 1;
			goto Exit;
		}

		p_profiles_conf = (osm_cc_profiles_conf_t *)p_item;
		p_vl_conf->vl_to_profiles[i++] = p_profiles_conf;
		list_iterator = cl_list_next(list_iterator);
	}

	if (!p_profiles_conf) {
		yyerror("Invalid vl profiles configuration");
		status = 1;
		goto Exit;
	}

	if (elements_num == 1) {
		for (; i < IB_MAX_NUM_VLS; i++)
			p_current_vl_conf->vl_to_profiles[i] = p_profiles_conf;
	}

	mark_vl_tag_parsed(VL_CONFIG_PROFILES);

Exit:
	destroy_tmp_str_list();
	cl_list_remove_all(&cc_tmp_parser_struct.str_list);
	return status;
}

static int parse_profile_param(osm_cc_profile_t *p_profile_conf, char *str_param_value,
			       int param_id, uint8_t bits_num, char *param_name)
{
	int res = 0;
	uint64_t tmp_value;

	if (is_profile_tag_parsed(param_id))
		goto Exit;

	tmp_value = strtoull(str_param_value, NULL, 0);
	if (validate_param_size_by_bits_num(tmp_value, bits_num, param_name)) {
		res = 1;
		goto Exit;
	}

	switch (param_id) {
	case PROFILE_MIN:
		p_profile_conf->min = tmp_value;
		break;
	case PROFILE_MAX:
		p_profile_conf->max = tmp_value;
		break;
	case PROFILE_PERCENT:
		p_profile_conf->percent = tmp_value;
		break;
	default:
		break;
	}

	mark_profile_tag_parsed(param_id);
Exit:
	return res;
}

static int set_sw_num_param(osm_cc_sw_conf_t *p_sw_conf, char *str_param_value,
			    int param_id, uint8_t bits_num, char *param_name)
{
	int res = 0;
	uint64_t tmp_value = strtoull(str_param_value, NULL, 0);

	if (validate_param_size_by_bits_num(tmp_value, bits_num, param_name)) {
		res = 1;
		goto Exit;
	}

	switch (param_id) {
	case SWITCH_CONFIG_GROUP:
		break;
	case SWITCH_CONFIG_ENABLE:
		p_sw_conf->enable = tmp_value;
		break;
	case SWITCH_CONFIG_AQS_WEIGHT:
		p_sw_conf->aqs_weight = tmp_value;
		break;
	case SWITCH_CONFIG_AQS_TIME:
		if (tmp_value > SWITCH_AQS_TIME_MAX) {
			yyerror("Switch aqs_time: %llu is to big. "
				"Valid values are: [0 - %llu]",
				tmp_value, (uint64_t)SWITCH_AQS_TIME_MAX);
			res = 1;
			goto Exit;
		}

		p_sw_conf->aqs_time = tmp_value;
		break;
	default:
		break;
	}

Exit:
	return res;
}

static int parse_sw_param(osm_cc_sw_conf_t *p_sw_conf, char *str_param_value,
			  int param_id, uint8_t bits_num, char *param_name)
{
	int res = 0;
	if (is_sw_tag_parsed(param_id))
		goto Exit;

	if (param_id == SWITCH_CONFIG_GROUP) {
		if (set_group_type(str_param_value, &p_sw_conf->group_type)) {
			res = 1;
			goto Exit;
		}
	} else {
		if (set_sw_num_param(p_sw_conf, str_param_value, param_id, bits_num, param_name)) {
			res = 1;
			goto Exit;
		}
	}

	mark_sw_tag_parsed(param_id);
Exit:
	return res;
}

static int validate_profile(osm_cc_profile_t *p_profile_conf)
{
	int res = 0;

	if (!p_profile_conf) {
		res = 1;
		goto Exit;
	}

	if (p_profile_conf->min > p_profile_conf->max) {
		yyerror("Profile min value: %llu can't be bigger "
			"than the max value: %llu", p_profile_conf->min,
			p_profile_conf->max);
		res = 1;
		goto Exit;
	}

	if (p_profile_conf->percent > 100) {
		yyerror("Profile percent value: %u has to be between 0 and 100",
			p_profile_conf->percent);
		res = 1;
		goto Exit;
	}

Exit:
	return res;
}

static int validate_profiles(osm_cc_profiles_conf_t *p_profiles_conf) {
	osm_cc_profile_t *p_profile;
	int res = 0;
	int i;

	if (p_profiles_conf->mode == PROFILE_PERCENT_MODE) {
		for (i = 0; i < OSM_CC_MAX_PROFILE_NUM; i++) {
			p_profile = &p_profiles_conf->profiles[i];

			if ((p_profile->min > 100) || (p_profile->max > 100)) {
				yyerror("When profiles mode set to percent mode(1), "
					"profile min value: %llu and, profile max value: %llu, "
					"have be between 0 and 100", p_profile->min, p_profile->max);
				res = 1;
				goto Exit;
			}
		}
	}

Exit:
	return res;
}

static osm_cc_sl_conf_t *create_sl_conf()
{
	osm_cc_sl_conf_t *p_sl_conf = NULL;

	p_sl_conf = calloc(1, sizeof(osm_cc_sl_conf_t));
	if (!p_sl_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E06: "
			"Failed allocating memory for new SL profiles conf\n");
		goto Exit;
	}

	cl_qlist_insert_tail(&p_osm->sm.cc_mgr.sl_profiles_conf, &p_sl_conf->list_item);
Exit:
	return p_sl_conf;
}

static osm_cc_vl_conf_t *create_vl_conf()
{
	osm_cc_vl_conf_t *p_vl_conf = NULL;

	p_vl_conf = calloc(1, sizeof(osm_cc_vl_conf_t));
	if (!p_vl_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E06: "
			"Failed allocating memory for new VL profiles conf\n");
		goto Exit;
	}

	cl_qlist_insert_tail(&p_osm->sm.cc_mgr.vl_profiles_conf, &p_vl_conf->list_item);
Exit:
	return p_vl_conf;
}

static osm_cc_sw_conf_t *create_sw_conf()
{
	osm_cc_sw_conf_t *p_sw_conf = NULL;

	p_sw_conf = calloc(1, sizeof(osm_cc_sw_conf_t));
	if (!p_sw_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E06: "
			"Failed allocating memory for new SW conf\n");
		goto Exit;
	}

	cl_qlist_insert_tail(&p_osm->sm.cc_mgr.sw_conf, &p_sw_conf->list_item);
Exit:
	return p_sw_conf;
}

static osm_cc_profile_t *create_profile()
{
	osm_cc_profile_t *p_profile_conf = NULL;

	if (current_profile_index >= OSM_CC_MAX_PROFILE_NUM) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E08: "
			"Too many profiles defined inside a triplet. "
			"Only 3 are allowed\n");
		goto Exit;
	}

	p_profile_conf = &p_current_profiles_conf->profiles[current_profile_index];

	current_profile_index++;

Exit:
	return p_profile_conf;
}

static int cc_parse_algo_parameters(osm_ppcc_ca_algo_t *p_ca_algo_conf)
{
	osm_ppcc_dictionary_entry_t *tmp_entry;
	uint32_t elements_num = cl_list_count(&cc_tmp_parser_struct.entry_list);
	int i = 0;
	int ret = 0;

	if (elements_num > PPCC_ALGO_MAX_PARAMS) {
		ret = 1;
		goto Exit;
	}

	while (!cl_is_list_empty(&cc_tmp_parser_struct.entry_list)) {
		tmp_entry = (osm_ppcc_dictionary_entry_t *) cl_list_remove_head(&cc_tmp_parser_struct.entry_list);
		p_ca_algo_conf->param_list[i++] = tmp_entry;
	}

	cl_list_remove_all(&cc_tmp_parser_struct.entry_list);
Exit:
	return ret;
}

static int cc_parse_algo_file_name(osm_ppcc_ca_algo_t *p_ca_algo_conf)
{
	cl_fmap_item_t *p_item;
	int ret = 0;
	char file_path[256];
	FILE *f;

	snprintf(file_path, sizeof(file_path), "%s/%s",
		 p_osm->subn.opt.ppcc_algo_dir, __cc_parser_yylval);

	f = fopen(file_path, "r");
	if (!f) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E15: Failed to open PPCC algo file '%s' - %s\n",
			file_path, strerror(errno));
		ret = 1;
		goto Exit;
	}

	fclose(f);

	p_item = cl_fmap_get(&p_osm->sm.cc_mgr.ca_algo_conf, __cc_parser_yylval);
	if (p_item == cl_fmap_end(&p_osm->sm.cc_mgr.ca_algo_conf)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E12: "
			"Algorithm file name '%s' is not the name "
			"of any of the successfully parsed algorithm files\n", __cc_parser_yylval);
		ret = 1;
		goto Exit;
	}

	snprintf(p_ca_algo_conf->file_name, sizeof(p_ca_algo_conf->file_name), "%s",
		__cc_parser_yylval);

	mark_ca_algo_tag_parsed(CA_ALGO_CONFIG_FILE_NAME);

Exit:
	return ret;
}

static int cc_parse_algo_entry(osm_ppcc_ca_algo_t *p_ca_algo_conf, int param_id)
{
	int ret = 0;

	if (!p_ca_algo_conf) {
		ret = 1;
		goto Exit;
	}

	if (is_ca_algo_tag_parsed(param_id)) {
		if (param_id == CA_ALGO_CONFIG_PARAMETERS) {
			destroy_tmp_entry_list();
			cl_list_remove_all(&cc_tmp_parser_struct.entry_list);
		}
		goto Exit;
	}

	switch (param_id) {
	case CA_ALGO_CONFIG_ALGO_ID:
		p_ca_algo_conf->algo_id = strtoull(__cc_parser_yylval, NULL, 0);
		if (p_ca_algo_conf->algo_id < 1 || p_ca_algo_conf->algo_id > PPCC_MAX_CA_ALGOS) {
			yyerror("Wrong algo id value: %u. The allowed values for algo_id are: 1-%u",
				p_ca_algo_conf->algo_id, PPCC_MAX_CA_ALGOS);
			ret = 1;
		}
		break;
	case CA_ALGO_CONFIG_FILE_NAME:
		ret = cc_parse_algo_file_name(p_ca_algo_conf);
		break;
	case CA_ALGO_CONFIG_PARAMETERS:
		ret = cc_parse_algo_parameters(p_ca_algo_conf);
		break;
	default:
		break;
	}
	if (ret)
		goto Exit;

	mark_ca_algo_tag_parsed(param_id);
Exit:

	return ret;
}

static int cc_parse_algo(osm_ppcc_ca_algo_t *p_ppcc_ca_algo_conf)
{
	osm_ppcc_dictionary_entry_t *entry;
	osm_cc_ca_algo_conf_t *p_ca_algo_conf;
	cl_fmap_item_t *p_item;
	cl_fmap_item_t *p_algo_item;
	int ret = 0;
	int i;

	if (!are_all_params_defined(CONF_CA_ALGO, algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST)) {
		ret = 1;
		goto Exit;
	}

	p_algo_item = cl_fmap_get(&p_osm->sm.cc_mgr.ca_algo_conf,
				  p_ppcc_ca_algo_conf->file_name);
	p_ca_algo_conf = (osm_cc_ca_algo_conf_t *) p_algo_item;
	for (i = 0; i < PPCC_ALGO_MAX_PARAMS; i++) {
		entry = p_ppcc_ca_algo_conf->param_list[i];
		if (!entry)
			break;

		p_item = cl_fmap_get(&p_ca_algo_conf->param_name_to_index, entry->str);
		if (p_item == cl_fmap_end(&p_ca_algo_conf->param_name_to_index)) {
			yyerror("Algo param with the name: %s is not defined in algo in file: %s",
				entry->str, p_ppcc_ca_algo_conf->file_name);
			ret = 1;
			goto Exit;
		}
	}

	if (i > p_ca_algo_conf->param_num) {
		yyerror("Algo set more params than defined for this algorithm in the file: %s", p_ppcc_ca_algo_conf->file_name);
		ret = 1;
	}

Exit:
	return ret;
}

static int cc_parse_dictionary_entry()
{
	int ret = 0;
	osm_ppcc_dictionary_entry_t *dictionary_entry = NULL;

	if (cl_list_count(&cc_tmp_parser_struct.entry_list) >= PPCC_ALGO_MAX_PARAMS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E16: Couldn't parse dictionary of PPCC algo parameters, "
			"max number of parameters is %d\n", PPCC_ALGO_MAX_PARAMS);
		ret = 1;
		goto Exit;
	}

	dictionary_entry = calloc(1, sizeof(osm_ppcc_dictionary_entry_t));
	if (!dictionary_entry) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E11: Couldn't allocate memory for "
			"param entry, Aborting parsing\n");
		ret = 1;
		goto Exit;
	}

	memcpy(dictionary_entry->str, cc_tmp_parser_struct.entry.str, sizeof(cc_tmp_parser_struct.entry.str));
	dictionary_entry->value = cc_tmp_parser_struct.entry.value;

	cl_list_insert_tail(&cc_tmp_parser_struct.entry_list, dictionary_entry);

Exit:
    return ret;
}

static osm_ppcc_ca_algos_import_conf_t *create_ca_algo_import_conf()
{
	osm_ppcc_ca_algos_import_conf_t *p_ca_algos_import_conf = NULL;

	p_ca_algos_import_conf = calloc(1, sizeof(osm_ppcc_ca_algos_import_conf_t));
	if (!p_ca_algos_import_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E09: "
			"Failed allocating memory for new CA algos import conf\n");
		goto Exit;
	}

	p_osm->sm.cc_mgr.ca_algo_import_conf = p_ca_algos_import_conf;
	cl_qmap_init(&p_osm->sm.cc_mgr.ca_algo_import_conf->ca_algos);
Exit:
	return p_ca_algos_import_conf;
}

static osm_ppcc_ca_algo_t *create_ca_algo_conf()
{
	osm_ppcc_ca_algo_t *p_ca_algo_conf = NULL;
	
	p_ca_algo_conf = calloc(1, sizeof(osm_ppcc_ca_algo_t));
	if (!p_ca_algo_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E10: "
			"Failed allocating memory for new CA algo conf\n");
		goto Exit;
	}

	p_current_ca_algos_import_conf->algo_num++;
Exit:
	return p_ca_algo_conf;
}

int __cc_parser_parse();
int __cc_parser_lex();

int parse_congestion_control_policy(struct osm_opensm *osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__cc_parser_in = NULL;
	parser_tmp_struct_init();
	cc_line_num = 0;
	cc_column_num = 0;

	if (!p_osm->subn.opt.congestion_control_policy_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E00: No Congestion control policy file provided - "
			"Quit cc parser\n");
		res = 1;
		goto Exit;
	}

	__cc_parser_in = fopen(p_osm->subn.opt.congestion_control_policy_file, "r");
	if (!__cc_parser_in) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E02: Congestion control policy file specified but not found\n");
		res = 1;
		goto Exit;
	}

        res =  __cc_parser_parse();
	if (res) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E04: Congestion control policy file parsing failed\n");
		osm_cc_mgr_destroy(&p_osm->sm.cc_mgr);
		/*
		 * If during parsing profiles the parser failed, need to destroy
		 * it separately because it's not inside the manager yet.
		 */
		destroy_profiles_conf(&p_current_profiles_conf);
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Congestion control "
			"policy file parsed successfully\n");
	}

Exit:
	if (__cc_parser_in) {
		__cc_parser_restart(__cc_parser_in);
		fclose(__cc_parser_in);
		__cc_parser_in = NULL;
	}
	clear_val(&__cc_parser_yylval);
	parser_tmp_struct_destroy();
	return res;
}

%}

%token COMMENT
%token COMMA
%token WORD
%token GROUP
%token EXNUMBER
%token NUMBER
%token NAME

%token PROFILES_START
%token PROFILES_END
%token MODE

%token PROFILE_START
%token PROFILE_END
%token PERCENT
%token MIN
%token MAX

%token SL_CONFIG_START
%token SL_CONFIG_END
%token SL_PROFILES

%token VL_CONFIG_START
%token VL_CONFIG_END
%token VL_PROFILES

%token SWITCH_CONFIG_START
%token SWITCH_CONFIG_END
%token ENABLE
%token AQS_WEIGHT
%token AQS_TIME

%token CA_CONFIG_START
%token CA_CONFIG_END
%token ENABLE_NOTIFY
%token ENABLE_REACT
%token ENABLE_CLAMP_TGT_RATE
%token ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC
%token RPG_TIME_RESET
%token RPG_BYTE_RESET
%token RPG_THRESHOLD
%token RPG_MAX_RATE
%token RPG_AI_RATE
%token RPG_HAI_RATE
%token RPG_GD
%token RPG_MIN_DEC_FAC
%token RPG_MIN_RATE
%token RATE_TO_SET_ON_FIRST_CNP
%token DCE_TCP_G
%token DCE_TCP_RTT
%token RATE_REDUCE_MONITOR_PERIOD
%token INITIAL_ALPHA_VALUE
%token RP_FLOW_DEFINITION_MODE

%token NP_FLOW_DEFINITION_MODE
%token MIN_TIME_BETWEEN_CNPS
%token CNP_SL_MODE
%token CNP_SL

%token PPCC_CA_CONFIG_START
%token PPCC_CA_CONFIG_END
%token RTT_SL_MODE
%token RTT_RESP_SL
%token ALGO_SL_MAP

%token CA_ALGO_IMPORT_START
%token CA_ALGO_IMPORT_END
%token ALGO_START
%token ALGO_END
%token ALGO_ID
%token ALGO_FILE_NAME
%token PARAMETERS
%token LEFT_PARENTHESIS
%token RIGHT_PARENTHESIS
%token LEFT_BRACKET
%token RIGHT_BRACKET

%%

cc_policy: configurations;

configurations: /* empty */
		| configurations configuration
		;

comment: COMMENT {};

configuration:    ca_algo_import
		| profiles
		| ca_config
		| ppcc_ca_config
		| switch_config
		| vl_config
		| sl_config
		| word
		| comment
		;


ca_algo_import: ca_algo_import_start ca_algo_import_entries ca_algo_import_end
	 ;

ca_algo_import_entries:   /* empty */
			| ca_algo_import_entries ca_algo_import_entry
		 	;

ca_algo_import_entry:    algo
			| comment
			;
algo: algo_start algo_entries algo_end
	 ;

algo_entries:   /* empty */
	      | algo_entries algo_entry
	      ;

algo_entry:    algo_id
	     | algo_file_name
	     | algo_parameters
	     | comment
	     ;

ca_algo_import_start: CA_ALGO_IMPORT_START {
	if (is_ca_algo_import_parsed) {
		yyerror("ca_algo_import already parsed");
		return 1;
	}
	p_current_ca_algos_import_conf = create_ca_algo_import_conf();
	if (!p_current_ca_algos_import_conf)
		return 1;

	clear_params_defined(ca_algo_import_conf_tags_parsed, CA_ALGO_IMPORT_CONFIG_LAST);
};

ca_algo_import_end: CA_ALGO_IMPORT_END {
	osm_ppcc_dictionary_entry_t *entry;
	cl_fmap_item_t *p_algo_item;
	cl_fmap_item_t *p_item;
	cl_map_item_t *p_map_item;
	osm_cc_algo_param_index_t *p_map_algo_param_index;
	osm_cc_ca_algo_conf_t *p_ca_algo_conf;
	osm_ppcc_ca_algo_t *p_algo_conf;
	osm_cc_algo_param_t *algo_param;
	osm_ppcc_algo_id_map_item_t *p_ppcc_algo_id_map_item;
	int j;

	if (p_current_ca_algos_import_conf->algo_num >= 1)
		ca_algo_import_conf_tags_parsed[CA_ALGO_IMPORT_CONFIG_ALGO] = TRUE;

	if (!are_all_params_defined(CONF_CA_ALGO_IMPORT, ca_algo_import_conf_tags_parsed, CA_ALGO_IMPORT_CONFIG_LAST))
		return 1;

	for (p_map_item = cl_qmap_head(&p_current_ca_algos_import_conf->ca_algos);
	     p_map_item != cl_qmap_end(&p_current_ca_algos_import_conf->ca_algos);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_algo_conf = (osm_ppcc_ca_algo_t *) p_map_item;
		p_algo_item = cl_fmap_get(&p_osm->sm.cc_mgr.ca_algo_conf, p_algo_conf->file_name);
		p_ca_algo_conf = (osm_cc_ca_algo_conf_t *) p_algo_item;

		p_ppcc_algo_id_map_item = calloc(1, sizeof(osm_ppcc_algo_id_map_item_t));
		if (!p_ppcc_algo_id_map_item)
			return 1;

		p_ppcc_algo_id_map_item->algo_id = p_algo_conf->algo_id;
		memcpy(&p_ppcc_algo_id_map_item->algo_version, &p_ca_algo_conf->algo_version,
		       sizeof(p_ppcc_algo_id_map_item->algo_version));

		p_item = cl_fmap_insert(&p_osm->sm.cc_mgr.algo_version_to_id_map,
					&p_ppcc_algo_id_map_item->algo_version,
					&p_ppcc_algo_id_map_item->fmap_item);
		if (p_item != &p_ppcc_algo_id_map_item->fmap_item) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E15: "
				"Fail to add algo_id_map_item, algo_info: id: %u, minor_version: %u, major_version: %u "
				"already exist, ignoring current parsing of ca_algos_import_conf\n",
				p_ppcc_algo_id_map_item->algo_version.algo_id, p_ppcc_algo_id_map_item->algo_version.algo_minor_version,
				p_ppcc_algo_id_map_item->algo_version.algo_major_version);
			free(p_ppcc_algo_id_map_item);
			return 1;
		}
		for (j = 0; j < PPCC_ALGO_MAX_PARAMS; j++) {
			entry = p_algo_conf->param_list[j];
			if (!entry)
				break;

			p_item = cl_fmap_get(&p_ca_algo_conf->param_name_to_index, entry->str);
			p_map_algo_param_index = (osm_cc_algo_param_index_t *) p_item;
			algo_param = &p_ca_algo_conf->algo_param_list[p_map_algo_param_index->param_index];
			if (entry->value < algo_param->min_value || entry->value > algo_param->max_value) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
					"ERR 8E14: The value of algo param '%s': %u is not in the allowed range %u-%u\n",
					entry->str, entry->value, algo_param->min_value, algo_param->max_value);
				return 1;
			}
		}
	}

	is_ca_algo_import_parsed = TRUE;
};

algo_start: ALGO_START {
	if (p_current_ca_algos_import_conf->algo_num >= PPCC_MAX_CA_ALGOS)
		return 1;

	p_current_ppcc_ca_algo_conf = create_ca_algo_conf();
	if (!p_current_ppcc_ca_algo_conf)
		return 1;

	p_current_ca_algos_import_conf->p_current_ca_algo = p_current_ppcc_ca_algo_conf;
	clear_params_defined(algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST);
};

algo_end: ALGO_END {
	cl_map_item_t *p_item;

	if (!p_current_ppcc_ca_algo_conf)
		return 1;

	/* The parameters entry is optional */
	mark_ca_algo_tag_parsed(CA_ALGO_CONFIG_PARAMETERS);
	if (cc_parse_algo(p_current_ppcc_ca_algo_conf))
		return 1;

	p_item = cl_qmap_insert(&p_current_ca_algos_import_conf->ca_algos,
				p_current_ppcc_ca_algo_conf->algo_id, &p_current_ppcc_ca_algo_conf->map_item);
	if (p_item != &p_current_ppcc_ca_algo_conf->map_item) {
		yyerror("Can't insert to algo map, algo_id: %d already exists", p_current_ppcc_ca_algo_conf->algo_id);
		return 1;
	}
	p_current_ca_algos_import_conf->p_current_ca_algo = NULL;
	p_current_ppcc_ca_algo_conf = NULL;
};

algo_id: ALGO_ID NUMBER {
	if (cc_parse_algo_entry(p_current_ppcc_ca_algo_conf, CA_ALGO_CONFIG_ALGO_ID))
		return 1;
};


algo_file_name: ALGO_FILE_NAME WORD {
	if (cc_parse_algo_entry(p_current_ppcc_ca_algo_conf, CA_ALGO_CONFIG_FILE_NAME))
		return 1;
};

algo_parameters: PARAMETERS entries {
	if (cc_parse_algo_entry(p_current_ppcc_ca_algo_conf, CA_ALGO_CONFIG_PARAMETERS)) {
		destroy_tmp_entry_list();
		cl_list_remove_all(&cc_tmp_parser_struct.entry_list);
		return 1;
	}
};

entries:  entry
	| entries COMMA entry
	;

entry:  LEFT_PARENTHESIS entry_str COMMA entry_value RIGHT_PARENTHESIS {
	if (cc_parse_dictionary_entry()) {
		destroy_tmp_entry_list();
		cl_list_remove_all(&cc_tmp_parser_struct.entry_list);
		return 1;
	}
};

entry_str: WORD {
	memcpy(cc_tmp_parser_struct.entry.str, __cc_parser_yylval,
	       strlen(__cc_parser_yylval) + 1);
}

entry_value: NUMBER {
	cc_tmp_parser_struct.entry.value = strtoull(__cc_parser_yylval, NULL, 0);	
}

ppcc_ca_config:   ppcc_ca_config_start ppcc_ca_config_entries ppcc_ca_config_end
		;

ppcc_ca_config_entries:   /* empty */
			| ppcc_ca_config_entries ppcc_ca_config_entry
			;

ppcc_ca_config_entry:     ppcc_ca_config_group
			| ppcc_ca_config_enable_notify
			| ppcc_ca_config_enable_react
			| ppcc_ca_config_min_time_between_cnps
			| ppcc_ca_config_cnp_sl_mode
			| ppcc_ca_config_cnp_sl
			| ppcc_ca_config_rtt_sl_mode
			| ppcc_ca_config_rtt_resp_sl
			| ppcc_ca_config_algo_sl_map
			| comment
			;

ppcc_ca_config_start: PPCC_CA_CONFIG_START {
	p_current_ca_conf = create_ca_conf(OSM_CC_V1_PPCC);
	if (!p_current_ca_conf)
		return 1;

	clear_params_defined(ppcc_ca_conf_tags_parsed, PPCC_CA_CONFIG_LAST);
};

ppcc_ca_config_end: PPCC_CA_CONFIG_END {
	if (!are_all_params_defined(CONF_PPCC_CA, ppcc_ca_conf_tags_parsed, PPCC_CA_CONFIG_LAST))
		return 1;
};

ppcc_ca_config_group: GROUP WORD {
	if (!is_ppcc_ca_tag_parsed(PPCC_CA_CONFIG_GROUP)) {
		if (set_group_type(__cc_parser_yylval, &p_current_ca_conf->group_type))
			return 1;

		if (set_ca_port_group(p_current_ca_conf, __cc_parser_yylval))
			return 1;

		mark_ppcc_ca_tag_parsed(PPCC_CA_CONFIG_GROUP);
	}
};

ppcc_ca_config_enable_notify: ENABLE_NOTIFY NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_ENABLE_NOTIFY, HCA_BITS_NUM_ENABLE_NOTIFY))
		return 1;
};

ppcc_ca_config_enable_react: ENABLE_REACT NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_ENABLE_REACT, HCA_BITS_NUM_ENABLE_REACT))
		return 1;
};

ppcc_ca_config_min_time_between_cnps: MIN_TIME_BETWEEN_CNPS NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_MIN_TIME_BETWEEN_CNPS,
				    HCA_NP_BITS_NUM_MIN_TIME_BETWEEN_CNPS))
		return 1;
};

ppcc_ca_config_cnp_sl_mode: CNP_SL_MODE NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_CNP_SL_MODE,
				    HCA_NP_BITS_NUM_CNP_SL_MODE))
		return 1;
};

ppcc_ca_config_cnp_sl: CNP_SL NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_CNP_SL,
				    HCA_NP_BITS_NUM_CNP_SL))
		return 1;
};

ppcc_ca_config_rtt_sl_mode: RTT_SL_MODE NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_RTT_SL_MODE,
				    HCA_BITS_NUM_RTT_SL_MODE))
		return 1;
};

ppcc_ca_config_rtt_resp_sl: RTT_RESP_SL NUMBER {
	if (parse_ppcc_ca_num_param(p_current_ca_conf, __cc_parser_yylval,
				    PPCC_CA_CONFIG_RTT_RESP_SL,
				    HCA_BITS_NUM_RTT_RESP_SL))
		return 1;
};

ppcc_ca_config_algo_sl_map: ALGO_SL_MAP LEFT_BRACKET num_list RIGHT_BRACKET {
	if (parse_ppcc_ca_sl_map(&p_current_ca_conf->ppcc_ca_conf,
				 PPCC_CA_CONFIG_ALGO_SL_MAP))
		return 1;
};

ca_config: ca_config_start ca_config_entries ca_config_end
	   ;

ca_config_entries:   /* empty */
		   | ca_config_entries ca_config_entry
		   ;

ca_config_entry:    ca_config_group
		  | ca_config_enable_notify
		  | ca_config_enable_react
		  | ca_config_enable_clamp_tgt_rate
		  | ca_config_enable_clamp_tgt_rate_after_time_inc
		  | ca_config_rpg_time_reset
		  | ca_config_rpg_byte_reset
		  | ca_config_rpg_threshold
		  | ca_config_rpg_max_rate
		  | ca_config_rpg_ai_rate
		  | ca_config_rpg_hai_rate
		  | ca_config_rpg_gd
		  | ca_config_rpg_min_dec_fac
		  | ca_config_rpg_min_rate
		  | ca_config_rate_to_set_on_first_cnp
		  | ca_config_dce_tcp_g
		  | ca_config_dce_tcp_rtt
		  | ca_config_rate_reduce_monitor_period
		  | ca_config_initial_alpha_value
		  | ca_config_min_time_between_cnps
		  | ca_config_cnp_sl_mode
		  | ca_config_cnp_sl
		  | comment
		  ;

ca_config_start: CA_CONFIG_START {
	p_current_ca_conf = create_ca_conf(OSM_CC_V0_CC);
	if (!p_current_ca_conf)
		return 1;

	clear_params_defined(ca_conf_tags_parsed, CA_CONFIG_LAST);
};

ca_config_end: CA_CONFIG_END {
	if (!are_all_params_defined(CONF_CA, ca_conf_tags_parsed, CA_CONFIG_LAST))
		return 1;
};

ca_config_group: GROUP WORD {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_GROUP, 0, "group"))
		return 1;
};

ca_config_enable_notify: ENABLE_NOTIFY NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_ENABLE_NOTIFY,
			   HCA_BITS_NUM_ENABLE_NOTIFY, "enable_notify"))
		return 1;
};

ca_config_enable_react: ENABLE_REACT NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_ENABLE_REACT,
			   HCA_BITS_NUM_ENABLE_REACT, "enable_react"))
		return 1;
};

ca_config_enable_clamp_tgt_rate: ENABLE_CLAMP_TGT_RATE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_ENABLE_CLAMP_TGT_RATE,
			   HCA_RP_BITS_NUM_CLAMP_TGT_RATE, "enable_clamp_tgt_rate"))
		return 1;
};

ca_config_enable_clamp_tgt_rate_after_time_inc: ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval,
			   CA_CONFIG_ENABLE_CLAMP_TGT_RATE_AFTER_TIME_INC,
			   HCA_RP_BITS_NUM_CLAMP_TGT_RATE_AFTER_TIME_INC,
			   "enable_clamp_tgt_rate_after_time_inc"))
		return 1;
};

ca_config_rpg_time_reset: RPG_TIME_RESET NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_TIME_RESET,
			   HCA_RP_BITS_NUM_RPG_TIME_RESET, "rpg_time_reset"))
		return 1;
};

ca_config_rpg_byte_reset: RPG_BYTE_RESET NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_BYTE_RESET,
			   HCA_RP_BITS_NUM_RPG_BYTE_RESET, "rpg_byte_reset"))
		return 1;
};

ca_config_rpg_threshold: RPG_THRESHOLD NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_THRESHOLD,
			   HCA_RP_BITS_NUM_RPG_THRESHOLD, "rpg_threshold"))
		return 1;
};

ca_config_rpg_max_rate:	RPG_MAX_RATE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_MAX_RATE,
			   HCA_RP_BITS_NUM_RPG_MAX_RATE, "rpg_max_rate"))
		return 1;
};

ca_config_rpg_ai_rate: RPG_AI_RATE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_AI_RATE,
			   HCA_RP_BITS_NUM_RPG_AI_RATE, "rpg_ai_rate"))
		return 1;
};

ca_config_rpg_hai_rate: RPG_HAI_RATE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval,
			   CA_CONFIG_RPG_HAI_RATE, HCA_RP_BITS_NUM_RPG_HAI_RATE, "rpg_hai_rate"))
		return 1;
};

ca_config_rpg_gd: RPG_GD NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_GD,
			   HCA_RP_BITS_NUM_RPG_GD, "rpg_gd"))
		return 1;
};

ca_config_rpg_min_dec_fac: RPG_MIN_DEC_FAC NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_MIN_DEC_FAC,
			   HCA_RP_BITS_NUM_RPG_MIN_DEC_FAC, "rpg_min_dec_fac"))
		return 1;
};

ca_config_rpg_min_rate:	RPG_MIN_RATE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RPG_MIN_RATE,
			   HCA_RP_BITS_NUM_RPG_MIN_RATE, "rpg_min_rate"))
		return 1;
};

ca_config_rate_to_set_on_first_cnp: RATE_TO_SET_ON_FIRST_CNP NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RATE_TO_SET_ON_FIRST_CNP,
			   HCA_RP_BITS_NUM_RATE_TO_SET_ON_FIRST_CNP, "rate_to_set_on_first_cnp"))
		return 1;
};

ca_config_dce_tcp_g: DCE_TCP_G NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_DCE_TCP_G,
			   HCA_RP_BITS_NUM_DCE_TCP_G, "dce_tcp_g"))
		return 1;
};

ca_config_dce_tcp_rtt: DCE_TCP_RTT NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_DCE_TCP_RTT,
			   HCA_RP_BITS_NUM_DCE_TCP_RTT, "dce_tcp_rtt"))
		return 1;
};

ca_config_rate_reduce_monitor_period: RATE_REDUCE_MONITOR_PERIOD NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_RATE_REDUCE_MONITOR_PERIOD,
			   HCA_RP_BITS_NUM_RATE_REDUCE_MONITOR_PERIOD, "rate_reduce_monitor_period"))
			return 1;
};

ca_config_initial_alpha_value: INITIAL_ALPHA_VALUE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_INITIAL_ALPHA_VALUE,
			   HCA_RP_BITS_NUM_INITIAL_ALPHA_VALUE, "initial_alpha_value"))
		return 1;
};

ca_config_min_time_between_cnps: MIN_TIME_BETWEEN_CNPS NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_MIN_TIME_BETWEEN_CNPS,
			   HCA_NP_BITS_NUM_MIN_TIME_BETWEEN_CNPS, "min_time_between_cnps"))
		return 1;
};

ca_config_cnp_sl_mode: CNP_SL_MODE NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_CNP_SL_MODE,
			   HCA_NP_BITS_NUM_CNP_SL_MODE, "sl_mode"))
		return 1;
};

ca_config_cnp_sl: CNP_SL NUMBER {
	if (parse_ca_param(p_current_ca_conf, __cc_parser_yylval, CA_CONFIG_CNP_SL,
			   HCA_NP_BITS_NUM_CNP_SL, "cnp_sl"))
		return 1;
};

switch_config: switch_config_start switch_config_entries switch_config_end
	       ;

switch_config_entries:    /* empty */
			| switch_config_entries switch_config_entry
		        ;

switch_config_entry:	  switch_config_group
			| switch_config_enable
			| switch_config_aqs_weight
			| switch_config_aqs_time
			| comment
			;

switch_config_start: SWITCH_CONFIG_START {
	p_current_sw_conf = create_sw_conf();
	if (!p_current_sw_conf)
		return 1;

	clear_params_defined(sw_conf_tags_parsed, SWITCH_CONFIG_LAST);
};

switch_config_end: SWITCH_CONFIG_END {
	if (!are_all_params_defined(CONF_SW, sw_conf_tags_parsed, SWITCH_CONFIG_LAST))
		return 1;
};

switch_config_group: GROUP WORD {
	if (parse_sw_param(p_current_sw_conf, __cc_parser_yylval,
	    SWITCH_CONFIG_GROUP, 0, "sw_group"))
		return 1;
};

switch_config_enable: ENABLE NUMBER {
	if (parse_sw_param(p_current_sw_conf, __cc_parser_yylval,
	    SWITCH_CONFIG_ENABLE, SWITCH_BITS_NUM_ENABLE, "sw_enable"))
		return 1;
};

switch_config_aqs_weight: AQS_WEIGHT NUMBER {
	if (parse_sw_param(p_current_sw_conf, __cc_parser_yylval,
	    SWITCH_CONFIG_AQS_WEIGHT, SWITCH_BITS_NUM_AQS_WEIGHT, "sw_aqs_weight"))
		return 1;
};

switch_config_aqs_time: AQS_TIME NUMBER {
	if (parse_sw_param(p_current_sw_conf, __cc_parser_yylval,
	    SWITCH_CONFIG_AQS_TIME, SWITCH_BITS_NUM_AQS_TIME, "sw_aqs_time"))
		return 1;
};

sl_config: sl_config_start sl_config_entries sl_config_end
	   ;

sl_config_start: SL_CONFIG_START {
	p_current_sl_conf = create_sl_conf();
	if (!p_current_sl_conf)
		return 1;

	clear_params_defined(sl_conf_tags_parsed, SL_CONFIG_LAST);
};

sl_config_end: SL_CONFIG_END {
	if (!are_all_params_defined(CONF_SL, sl_conf_tags_parsed, SL_CONFIG_LAST))
		return 1;
};

sl_config_entries:   /* empty */
		   | sl_config_entries sl_config_entry
		   ;

sl_config_entry:    sl_config_group
		  | sl_config_profile_nums
		  | comment
		  ;

sl_config_group: GROUP WORD {
	if (!is_sl_tag_parsed(SL_CONFIG_GROUP)) {
		if (set_group_type(__cc_parser_yylval, &p_current_sl_conf->group_type))
			return 1;

		mark_sl_tag_parsed(SL_CONFIG_GROUP);
	}
};

sl_config_profile_nums: SL_PROFILES num_list /* list of profile numbers */{
	if (parse_profile_nums(p_current_sl_conf))
		return 1;
};

vl_config: vl_config_start vl_config_entries vl_config_end
	   ;

vl_config_start: VL_CONFIG_START {
	p_current_vl_conf = create_vl_conf();
	if (!p_current_vl_conf)
		return 1;

	clear_params_defined(vl_conf_tags_parsed, VL_CONFIG_LAST);
};

vl_config_end: VL_CONFIG_END {
	if (!are_all_params_defined(CONF_VL, vl_conf_tags_parsed, VL_CONFIG_LAST))
		return 1;
};

vl_config_entries:   /* empty */
		   | vl_config_entries vl_config_entry
		   ;

vl_config_entry:    vl_config_group
		  | vl_config_profiles
		  | comment
		  ;

vl_config_group: GROUP WORD {
	if (!is_vl_tag_parsed(VL_CONFIG_GROUP)) {
		if (set_group_type(__cc_parser_yylval, &p_current_vl_conf->group_type))
			return 1;

		mark_vl_tag_parsed(VL_CONFIG_GROUP);
	}
};

vl_config_profiles: VL_PROFILES string_list {
	if (parse_vl_profiles(p_current_vl_conf))
		return 1;

};

profiles: profiles_start profiles_entries profiles_end
	  ;

profiles_entries:   /* empty */
		  | profiles_entries profiles_entry
		  ;

profiles_entry:    profiles_name
		 | profiles_mode
		 | profile
		 | comment
		 ;

profiles_start: PROFILES_START {
	free(p_current_profiles_conf);
	p_current_profiles_conf = create_profiles_conf();
	if (!p_current_profiles_conf)
		return 1;

	clear_params_defined(profiles_conf_tags_parsed, PROFILE_LAST);
}

profiles_end: PROFILES_END {
	cl_fmap_item_t *p_item;

	if (current_profile_index == OSM_CC_MAX_PROFILE_NUM)
		profiles_conf_tags_parsed[PROFILES_PROFILE] = TRUE;

	if (!are_all_params_defined(CONF_PROFILES, profiles_conf_tags_parsed, PROFILE_LAST))
		return 1;

	if (validate_profiles(p_current_profiles_conf))
		return 1;

	p_item = cl_fmap_insert(&p_osm->sm.cc_mgr.cc_profiles_map,
				&p_current_profiles_conf->name,
				&p_current_profiles_conf->fmap_item);
	if (p_item != &p_current_profiles_conf->fmap_item) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E13: "
			"Fail to add profiles_conf, name: %s of profiles_conf "
			"already exist, ignoring current parsing of profiles_conf\n",
			p_current_profiles_conf->name);
		destroy_profiles_conf(&p_current_profiles_conf);
	}

	p_current_profiles_conf = NULL;
}

profiles_name: NAME WORD {
	if (!p_current_profiles_conf)
		return 1;

	if (!is_profiles_tag_parsed(PROFILES_NAME)) {
		snprintf(p_current_profiles_conf->name, sizeof(p_current_profiles_conf->name),
			 "%s", __cc_parser_yylval);
		mark_profiles_tag_parsed(PROFILES_NAME);
	}
};

profiles_mode: MODE NUMBER {
	if (!p_current_profiles_conf)
		return 1;

	if (!is_profiles_tag_parsed(PROFILES_MODE)) {
		if (strlen(__cc_parser_yylval) <= CC_MAX_MODE_LEN) {
			uint64_t tmp_value = strtoull(__cc_parser_yylval, NULL, 0);
			p_current_profiles_conf->mode = tmp_value;
		} else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E07: Illegal "
				"mode, mode qualifier length must be up to "
				"%d, Aborting congestion control parsing!!!\n",
				CC_MAX_MODE_LEN);
			return 1;
		}
		mark_profiles_tag_parsed(PROFILES_MODE);
	}
};

profile: profile_start profile_entries profile_end
	 ;

profile_entries: /* empty */
		 | profile_entries profile_entry
		 ;

profile_entry:    profile_percent
		| profile_min
		| profile_max
		| comment
		;

profile_start: PROFILE_START {
	p_current_profile_conf = create_profile();
	if (!p_current_profile_conf)
		return 1;

	clear_params_defined(profile_conf_tags_parsed, PROFILE_LAST);
};

profile_end: PROFILE_END {
	if (!are_all_params_defined(CONF_PROFILE, profile_conf_tags_parsed, PROFILE_LAST))
		return 1;

	if (validate_profile(p_current_profile_conf))
		return 1;
};

profile_min: MIN NUMBER {
	if (parse_profile_param(p_current_profile_conf, __cc_parser_yylval,
				PROFILE_MIN, PROFILE_BITS_NUM_MIN, "profile_min"))
		return 1;
};

profile_max: MAX NUMBER {
	if (parse_profile_param(p_current_profile_conf,__cc_parser_yylval,
				PROFILE_MAX, PROFILE_BITS_NUM_MAX, "profile_max"))
		return 1;
};

profile_percent: PERCENT NUMBER {
	if (parse_profile_param(p_current_profile_conf,__cc_parser_yylval,
				PROFILE_PERCENT, PROFILE_BITS_NUM_PERCENT, "profile_percent"))
		return 1;
};

num_list:   number
	  | num_list COMMA number
	  ;

number: NUMBER {
	uint64_t *p_num = (uint64_t*)malloc(sizeof(uint64_t));
	if (!p_num) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E05: Couldn't allocate memory, Aborting parsing\n");
		return 1;
	}

	*p_num = strtoull(__cc_parser_yylval, NULL, 0);
	cl_list_insert_tail(&cc_tmp_parser_struct.num_list, p_num);
};

string_list:   string
	     | string_list COMMA string
	     ;

string: WORD {
	char *tmp_str = strdup(__cc_parser_yylval);
	if (!tmp_str) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 8E05: Couldn't allocate memory for "
			"profile triplet name, Aborting parsing\n");
		return 1;
	}

	cl_list_insert_tail(&cc_tmp_parser_struct.str_list, tmp_str);
};

word: WORD {
	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 8E05: Unexpected "
		"input |%.200s|, Aborting parsing!\n", __cc_parser_yylval);
	return 1;
};
