/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of hierarchy information and planarization related functionality.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */


#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_HIERARCHY_INFO_C
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_advanced_routing.h>

#define hierarchy_info_level_mask(val, msb, lsb) \
	(((val) & (((1 << ((msb) - (lsb) + 1)) - 1) << lsb)) >> lsb)

#define SHORT_STR_LEN		10

extern boolean_t is_sm_port(IN osm_sm_t *sm, IN osm_node_t * p_node, IN const uint8_t port_num);

static boolean_t osm_is_hierarchy_info_supported(IN osm_physp_t * p_physp)
{
	ib_net32_t capability_mask;
	osm_physp_t *p_physp0;

	if (!p_physp->p_node)
	       return FALSE;

	if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_physp->p_node, 0);
		capability_mask = p_physp0->port_info.capability_mask;
	} else {
		capability_mask = p_physp->port_info.capability_mask;
	}

	return (capability_mask & IB_PORT_CAP_HAS_HIER_INFO);
}

static void process_ndr_hierarchy_info(IN struct osm_sm *p_sm, IN osm_physp_t * p_physp)
{
	osm_hierarchy_info_t *p_hierarchy;
	char split_str[SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8;
	int split, port, cage, slot_type, slot_serial_num, port_type;
	int bdf_bus, bdf_device, bdf_port;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_NDR)
		return;

	split = port = cage = slot_type = slot_serial_num = port_type =
	bdf_bus = bdf_device = bdf_port = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_SPLIT:
				split = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_PORT:
				port = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_ASIC:
				p_hierarchy->asic = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_SLOT:
				/* bits lsb 0, msb 1 */
				slot_type = hierarchy_info_level_mask(level, 1, 0);
				slot_serial_num = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_TYPE:
				/* bits lsb 0, msb 0 */
				port_type = hierarchy_info_level_mask(level, 0, 0);
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_BDF:
				/* bits lsb 16, msb 23 */
				bdf_bus = hierarchy_info_level_mask(level, 23, 16);
				/* bits lsb 11, msb 15 */
				bdf_device = hierarchy_info_level_mask(level, 15, 11);
				/* bits lsb 8, msb 10 */
				bdf_port = hierarchy_info_level_mask(level, 10, 8);
				break;
			default:
				break;
		}
	}
	/* switch and hca have to have 'cage' and 'port' for port label */
	if (cage == -1 || port == -1 ||
	    (node_type == IB_NODE_TYPE_SWITCH && p_hierarchy->asic == -1))
		return;

	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	if (split != -1) {
		snprintf(split_str, SHORT_STR_LEN, "/%d", split);
	}

	if (node_type == IB_NODE_TYPE_SWITCH) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%d/%d/%d%s",
			 p_hierarchy->asic, cage, port, split_str);
	} else if (node_type == IB_NODE_TYPE_CA) {
		/* slot type is physical */
		if (slot_type == 1 && slot_serial_num != -1) {
			snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN,
					"Physical%d/%d/%d%s",
					slot_serial_num, cage, port, split_str);
		} else if (bdf_bus != -1 && bdf_device != -1 &&	bdf_port != -1) {
			snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN,
					"B%dD%dF%d/%d/%d%s",
					bdf_bus, bdf_device,
					bdf_port, cage, port,
					split_str);
		}
	}
	OSM_LOG_EXIT(p_sm->p_log);
}

#define OSM_HIERARCHY_INFO_PORT_TYPE_NUM       6
const char* g_xdr_port_type_strings[OSM_HIERARCHY_INFO_PORT_TYPE_NUM] = {"", "FNM1", "FNM", "sw", "HCA", "Access"};

static boolean_t check_xdr_template_guid_validity(IN struct osm_sm *p_sm,
						  IN OUT osm_physp_t * p_physp,
						  IN int port_type,
						  uint64_t template_guid)
{
	int aport = p_physp->hierarchy_info.aport;
	int plane_number = p_physp->hierarchy_info.plane_number;
	int num_of_planes = p_physp->hierarchy_info.num_of_planes;
	int asic = p_physp->hierarchy_info.asic;
	uint8_t num_of_ports = osm_node_get_num_physp(p_physp->p_node);
	uint8_t port_num = p_physp->port_num;
	osm_node_t *p_node = p_physp->p_node;
	boolean_t is_planarized = TRUE;

	/* this port has no planarization info */
	if (aport == -1) {
		p_physp->hierarchy_info.is_planarized = FALSE;
		return p_physp->hierarchy_info.is_planarized;
	}

	/* the aport is present, so the port is expected to be planarized */
	if (aport == 0 || aport >= num_of_ports) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A007: "
			"Node 0x%" PRIx64 " port %u has invalid aport number %d "
			"(number of ports %u).\n",
			osm_node_get_node_guid_ho(p_node), port_num, aport, num_of_ports);
		is_planarized = FALSE;
	}

	if (num_of_planes <= 0 || num_of_planes > IB_MAX_NUM_OF_PLANES_XDR) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A008: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid number of planes %d.\n",
			osm_node_get_node_guid_ho(p_node), aport,
			p_physp->port_num, num_of_planes);
		is_planarized = FALSE;
	}

	if (plane_number <= 0 || plane_number > IB_MAX_NUM_OF_PLANES_XDR) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A004: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid plane number %d.\n",
			osm_node_get_node_guid_ho(p_node), aport,
			port_num, plane_number);
		is_planarized = FALSE;
	}

	if (template_guid == OSM_HI_TEMPLATE_GUID_4 &&
	    (asic <= 0 || asic > OSM_PLANAR_MAX_NUM_OF_ASICS)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A009: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid asic number %d.\n",
			osm_node_get_node_guid_ho(p_node), aport,
			port_num, asic);
		is_planarized = FALSE;
	}

	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A00A: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid port type value %d.\n",
			osm_node_get_node_guid_ho(p_node), aport,
			port_num, port_type);
	}

	p_physp->hierarchy_info.is_planarized = is_planarized;

	return p_physp->hierarchy_info.is_planarized;
}

static void process_hierarchy_info_template_guid_5(IN osm_sm_t *p_sm, IN osm_physp_t * p_physp,
						   IN uint64_t template_guid)
{
	osm_hierarchy_info_t *p_hierarchy;
	char cage_str[SHORT_STR_LEN] = {0}, ipil_str[SHORT_STR_LEN] = {0};
	char split_str[SHORT_STR_LEN] = {0}, bdf_str[2 * SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8, value_15_8;
	int port_type, cage, ipil, split;
	int bdf_bus, bdf_device, bdf_port;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_5)
		return;

	port_type = cage = ipil = split = bdf_bus = bdf_device = bdf_port = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		value_15_8 = hierarchy_info_level_mask(level, 15, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_PORT_TYPE:
				port_type = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_IPIL:
				ipil = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_SPLIT:
				split = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_BDF:
				bdf_bus = hierarchy_info_level_mask(level, 23, 16);
				bdf_device = hierarchy_info_level_mask(level, 15, 11);
				bdf_port = hierarchy_info_level_mask(level, 10, 8);
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_PLANE:
				p_hierarchy->plane_number = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_NUM_OF_PLANES:
				p_hierarchy->num_of_planes = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_APORT:
				p_hierarchy->aport = value_23_8;
				break;
			default:
				break;
		}
	}

	check_xdr_template_guid_validity(p_sm, p_physp, port_type, template_guid);

	/* switch and hca have to have 'port_type'. If none, use empty string */
	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM)
		port_type = 0;

	if (cage != -1)
		snprintf(cage_str, sizeof(cage_str), "%d", cage);

	if (ipil != -1)
		snprintf(ipil_str, sizeof(ipil_str), "p%d", ipil);

	if (split != -1)
		snprintf(split_str, sizeof(split_str), "s%d", split);

	if (bdf_bus != -1 && bdf_device != -1 && bdf_port != -1)
		snprintf(bdf_str, sizeof(bdf_str), "ibB%dD%dF%d",
			 bdf_bus, bdf_device, bdf_port);

	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	/* ibB{bus}D{device}F{function}{PortType}{$cage}p{IPIL}s{split} */
	if (node_type == IB_NODE_TYPE_CA) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%s%s%s%s%s",
			 bdf_str, g_xdr_port_type_strings[port_type],
			 cage_str, ipil_str, split_str);
	}

	OSM_LOG_EXIT(p_sm->p_log);
}

static void process_hierarchy_info_template_guid_4(IN osm_sm_t *p_sm, IN osm_physp_t * p_physp,
						   IN uint64_t template_guid)
{
	osm_hierarchy_info_t *p_hierarchy;
	char asic_name_str[SHORT_STR_LEN] = {0}, ibport_str[SHORT_STR_LEN] = {0};
	char cage_str[SHORT_STR_LEN] = {0}, ipil_str[SHORT_STR_LEN] = {0};
	char split_str[SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8, value_15_8;
	int port_type, asic_name, ib_port, cage, ipil, split;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_4)
		return;

	port_type = asic_name = ib_port = cage = ipil = split = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		value_15_8 = hierarchy_info_level_mask(level, 15, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_PORT_TYPE:
				port_type = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_ASIC_NAME:
				asic_name = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_IB_PORT:
				ib_port = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_SWITCH_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_IPIL:
				ipil = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_SPLIT:
				split = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_ASIC:
				p_hierarchy->asic = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_TYPE:
				p_hierarchy->port_location_type =
					hierarchy_info_level_mask(level, 7, 0);
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_PLANE:
				p_hierarchy->plane_number = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_NUM_OF_PLANES:
				p_hierarchy->num_of_planes = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_APORT:
				p_hierarchy->aport = value_23_8;
				break;
			default:
				break;
		}
	}

	check_xdr_template_guid_validity(p_sm, p_physp, port_type, template_guid);

	/* switch and hca have to have 'port_type'. If none, use empty string */
	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM)
		port_type = 0;

	if (asic_name >= 1 && asic_name <= 4)
		snprintf(asic_name_str, sizeof(asic_name_str), "A%d", asic_name);
	else if (asic_name == 256)
		snprintf(asic_name_str, sizeof(asic_name_str), "A");
	else if (asic_name == 257)
		snprintf(asic_name_str, sizeof(asic_name_str), "B");
	else if (asic_name != -1)
		snprintf(asic_name_str, sizeof(asic_name_str), "%d", asic_name);

	if (ib_port != -1)
		snprintf(ibport_str, sizeof(ibport_str), "P%d", ib_port);

	if (cage != -1)
		snprintf(cage_str, sizeof(cage_str), "%d", cage);

	if (ipil != -1)
		snprintf(ipil_str, sizeof(ipil_str), "p%d", ipil);

	if (split != -1)
		snprintf(split_str, sizeof(split_str), "s%d", split);


	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	/* {PortType}{AsicName}P{Ibport}{SwitchCage}p{IPIL}sp{Split} */
	if (node_type == IB_NODE_TYPE_SWITCH) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%s%s%s%s%s%s",
			 g_xdr_port_type_strings[port_type], asic_name_str, ibport_str,
			 cage_str, ipil_str, split_str);

	}

	OSM_LOG_EXIT(p_sm->p_log);
}

void osm_get_hierarchy_info(IN osm_sm_t * sm, IN osm_node_t * p_node,
			    IN osm_port_t * p_port,
			    IN osm_physp_t * p_physp,
			    IN uint8_t index)
{
	osm_madw_context_t context;
	osm_physp_t *p_physp0;
	ib_net64_t m_key;
	osm_madw_t *p_madw;
	uint8_t port_num = osm_physp_get_port_num(p_physp);
	uint32_t attr_mod = (index << IB_HIERARCHY_INFO_ATTR_MOD_INDEX_SHIFT); /* bits 8 - 15 */

	p_physp->hierarchy_info_updated =
		p_physp->hierarchy_info_updated &&
		!sm->p_subn->coming_out_of_standby &&
		!p_physp->need_update;

	if (p_physp->hierarchy_info_updated)
		return;

	/* check capability only for the first block, assuming that index 0 is always sent first */
	if (index == 0 && !osm_is_hierarchy_info_supported(p_physp))
		return;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_node, 0);
		m_key = ib_port_info_get_m_key(&p_physp0->port_info);
		/* bits 0-7 */
		attr_mod |= (IB_HIERARCHY_INFO_ATTR_MOD_PORT_MASK & port_num);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"SWITCH HI Node 0x%" PRIx64 " port number %d attr_mod %u\n",
			osm_node_get_node_guid_ho(p_node), port_num, attr_mod);
	} else {
		m_key = ib_port_info_get_m_key(&p_physp->port_info);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"CA HI Node 0x%" PRIx64 " port number %d attr_mod %u\n",
			osm_node_get_node_guid_ho(p_node), port_num, attr_mod);
	}

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = FALSE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;
	context.pi_context.client_rereg = FALSE;
	context.pi_context.port_num = port_num;

	p_madw = osm_prepare_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			IB_MAD_ATTR_HIERARCHY_INFO,
			cl_hton32(attr_mod), /* query index 0 */
			FALSE, m_key, 0, CL_DISP_MSGID_NONE, 
			osm_physp_get_node_guid(p_physp), &context);
	if (!p_madw)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A001: "
			"Failure initiating Hierarchical Info request\n");
	else
		cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
}

static boolean_t validate_hierarchy_info(osm_log_t *p_log, osm_node_t *p_node,
					 uint8_t port_num, ib_hierarchy_info_t *p_hi,
					 uint8_t hierarchy_index)
{
	if (p_hi->max_active_index > p_hi->max_supported_index) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A00B: "
			"Port 0x%" PRIx64 " port num %u, hierarchy index %u, max active index (%u)"
			" is greater than max supported index (%u).\n",
			osm_node_get_node_guid_ho(p_node), port_num, hierarchy_index,
			p_hi->max_active_index, p_hi->max_supported_index);
		return FALSE;
	}
	return TRUE;
}

void osm_hierarchy_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_hierarchy_info_t *p_hi;
	ib_smp_t *p_smp;
	osm_pi_context_t *p_context;
	uint8_t port_num;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t hierarchy_index;
	uint64_t template_guid;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_HIERARCHY_INFO);

	p_context = osm_madw_get_pi_context_ptr(p_madw);
	p_hi = ib_smp_get_payload_ptr(p_smp);

	port_num = p_context->port_num;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_port = osm_get_port_by_guid(sm->p_subn, p_context->port_guid);
	if (PF(!p_port)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A002: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(p_context->port_guid),
			cl_ntoh64(p_context->node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, port_num);

	OSM_ASSERT(p_physp);

	/* process only valid hierarchy info responses */
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received. Send next mad.\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	/* reset previous information about aport */
	p_physp->hierarchy_info.p_aport = NULL;

	/* bits 15 - 8 */
	hierarchy_index = (cl_ntoh32(p_smp->attr_mod) & IB_HIERARCHY_INFO_ATTR_MOD_INDEX_MASK) >>
		IB_HIERARCHY_INFO_ATTR_MOD_INDEX_SHIFT;

	if (!validate_hierarchy_info(sm->p_log, p_node, port_num, p_hi, hierarchy_index))
		goto Exit;

	template_guid = cl_ntoh64(p_hi->template_guid);
	if (template_guid == OSM_HI_TEMPLATE_GUID_NDR) {
		p_physp->hierarchy_info.block = *p_hi;
		process_ndr_hierarchy_info(sm, p_physp);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Node 0x%" PRIx64 " port number %d label %s\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			p_physp->hierarchy_info.port_label);
	} else if (template_guid == OSM_HI_TEMPLATE_GUID_4) {
		p_physp->hierarchy_info.block = *p_hi;
		process_hierarchy_info_template_guid_4(sm, p_physp, template_guid);
	} else if (template_guid == OSM_HI_TEMPLATE_GUID_5) {
		p_physp->hierarchy_info.block = *p_hi;
		process_hierarchy_info_template_guid_5(sm, p_physp, template_guid);
	}

	if ((hierarchy_index + 1) <= p_hi->max_active_index) {
		osm_get_hierarchy_info(sm, p_node, p_port, p_physp, (hierarchy_index + 1));
	}

	/* if the block with the maximal index is received, all blocks were received */
	if (hierarchy_index == p_hi->max_active_index) {
		p_physp->hierarchy_info_updated = TRUE;
	}

Exit:
	/*
	 * Send next mad from the queue both on successful MAD and one with error status,
	 * in order not to disturb discovery.
	 */
	osm_port_send_next_mad(sm, p_port);

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}

osm_aports_system_t *osm_aports_system_new(IN const ib_node_info_t *p_ni,
					   IN osm_aports_system_type_t type)
{
	uint32_t size;
	osm_aports_system_t *p_aports_system;
	osm_switch_system_t *p_sw_system;
	osm_ca_system_t *p_ca_system;
	osm_aport_t *aports_array;

	OSM_ASSERT(type == OSM_APORTS_SYSTEM_TYPE_CA ||
		   type == OSM_APORTS_SYSTEM_TYPE_SWITCH);

	/*
	 * Create number of aports as the maximal possible number,
	 * which is equal to number of ports.
	 * The real number of aports will be most likely less than that number.
	 */
	size = p_ni->num_ports;
	aports_array = malloc(sizeof(osm_aport_t) * size);
	if (!aports_array)
		return NULL;
	memset(aports_array, 0, sizeof(osm_aport_t) * size);

	/* Allocate the switch/CA subclass, as it contains the aports system parent class */
	if (type == OSM_APORTS_SYSTEM_TYPE_SWITCH) {
		p_sw_system = malloc(sizeof(*p_sw_system));
		if (!p_sw_system) {
			free(aports_array);
			return NULL;
		}

		memset(p_sw_system, 0, sizeof(*p_sw_system));

		p_aports_system = &p_sw_system->aports;
		p_aports_system->guid_id = p_ni->sys_guid;
	} else {
		p_ca_system = malloc(sizeof(*p_ca_system));
		if (!p_ca_system) {
			free(aports_array);
			return NULL;
		}

		memset(p_ca_system, 0, sizeof(*p_ca_system));

		p_aports_system = &p_ca_system->aports;
		p_aports_system->guid_id = p_ni->node_guid;
	}

	p_aports_system->aport_table = aports_array;
	p_aports_system->type = type;
	p_aports_system->aports_number = size;

	return p_aports_system;
}

static void osm_build_aport(cl_map_item_t * const p_item, void *context)
{
	cl_qmap_t *p_aports_system_tbl;
	osm_node_t *p_node = (osm_node_t *)p_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	uint8_t port_num;
	osm_physp_t *p_physp;
	osm_hierarchy_info_t *p_hierarchy = NULL;
	ib_node_info_t *node_info = &p_node->node_info;
	ib_net64_t guid_id;
	osm_aports_system_t *p_aports_system;
	osm_switch_system_t *p_sw_system = NULL;
	osm_ca_system_t *p_ca_system = NULL;
	osm_aport_t *p_aport;
	osm_aports_system_type_t aports_system_type;
	uint8_t num_of_ports, num_internal_fnm_ports = 0;
	uint8_t plane_mask = 0;

	num_of_ports = osm_node_get_num_physp(p_node);

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
		aports_system_type = OSM_APORTS_SYSTEM_TYPE_CA;
		p_aports_system_tbl = &sm->p_subn->aports_systems_tables.ca_aports_systems;
		guid_id = node_info->node_guid;
	} else {
		aports_system_type = OSM_APORTS_SYSTEM_TYPE_SWITCH;
		p_aports_system_tbl = &sm->p_subn->aports_systems_tables.switch_aports_systems;
		guid_id = node_info->sys_guid;
	}

	for (port_num = 1; port_num < num_of_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);

		if (!p_physp)
			continue;

		if (osm_planar_link_is_internal_system_fnm_link(sm->p_subn, p_physp))
			num_internal_fnm_ports++;

		/* this port has no valid planarization info */
		if (!osm_hierarchy_is_planarized_physp(p_physp))
			continue;

		p_hierarchy = &p_physp->hierarchy_info;

		plane_mask |= (1 << p_hierarchy->plane_number);
		p_aports_system = (osm_aports_system_t *) cl_qmap_get(p_aports_system_tbl, guid_id);
		if (p_aports_system == (osm_aports_system_t *)cl_qmap_end(p_aports_system_tbl)) {
			p_aports_system = osm_aports_system_new(&p_node->node_info,
								aports_system_type);
			if (PF(p_aports_system == NULL)) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A003: "
					"Unable to create new aports system object\n");
				return;
			}
			cl_qmap_insert(p_aports_system_tbl, guid_id, &p_aports_system->map_item);
		}

		/* for precaution, all nodes of aports system are supposed to have the same number of ports */
		if (p_hierarchy->aport >= p_aports_system->aports_number) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A006: "
				"Node 0x%" PRIx64 " port num %u's  hierarchy info with aport %u, "
				"that is greater than number of ports (%u) on nodes of %s "
				"aports system 0x%" PRIx64 ".\n",
				osm_node_get_node_guid_ho(p_node), port_num, p_hierarchy->aport,
				p_aports_system->aports_number,
				osm_aports_system_get_type_str(p_aports_system),
				cl_ntoh64(guid_id));
			continue;
		}

		if (aports_system_type == OSM_APORTS_SYSTEM_TYPE_SWITCH) {
			p_sw_system = (osm_switch_system_t *)p_aports_system;
		} else {
			p_ca_system = (osm_ca_system_t *)p_aports_system;
		}

		p_aport = &p_aports_system->aport_table[p_hierarchy->aport];

		p_aport->p_aports_system = p_aports_system;

		/* all ports on the node are supposed to have the same asic number, so we can use last port's data */
		if (p_hierarchy->asic > -1 && aports_system_type == OSM_APORTS_SYSTEM_TYPE_SWITCH) {
			p_sw_system->nodes[p_hierarchy->asic] = p_node;
			if (!p_sw_system->min_asic_available ||
			    p_sw_system->min_asic_available > p_hierarchy->asic)
				p_sw_system->min_asic_available = p_hierarchy->asic;
			if (p_sw_system->max_asic_available < p_hierarchy->asic) {
				p_sw_system->max_asic_available = p_hierarchy->asic;
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"node GUID 0x%016" PRIx64 " asic %d p_sw_system->max_asic %d\n",
					osm_node_get_node_guid_ho(p_node),
					p_hierarchy->asic, p_sw_system->max_asic_available);
			}
		} else if (aports_system_type == OSM_APORTS_SYSTEM_TYPE_CA) {
			p_ca_system->node = p_node;
		}

		/* check if aport has already physp there */
		if (p_aport->plane_ports[p_hierarchy->plane_number]) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A005: "
				"%s aports system 0x%" PRIx64 " node 0x%" PRIx64
				" APort %d port %u has the same plane %d as port %u.\n",
				osm_aports_system_get_type_str(p_aports_system),
				cl_ntoh64(guid_id), osm_node_get_node_guid_ho(p_node),
				p_hierarchy->aport, p_physp->port_num, p_hierarchy->plane_number,
				p_aport->plane_ports[p_hierarchy->plane_number]->port_num);
		} else {
			p_aport->plane_ports[p_hierarchy->plane_number] = p_physp;

			/* update reference from p_physp to its aport */
			p_hierarchy->p_aport = p_aport;

			if (p_hierarchy->aport > p_aports_system->max_aport_available)
				p_aports_system->max_aport_available = p_hierarchy->aport;

			p_aport->aport_num = p_hierarchy->aport;

			if (p_hierarchy->plane_number >	p_aport->max_found_plane_number)
				p_aport->max_found_plane_number = p_hierarchy->plane_number;

			if (p_aport->min_found_plane_number == 0 ||
			    p_hierarchy->plane_number < p_aport->min_found_plane_number)
				p_aport->min_found_plane_number = p_hierarchy->plane_number;

			if (p_aport->min_port_num == 0 || port_num < p_aport->min_port_num)
				p_aport->min_port_num = port_num;
		}
	}
	if (p_node->sw) {
		p_node->sw->plane_mask = plane_mask;
	}

	if (p_sw_system)
		p_sw_system->num_internal_fnm_ports += num_internal_fnm_ports;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "node GUID 0x%016" PRIx64 " plane mask 0x%x\n",
			osm_node_get_node_guid_ho(p_node), plane_mask);
}

void osm_aports_system_delete(IN OUT osm_aports_system_t ** p_aports_system)
{
	OSM_ASSERT(p_aports_system && *p_aports_system);

	if ((*p_aports_system)->aport_table) {
		free((*p_aports_system)->aport_table);
		(*p_aports_system)->aport_table = NULL;
	}

	if ((*p_aports_system)->type == OSM_APORTS_SYSTEM_TYPE_SWITCH) {
		free((osm_switch_system_t *)(*p_aports_system));
	} else {
		free((osm_ca_system_t *)(*p_aports_system));
	}

	*p_aports_system = NULL;
}

extern void osm_subn_destroy_aports_systems_tables(IN osm_subn_t * p_subn);
extern void osm_switch_clear_planes(cl_map_item_t *p_item, void *ctx);

void osm_build_aports(osm_sm_t * sm)
{
	uint32_t aports_systems_num = 0;

	OSM_LOG_ENTER(sm->p_log);

	/* clear previous db */
	osm_subn_destroy_aports_systems_tables(sm->p_subn);

	/* clear previous plane bit mask from switches */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, osm_switch_clear_planes, NULL);

	cl_qmap_apply_func(&sm->p_subn->node_guid_tbl, osm_build_aport, sm);

	aports_systems_num =
		cl_qmap_count(&sm->p_subn->aports_systems_tables.switch_aports_systems) +
		cl_qmap_count(&sm->p_subn->aports_systems_tables.ca_aports_systems);
	if (aports_systems_num) {
		sm->p_subn->is_planarized = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Planarized systems found. %u aports systems.\n",
			aports_systems_num);
	}

	OSM_LOG_EXIT(sm->p_log);
}

osm_aport_iter_t osm_aports_system_end(IN osm_aports_system_t * p_aports_system)
{
	/*
	 * HACK: Compiler warns p_aports_system is unused. Fake a usage to avoid warning.
	 * p_aports_system is kept as a parameter in case it is needed in the future.
	 */
	(void)p_aports_system;

	return NULL;
}

osm_aport_iter_t osm_aports_system_begin(IN osm_aports_system_t * p_aports_system)
{
	osm_aport_t *p_aport;
	uint8_t aport_num;

	for (aport_num = 1; aport_num <= p_aports_system->max_aport_available; aport_num++) {
		p_aport = &p_aports_system->aport_table[aport_num];
		if (!osm_aport_is_valid(p_aport))
			continue;
		return p_aport;
	}
	return osm_aports_system_end(p_aports_system);
}

osm_aport_iter_t osm_aport_iter_next(IN osm_aport_iter_t aport_iter)
{
	osm_aport_t *p_aport = (osm_aport_t *)aport_iter;
	osm_aports_system_t *p_aports_system = p_aport->p_aports_system;
	uint8_t aport_num;

	if (!osm_aport_is_valid(p_aport))
		return osm_aports_system_end(p_aports_system);

	for (aport_num = p_aport->aport_num + 1; aport_num <= p_aports_system->max_aport_available;
	     aport_num++) {
		p_aport = &p_aports_system->aport_table[aport_num];
		if (!osm_aport_is_valid(p_aport))
			continue;
		return p_aport;
	}
	return osm_aports_system_end(p_aports_system);
}

static void osm_unhealthy_aport_process(IN osm_sm_t *sm,
					IN osm_aport_t *p_aport,
					IN osm_physp_t *p_plane_uh_physp,
					IN osm_unhealthy_port_t *p_plane_uhp)
{
	uint8_t plane_port;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_unhealthy_port_t *p_uhp_by_aport, *p_uhp_exist;
	cl_fmap_item_t *p_item;

	for (plane_port = 1; plane_port <= p_aport->max_found_plane_number; plane_port++) {
		p_physp = p_aport->plane_ports[plane_port];

		/* skip the originally found unhealthy port */
		if (!p_physp || p_physp == p_plane_uh_physp)
			continue;

		p_node = p_physp->p_node;

		p_uhp_by_aport =
			osm_unhealthy_port_new(0, 0, NULL, 0,
					osm_node_get_node_guid(p_node),
					p_physp->port_num,
					p_node->print_desc);

		if (!p_uhp_by_aport) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A00E: "
				"Couldn't create new unhealthy port UNKNOWN[0]<-->0x%016"
				PRIx64 "[%d]\n",
				osm_node_get_node_guid(p_node),
				p_physp->port_num);
			return;
		}

		p_uhp_by_aport->cond_mask = p_plane_uhp->cond_mask;
		p_uhp_by_aport->report = p_plane_uhp->report;
		p_uhp_by_aport->ignore = p_plane_uhp->ignore;
		p_uhp_by_aport->disable = p_plane_uhp->disable;
		p_uhp_by_aport->isolate = p_plane_uhp->isolate;
		p_uhp_by_aport->no_discover = p_plane_uhp->no_discover;
		p_uhp_by_aport->manual_action = p_plane_uhp->manual_action;

		p_item = cl_fmap_insert(&sm->p_subn->unhealthy_ports_tbl,
				&p_uhp_by_aport->p_obj,
				&p_uhp_by_aport->fmap_item);

		uint8_t aport_num = p_physp->hierarchy_info.aport;

		if (p_item != &p_uhp_by_aport->fmap_item) {
			p_uhp_exist = (osm_unhealthy_port_t *)p_item;
			osm_unhealthy_port_remove(sm, &p_uhp_by_aport);
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A00F: "
				"APort %u on System GUID 0x%" PRIx64
				" node GUID 0x%" PRIx64	" port number %u "
				"failed to add unhealthy, existing 0x%" PRIx64
				" port %u peer 0x%" PRIx64 " port %u.\n",
				aport_num, cl_ntoh64(p_physp->p_node->node_info.sys_guid),
				osm_node_get_node_guid_ho(p_physp->p_node),
				p_physp->port_num, cl_ntoh64(p_uhp_exist->guid), p_uhp_exist->port_num,
				cl_ntoh64(p_uhp_exist->peer_guid), p_uhp_exist->peer_port_num);
		} else {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"APort %u on System GUID 0x%" PRIx64
				" node GUID 0x%" PRIx64	" port number %u "
				"succeeded to add unhealthy\n",
				aport_num, cl_ntoh64(p_physp->p_node->node_info.sys_guid),
				osm_node_get_node_guid_ho(p_physp->p_node),
				p_physp->port_num);
		}
	}
}

static void osm_analyze_aports(cl_map_item_t * const p_item, void *context)
{
	osm_aports_system_t *p_aports_system = (osm_aports_system_t *)p_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_aport_iter_t aport_iter;
	osm_aport_t *p_aport;
	osm_physp_t *p_physp, *p_first_unhealthy_physp;
	osm_node_t *p_node, *p_sm_node;
	uint8_t first_aport_num_of_planes, number_of_found_planes;
	uint8_t first_plane_port, plane_port, first_port, sm_port_number, first_unhealthy_plane_port;
	uint8_t first_port_state = 0, phys_port_state = 0;
	uint16_t first_port_link_speed_active, port_link_speed_active;
	uint8_t first_link_width_active = 0, first_mtu_cap = 0, first_vl_cap = 0;
	boolean_t first_port_found = FALSE;
	osm_unhealthy_port_t *p_uhp, *p_first_uhp;
	boolean_t is_partially_unhealthy;
	boolean_t has_fdr10 = (sm->p_subn->opt.mepi_enabled_speeds & FDR10_BIT) ? TRUE : FALSE;
	uint64_t first_node_guid;
	boolean_t is_sm_aport;

	OSM_LOG_ENTER(sm->p_log);

	for (aport_iter = osm_aports_system_begin(p_aports_system);
	     aport_iter != osm_aports_system_end(p_aports_system);
	     aport_iter = osm_aport_iter_next(aport_iter)) {
		p_aport = (osm_aport_t *)aport_iter;

		if (!osm_aport_is_valid(p_aport))
			continue;

		/* reset for new aport */
		first_port_state = phys_port_state = first_port_link_speed_active = port_link_speed_active =
		first_link_width_active = first_mtu_cap = first_vl_cap = first_aport_num_of_planes =
		number_of_found_planes = first_node_guid = first_plane_port = first_port = sm_port_number =
			first_unhealthy_plane_port = 0;

		first_port_found = FALSE;
		p_sm_node = NULL;
		is_sm_aport = FALSE;
		p_first_uhp = NULL;
		is_partially_unhealthy = FALSE;
		p_first_unhealthy_physp = NULL;

		for (plane_port = 1; plane_port <= p_aport->max_found_plane_number; plane_port++) {
			p_physp = p_aport->plane_ports[plane_port];

			if (!p_physp || !p_physp->p_node) {
				continue;
			}

			p_node = p_physp->p_node;

			if (!is_sm_aport && is_sm_port(sm, p_node, p_physp->port_num)) {
				is_sm_aport = TRUE;
				p_sm_node = p_node;
				sm_port_number = p_physp->port_num;
			}

			/*
			 * look at parameters:
			 * Relevant attribute values are:
			 * active speed, active width, MTU cap and VL cap.
			 */
			phys_port_state = ib_port_info_get_port_state(&p_physp->port_info);
			number_of_found_planes++;

			p_uhp = osm_hm_get_unhealthy_port_by_guid(sm,
								  osm_node_get_node_guid(p_node),
								  p_physp->port_num);

			if (p_uhp && (p_uhp->isolate || p_uhp->no_discover)) {
				/*
				 * if all plane ports of this APort are unhealthy, then everything
				 * is symmetrical.
				 */
				if (!p_first_uhp) {
					/*
					 * This section is applied only for the first unhealthy plane port
					 * found for this APort
					 */
					p_first_uhp = p_uhp;
					first_unhealthy_plane_port = plane_port;
					p_first_unhealthy_physp = p_physp;

					/* other healthy planes exist on this APort */
					if (number_of_found_planes > 1)
						is_partially_unhealthy = TRUE;
				}

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found unhealthy plane %u on APort %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64	" port number %u, "
					"isolate %u no_discover %u (APort's partially unhealthy %u)\n",
					plane_port,
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num,
					p_uhp->isolate, p_uhp->no_discover, is_partially_unhealthy);

			} else if (!p_uhp && p_first_uhp) {
				/* found APort that some plane ports are healthy and some are not */
				is_partially_unhealthy = TRUE;

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found partially unhealthy APort %u on System GUID 0x%" PRIx64
					", healthy node GUID 0x%" PRIx64 " port number %u plane %u, "
					"while unhealthy for plane %u: isolate %u no_discover %u.\n",
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num, plane_port, first_unhealthy_plane_port,
					p_first_uhp->isolate, p_first_uhp->no_discover);
			}

			/* first found port, lets remember its values for comparison */
			if (!first_port_found) {
				first_plane_port = plane_port;
				first_port = p_physp->port_num;
				first_port_state = phys_port_state;
				first_port_link_speed_active =
					osm_physp_get_highest_active_speed(p_physp, has_fdr10);
				first_link_width_active = p_physp->port_info.link_width_active;
				first_mtu_cap = ib_port_info_get_mtu_cap(&p_physp->port_info);
				first_vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
				first_port_found = TRUE;
				first_aport_num_of_planes = p_physp->hierarchy_info.num_of_planes;
				first_node_guid = osm_node_get_node_guid_ho(p_node);
				continue;
			}

			port_link_speed_active =
				osm_physp_get_highest_active_speed(p_physp, has_fdr10);

			/*
			 * Check if one of the plane ports has DOWN state while others don't.
			 * Check if the APort is not marked as asymmetric yet.
			 */
			if (!osm_hierarchy_is_asymmetric_aport(p_aport) &&
			    phys_port_state != first_port_state &&
			    (phys_port_state == IB_LINK_DOWN || first_port_state == IB_LINK_DOWN)) {
				p_aport->asymmetric = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Found state asymmetric APort %u on System GUID 0x%016" PRIx64
					" node GUID 0x%016" PRIx64 " port number %u plane %u: "
					"state %u compared to node GUID 0x%016" PRIx64
					" port number %u plane %u: state %u.\n",
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num, plane_port, phys_port_state,
					first_node_guid, first_port, first_plane_port,
					first_port_state);
			}

			/* Check if the APort is not marked as asymmetric yet. */
			if (!osm_hierarchy_is_asymmetric_aport(p_aport) &&
			    (port_link_speed_active != first_port_link_speed_active ||
			    p_physp->port_info.link_width_active != first_link_width_active ||
			    ib_port_info_get_mtu_cap(&p_physp->port_info) != first_mtu_cap ||
			    ib_port_info_get_vl_cap(&p_physp->port_info) != first_vl_cap ||
			    p_physp->hierarchy_info.num_of_planes != first_aport_num_of_planes)) {

				p_aport->asymmetric = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A00C: "
					"Found attributes asymmetric APort %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64
					" port number %u plane %u: speed %u, width %u, mtu_cap %u, vl_cap %u,"
					" number of planes %u compared to node GUID 0x%016" PRIx64
					" port number %u plane %u: speed %u, width %u, mtu_cap %u, vl_cap %u,"
					" number of planes %u.\n",
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num, plane_port,
					port_link_speed_active, p_physp->port_info.link_width_active,
					ib_port_info_get_mtu_cap(&p_physp->port_info),
					ib_port_info_get_vl_cap(&p_physp->port_info),
					p_physp->hierarchy_info.num_of_planes,
					first_node_guid, first_port, first_plane_port,
					first_port_link_speed_active, first_link_width_active,
					first_mtu_cap, first_vl_cap, first_aport_num_of_planes);
			}

		} /* end of planes loop */

		/* make sure to mark all other plane ports of this APort as unhealthy */
		if (is_partially_unhealthy) {
			osm_unhealthy_aport_process(sm, p_aport,
						    p_first_unhealthy_physp, p_first_uhp);
		}

		/*
		 * Check APort that wasn't marked yet as asymmetric,
		 * that all plane ports exist for it.
		 */
		if (!osm_hierarchy_is_asymmetric_aport(p_aport) &&
		    (!number_of_found_planes || number_of_found_planes != first_aport_num_of_planes)) {
			p_aport->asymmetric = TRUE;
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Found asymmetric APort %u on %s aports system ID 0x%016" PRIx64
				" with missing planes. Found %u planes out of %u planes.\n",
				p_aport->aport_num, osm_aports_system_get_type_str(p_aports_system),
				cl_ntoh64(p_aports_system->guid_id),
				number_of_found_planes, first_aport_num_of_planes);
		}

		if (osm_hierarchy_is_asymmetric_aport(p_aport) && is_sm_aport) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A00D: SM port/peer "
				"0x%" PRIx64 "[%u] found to belong to asymmetric APort %u on %s "
				"aports system 0x%" PRIx64 ". Ignoring this state (Not allowed).\n",
				osm_node_get_node_guid_ho(p_sm_node), sm_port_number,
				p_aport->aport_num, osm_aports_system_get_type_str(p_aports_system),
				cl_ntoh64(p_aports_system->guid_id));
			p_aport->asymmetric = FALSE;
		}

	} /* end of APorts loop */

	OSM_LOG_EXIT(sm->p_log);
}

static void build_dest_lids_to_groups(
		IN osm_subn_t *p_subn,
		OUT uint16_t p_lid_to_ar_groups[IB_LID_UCAST_END_HO + 1][OSM_PLANAR_MAX_NUM_OF_ASICS + 1],
		OUT uint16_t p_dense_lids[IB_LID_UCAST_END_HO + 1],
		OUT unsigned int *p_max_num_of_dest_lids)
{
	cl_map_item_t *p_item;
	osm_switch_system_t *p_sw_system;
	osm_physp_t *p_physp, *p_remote_physp;
	osm_node_t *p_node, *p_remote_node;
	uint16_t group_id;
	uint8_t port_num, remote_port;
	uint16_t lid, sw_lid, remote_lid, max_remote_lid;
	ib_special_port_type_t special_type;
	uint8_t asic;
	unsigned int i = 0, max_num_of_dest_lids = 0;
	uint16_t num_lids = p_subn->max_lid + 1;

	/*
	 * Go over non-prism leaf switches and map HCA lids of asics of the same system
	 * to the ar_groups ids of those asics on this system.
	 */
	for (p_item = cl_qmap_head(&p_subn->aports_systems_tables.switch_aports_systems);
	     p_item != cl_qmap_end(&p_subn->aports_systems_tables.switch_aports_systems);
	     p_item = cl_qmap_next(p_item)) {
		p_sw_system = (osm_switch_system_t *)p_item;

		for (asic = p_sw_system->min_asic_available;
		     asic <= p_sw_system->max_asic_available;
		     asic ++) {
			if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
				continue;
			}

			/* skip prism switch system */
			if (osm_switch_is_prism_switch(p_sw_system->nodes[asic]->sw, NULL)) {
				break;
			}

			p_node = p_sw_system->nodes[asic];
			p_physp = osm_node_get_physp_ptr(p_node, 0);
			sw_lid = osm_node_get_base_lid_ho(p_node, 0);
			group_id = osm_ucast_mgr_get_groupid(&p_subn->p_osm->sm.ucast_mgr,
							     sw_lid, 0);

			/*
			 * This system doesn't have ar group assigned,
			 * maybe it is not leaf, skip it and move to next system.
			 */
			if (group_id == UCAST_MGR_INVALID_GROUP_ID)
				break;

			/* also add HCAs' lids, that are connected to this switch */
			for (port_num = 1; port_num < osm_node_get_num_physp(p_node); ++port_num) {
				p_physp = osm_node_get_physp_ptr(p_node, port_num);
				p_remote_node = osm_node_get_remote_node(p_node, port_num,
									 &remote_port);

				if (!p_remote_node || !p_physp || !osm_link_is_healthy(p_physp))
					continue;

				if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH)
					continue;

				p_remote_physp = osm_physp_get_remote(p_physp);

				/* don't add Aggregation Nodes or Routers as valid destination for AR/FR */
				special_type =
					osm_physp_get_special_port_type(p_physp->p_remote_physp);
				if (special_type == IB_SPECIAL_PORT_TYPE_AN ||
						special_type == IB_SPECIAL_PORT_TYPE_ROUTER)
					continue;

				remote_lid = osm_node_get_base_lid_ho(p_remote_node, remote_port);
				max_remote_lid = remote_lid + (1 << osm_physp_get_lmc(p_remote_physp));

				for (lid = remote_lid; lid < max_remote_lid; lid++) {
					p_lid_to_ar_groups[lid][asic] = group_id;
					/* at index 0 mark that values for this lid exist for some asic */
					p_lid_to_ar_groups[lid][0] = group_id;
				}
			}
		} /* end of asics loop */
	} /* end of aports systems loop */

	/* fill compressed array of destination lids */
	for (lid = 0, i = 0; lid < num_lids; lid++) {
		if (p_lid_to_ar_groups[lid][0] != AR_INVALID_GROUP_ID) {
			p_dense_lids[i] = lid;
			i++;
			max_num_of_dest_lids++;
		}
	}

	*p_max_num_of_dest_lids = max_num_of_dest_lids;
}

void osm_build_system_ar_group(IN osm_subn_t *p_subn)
{
	uint16_t union_group_id, group_id, orig_group_id, group_top, group_cap;
	uint8_t port_num, state;
	uint16_t lid;
	osm_switch_t *p_sw;
	osm_advanced_routing_data_t *p_ar_data;
	cl_map_item_t *p_sw_item;
	uint16_t lids = p_subn->max_lid + 1;
	uint16_t (*lid_to_ar_groups)[OSM_PLANAR_MAX_NUM_OF_ASICS + 1];
	uint16_t *dense_lids;
	unsigned int i, max_num_of_dest_lids = 0, asic = 0;
	osm_ar_subgroup_t *p_subgroup, *p_union_subgroup;
	boolean_t union_group_created;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	lid_to_ar_groups = malloc(sizeof(uint16_t[lids][OSM_PLANAR_MAX_NUM_OF_ASICS + 1]));

	if (!lid_to_ar_groups) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A00C: "
			"Failed to allocate memory\n");
		goto Exit2;
	}

	dense_lids = malloc(lids * sizeof(uint16_t));
	if (!dense_lids) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A00D: "
			"Failed to allocate memory\n");
		goto Exit1;
	}

	/*
	 * Maps lids of the leaf switches (switches of the non-prism planarized system)
	 * to their ar_group id.
	 */
	memset(lid_to_ar_groups, AR_INVALID_GROUP, sizeof(uint16_t[lids][OSM_PLANAR_MAX_NUM_OF_ASICS + 1]));
	memset(dense_lids, 0, (lids * sizeof(uint16_t)));

	build_dest_lids_to_groups(p_subn, lid_to_ar_groups, dense_lids, &max_num_of_dest_lids);
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"max_num_of_dest_lids %u\n", max_num_of_dest_lids);

	/*
	 * On every prism switch create one union ar group for non-prism destination leaf.
	 * Go over leaf switches, and their HCAs, and build this union ar group,
	 * take the id of the smallest asic in BM and assign it to this ar group,
	 * on this prism switch. Make other asics' ar groups empty on this switch.
	 */
	for (p_sw_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw_item = cl_qmap_next(p_sw_item)) {
		p_sw = (osm_switch_t *)p_sw_item;

		if (!osm_switch_is_prism_switch(p_sw, NULL) ||
		    p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED) {
			continue;
		}

		p_ar_data = &p_sw->ar_data;
		group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);
		group_top = ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info);
		union_group_id = AR_INVALID_GROUP_ID;

		for (i = 0; i <= max_num_of_dest_lids; i++) {
			union_group_created = FALSE;

			/* skip switch that doesn't have capability to assign id for this new union group */
			if (group_top + 1 > group_cap) {
				break;
			}

			/* destination lid */
			lid = dense_lids[i];

			/* TBD: currently supports only plft 0 (ar_updn, not dfp) */
			osm_ar_lft_entry_get(p_sw, 0, lid, &port_num, &state, &orig_group_id);

			/* no routing for this lid */
			if (port_num == OSM_NO_PATH || orig_group_id == AR_INVALID_GROUP_ID) {
				continue;
			}

			/* go over groups of this destination lid and create union of all these groups */

			/* TBD: This is currently hack.
			 * The union group id (per system) has to be unique over the fabric,
			 * for not to break pfrn,
			 * so that if some other switch uses the same group id for another system,
			 * we won't make pfrn storm by sending this group id.
			 * This will be resolved when all the asics of the same leaf system
			 * will get the same group id.
			 */
			union_group_id = group_top + 1;
			/* TBD: currently supports only plft 0 (ar_updn, not dfp) and only sub group 0 */
			p_union_subgroup = osm_ar_subgroup_get(p_ar_data, union_group_id, 0);

			/*
			 * Create a separate group id for this union group.
			 * TBD: do not create separate AR group id for BlackMamba system,
			 * to use already existing group id that is used for one of its asics.
			 */
			for (asic = 1; asic <= OSM_PLANAR_MAX_NUM_OF_ASICS; asic++) {
				group_id = lid_to_ar_groups[lid][asic];

				if (group_id == AR_INVALID_GROUP_ID)
					continue;

				/* TBD: currently supports only plft 0 (ar_updn, not dfp) */
				p_subgroup = osm_ar_subgroup_get(p_ar_data, group_id, 0);
				osm_ar_subgroup_add_subgroup(p_union_subgroup, p_subgroup);
				union_group_created = TRUE;
			}

			if (union_group_created) {
				/* update group_top for this switch */
				group_top = union_group_id;
				osm_ar_set_required_group_top(p_ar_data, group_top);
				/*
				 * Change the group id to group, which is union towards
				 * multi asic, single plane asic system.
				 */
				osm_ar_lft_set(p_ar_data, lid, 0, port_num, state, union_group_id);
			}

		} /* end of destination LIDs loop */
	} /* end of switches loop */

	free(dense_lids);
Exit1:
	free(lid_to_ar_groups);
Exit2:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_build_planarized_hcas_ar_groups(IN osm_subn_t * p_subn)
{
	cl_qmap_t *p_aports_systems_tbl;
	cl_map_item_t *p_item;
	osm_aports_system_t *p_aports_system, *p_remote_aports_system;
	osm_node_t *p_node;
	osm_switch_t *p_remote_sw;
	osm_aport_iter_t aport_iter;
	osm_aport_t *p_aport, *p_remote_aport;
	osm_physp_t *p_physp;
	osm_advanced_routing_data_t *p_ar_data;
	osm_ar_subgroup_t *p_subgroup;
	uint16_t group_id, prev_group_id, group_cap, lid, max_lid;
	uint8_t prev_static_port_num, plane_port_num, plane, prev_state;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/*
	 * Iterating over planarized HCAs and getting their remote switch is faster than iterating
	 * over prism switch aports.
	 */
	p_aports_systems_tbl = &p_subn->aports_systems_tables.ca_aports_systems;
	for (p_item = cl_qmap_head(p_aports_systems_tbl);
	     p_item != cl_qmap_end(p_aports_systems_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_aports_system = (osm_aports_system_t *)p_item;

		p_node = osm_get_node_by_guid(p_subn, p_aports_system->guid_id);
		if (!p_node)
			continue;

		/* To use AR, the HCA must support out-of-order receiving */
		if (p_subn->opt.enable_ar_by_device_cap &&
		    !ib_general_info_is_cap_enabled(&p_node->general_info,
						    IB_GENERAL_INFO_IS_GLOBAL_OOO_SUPPORTED_BIT))
			continue;

		for (aport_iter = osm_aports_system_begin(p_aports_system);
		     aport_iter != osm_aports_system_end(p_aports_system);
		     aport_iter = osm_aport_iter_next(aport_iter)) {
			p_aport = (osm_aport_t *)aport_iter;
			if (!osm_aport_is_valid(p_aport))
				continue;

			p_remote_aport = osm_aport_get_remote(p_aport);
			if (!p_remote_aport || !osm_aport_is_valid(p_remote_aport) ||
			    p_remote_aport->min_port_num == 0)
				continue;

			/* NULL if remote is not a prism switch */
			p_remote_sw = osm_aport_get_prism_switch(p_remote_aport);
			if (!p_remote_sw || p_remote_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
				continue;

			p_remote_aports_system = osm_aport_get_aports_system(p_remote_aport);
			if (p_remote_aport->max_found_plane_number ==
			    p_remote_aport->min_found_plane_number) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"APort %u of switch system 0x%" PRIx64 " has only one plane"
					"(%u). Setting static routing even if it's an APort from "
					"prism switch to planarized HCA.\n",
					p_remote_aport->aport_num,
					cl_ntoh64(p_remote_aports_system->guid_id),
					p_remote_aport->min_found_plane_number);
				continue;
			}

			/*
			 * Create (or get existing) AR group ID for the switch aport by using the
			 * lowest-numbered port in the aport as the aport's representative.
			 * That is a unique representative on this switch, but not unique between
			 * switches.
			 * That is fine, since this AR group ID is only used from a leaf switch to
			 * its HCAs, and not for routing between switches.
			 */
			group_id = osm_ucast_mgr_make_port_num_groupid(&p_subn->p_osm->sm.ucast_mgr,
								       p_remote_aport->min_port_num);
			if (group_id == UCAST_MGR_INVALID_GROUP_ID) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A010: "
					"Can't allocate AR group ID for port number %u, "
					"which is the representative of switch system 0x%" PRIx64
					" APort number %u (prism switch toward planarized HCA).\n",
					p_remote_aport->min_port_num,
					cl_ntoh64(p_remote_aports_system->guid_id),
					p_remote_aport->aport_num);
				continue;
			}

			p_ar_data = &p_remote_sw->ar_data;
			group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);
			if (group_id > group_cap) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"AR group ID %u allocated for switch system 0x%" PRIx64
					" APort number %u (prism switch toward planarized HCA) "
					"is over this node's group_cap (%u). Skipping this APort.\n",
					group_id,
					cl_ntoh64(p_remote_aports_system->guid_id),
					p_remote_aport->aport_num, group_cap);
				continue;
			}

			if (group_id > ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info)) {
				osm_ar_set_required_group_top(p_ar_data, group_id);
			}

			/* TODO: Support subgroup 1. Currently only subgroup 0 supported */
			p_subgroup = osm_ar_subgroup_get(p_ar_data, group_id, 0);
			memset(p_subgroup, 0, sizeof(osm_ar_subgroup_t));

			for (plane = p_remote_aport->min_found_plane_number;
			     plane <= p_remote_aport->max_found_plane_number;
			     plane++) {
				plane_port_num = osm_physp_get_port_num(p_remote_aport->plane_ports[plane]);
				osm_ar_subgroup_add_port(p_subgroup, plane_port_num);
			}

			/* LID should be the same for all aport planes. Check only one plane. */
			p_physp = p_aport->plane_ports[p_aport->min_found_plane_number];
			lid = cl_ntoh16(osm_physp_get_base_lid(p_physp));
			max_lid = lid + (1 << osm_physp_get_lmc(p_physp)) - 1;

			for (; lid <= max_lid; lid++) {
				/* TODO: Add support for plfts other than plft 0 */
				osm_ar_lft_entry_get(p_remote_sw, 0, lid, &prev_static_port_num,
						     &prev_state, &prev_group_id);

				/*
				 * NOTE: This function is called after checking validity of aports,
				 * and removing routing for invalid ones (asymmetric aports).
				 * Update AR group only for HCA that previously had routing
				 * calculated to them.
				 */
				if (prev_static_port_num == OSM_NO_PATH)
					continue;

				/* The static port should be part of the AR group */
				osm_ar_lft_set(p_ar_data, lid, 0, p_remote_aport->min_port_num,
					       IB_AR_LID_STATE_FREE, group_id);
			} /* End of lids loop */
		} /* End of aports loop */
	} /* End of aports systems loop */

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_analyze_aports_systems_aports(IN osm_sm_t * sm)
{
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.switch_aports_systems,
			   osm_analyze_aports, sm);
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.ca_aports_systems,
			   osm_analyze_aports, sm);
}

boolean_t osm_hierarchy_is_planarized_physp(IN const struct osm_physp * p_physp)
{
	return osm_physp_is_valid(p_physp) && p_physp->hierarchy_info.is_planarized;
}

osm_aport_t * osm_physp_get_aport(IN struct osm_physp * p_physp)
{
	return p_physp->hierarchy_info.p_aport;
}

osm_aport_t * osm_aport_get_remote(IN osm_aport_t * p_aport)
{
	/*
	 * TODO: Verify link validity, either here or in aport construction.
	 * All aport physps should lead to the same remote aport.
	 * current implementation does minimal checks.
	 */
	osm_physp_t *p_physp, *p_remote_physp;
	OSM_ASSERT(p_aport);
	OSM_ASSERT(osm_aport_is_valid(p_aport));

	p_physp = p_aport->plane_ports[p_aport->min_found_plane_number];
	if (!p_physp)
		return NULL;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		return NULL;

	return osm_physp_get_aport(p_remote_physp);
}

osm_aports_system_t * osm_aport_get_aports_system(IN osm_aport_t * p_aport)
{
	OSM_ASSERT(p_aport);
	OSM_ASSERT(osm_aport_is_valid(p_aport));
	return p_aport->p_aports_system;
}

osm_switch_t * osm_aport_get_prism_switch(IN osm_aport_t * p_aport)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;

	OSM_ASSERT(p_aport);
	OSM_ASSERT(osm_aport_is_valid(p_aport));

	p_physp = p_aport->plane_ports[p_aport->min_found_plane_number];
	p_sw = osm_physp_get_node_ptr(p_physp)->sw;
	if (!p_sw || !osm_switch_is_prism_switch(p_sw, NULL))
		return NULL;

	return p_sw;
}

boolean_t osm_hierarchy_is_asymmetric_aport_physp(IN osm_physp_t *p_physp)
{
	if (p_physp && osm_hierarchy_is_planarized_physp(p_physp) &&
	    osm_physp_get_aport(p_physp) &&
	    osm_hierarchy_is_asymmetric_aport(osm_physp_get_aport(p_physp)))
		return TRUE;

	return FALSE;
}

boolean_t osm_hierarchy_is_planarized(IN struct osm_switch *p_sw)
{
	return p_sw->plane_mask != 0;
}

osm_switch_t *
osm_hierarchy_get_sw_by_plane(osm_subn_t *p_subn, osm_switch_t *p_sw, uint8_t plane)
{
	cl_qmap_t *p_sw_aports_systems_tbl;
	osm_node_t *p_sw_node;
	osm_switch_t *p_req_sw = NULL;
	ib_net64_t sys_guid;
	osm_switch_system_t *p_sw_system;
	boolean_t is_planarized;
	uint8_t sw_plane = 0, node_idx, plane_mask;

	OSM_ASSERT(p_sw && p_sw->p_node);

	/*
	 * for non planarized system or for prism switch (has all planes),
	 * or the given switch is already what we need, just return the given switch
	 */
	is_planarized = osm_hierarchy_is_planarized(p_sw);

	if (!is_planarized ||
	    osm_switch_is_prism_switch(p_sw, &sw_plane) ||
	    sw_plane == plane) {
		p_req_sw = p_sw;
		goto Exit;
	}

	p_sw_aports_systems_tbl = &p_subn->aports_systems_tables.switch_aports_systems;
	sys_guid = p_sw->p_node->node_info.sys_guid;
	p_sw_system = (osm_switch_system_t *) cl_qmap_get(p_sw_aports_systems_tbl, sys_guid);
	/* given switch is planarized and non-prism switch, but we fail to find it */
	if (p_sw_system == (osm_switch_system_t *)cl_qmap_end(p_sw_aports_systems_tbl)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 0A2B: "
			"Failed to find planarized system for system GUID 0x%" PRIx64 "\n",
			cl_ntoh64(sys_guid));
		goto Exit;
	}

	plane_mask = 1 << plane;
	for (node_idx = p_sw_system->min_asic_available; node_idx <= p_sw_system->max_asic_available;
	     node_idx++) {
		p_sw_node = p_sw_system->nodes[node_idx];
		if (!p_sw_node || !p_sw_node->sw)
			continue;
		if (p_sw_node->sw->plane_mask == plane_mask) {
			/* found switch asic with desired plane, return it */
			p_req_sw = p_sw_node->sw;
			break;
		}
	}

Exit:
	return p_req_sw;
}

osm_port_t *
osm_hierarchy_get_sw_port0_by_plane(osm_subn_t *p_subn, osm_port_t *p_sw_port, uint8_t plane)
{
	osm_switch_t *p_req_sw;
	osm_port_t *p_port = NULL;

	OSM_ASSERT(p_sw_port && p_sw_port->p_node);

	/* for not a switch node just return NULL */
	if (!p_sw_port->p_node->sw)
		goto Exit;

	p_req_sw = osm_hierarchy_get_sw_by_plane(p_subn, p_sw_port->p_node->sw, plane);
	if (p_req_sw && p_req_sw->p_node) {
		/* found switch asic with desired plane, return its port */
		p_port =
			osm_get_port_by_guid(p_subn, p_req_sw->p_node->node_info.port_guid);
	}

Exit:
	return p_port;
}

osm_port_t *
osm_hierarchy_get_an_by_plane(osm_subn_t * p_subn,
			      osm_switch_t *p_sw,
			      osm_port_t *p_an_port,
			      uint8_t plane)
{
	osm_switch_t *p_plane_sw;
	osm_node_t *p_plane_sw_node;
	uint8_t sw_plane, port_num;
	boolean_t is_planarized;
	osm_physp_t *p_physp, *p_remote_physp;
	osm_port_t *p_plane_an_port = NULL;

	if (!p_an_port || !p_sw)
		goto Exit;

	is_planarized = osm_hierarchy_is_planarized(p_sw);

	/*
	 * For non planarized or prism switch or switch has the required plane
	 * return the received AN port.
	 */
	if (!is_planarized || osm_switch_is_prism_switch(p_sw, &sw_plane) ||
	    sw_plane == plane) {
		p_plane_an_port = p_an_port;
		goto Exit;
	}

	p_plane_sw = osm_hierarchy_get_sw_by_plane(p_subn, p_sw, plane);
	if (!p_plane_sw || !p_plane_sw->p_node)
		goto Exit;

	p_plane_sw_node = p_plane_sw->p_node;

	for (port_num = osm_node_get_num_physp(p_plane_sw_node) - 1; port_num > 0; port_num--) {
		p_physp = osm_node_get_physp_ptr(p_plane_sw_node, port_num);
		if (p_physp &&
		    ib_mlnx_epi_get_special_port_type(&p_physp->ext_port_info) ==
			IB_SPECIAL_PORT_TYPE_AN) {
			p_remote_physp = osm_physp_get_remote(p_physp);
			p_plane_an_port = osm_get_port_by_guid(p_subn, p_remote_physp->port_guid);
			break;
		}
	}

Exit:
	return p_plane_an_port;
}

uint8_t osm_hierarchy_get_num_of_planes(IN osm_physp_t *p_physp)
{
	return p_physp->hierarchy_info.num_of_planes < 0 ?
		0 : p_physp->hierarchy_info.num_of_planes;
}

boolean_t osm_hierarchy_is_port_allowed_to_route(IN osm_subn_t *p_subn,
						 IN osm_switch_t *p_sw,
						 IN osm_physp_t *p_physp,
						 IN osm_switch_t *p_other_sw)
{
	boolean_t is_prism_sw = FALSE;
	boolean_t is_sibling_sw = FALSE;
	boolean_t is_fnm_port = FALSE;
	uint8_t sw_plane = 0;

	/*
	 * For planarized non-prism switch routing to other switches on the same system (sibling)
	 * is allowed only through FNM (aka service) ports.
	 * Routing to any other destination switch is forbidden through FNM port.
	 * On planarized non-prism switch (ex. Black Mamba) port is allowed for routing if:
	 * destination is sibling and the port is fnm port
	 * or
	 * destination is not sibling and the port is not fnm port
	 */
	if (osm_hierarchy_is_planarized(p_sw)) {
		is_prism_sw =
			osm_switch_is_prism_switch(p_sw, &sw_plane);
		/* This is planarized non-prism switch (ex. Black Mamba) */
		if (!is_prism_sw && sw_plane) {
			is_sibling_sw =
				(p_sw->p_node->node_info.sys_guid ==
				 p_other_sw->p_node->node_info.sys_guid);
			is_fnm_port = osm_physp_is_fnm(p_subn, p_physp);

			if ((is_sibling_sw && is_fnm_port) || (!is_sibling_sw && !is_fnm_port))
				return TRUE;
			return FALSE;
		}
	}

	/* all other cases: non planarized or prism switch there is no limitations on routing */
	return TRUE;
}
