/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MADS_ENGINE_C
#include <opensm/osm_mads_engine.h>
#include <opensm/osm_opensm.h>


static void mads_engine_poller_send(osm_mads_engine_t * p_mads_engine, osm_madw_t * p_madw)
{
	osm_opensm_t *p_osm = p_mads_engine->p_osm;
	cl_status_t sts;
	ib_api_status_t status;
	uint8_t mgmt_class = p_madw->p_mad->mgmt_class;
	osm_madw_context_t mad_context = p_madw->context;
	uint16_t attr_id = p_madw->p_mad->attr_id;
	boolean_t resp_expected;

	if (mgmt_class != p_mads_engine->mgmt_class) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B202: "
			"%s manager got MAD of management class (%u). "
			"Failed to send mad\n", p_mads_engine->class_name, mgmt_class);
		return;
	}

	resp_expected = p_madw->resp_expected;
	status = osm_vendor_send(p_mads_engine->bind_handle[p_osm->subn.sm_port_idx], p_madw, p_madw->resp_expected);
	if (status == IB_SUCCESS) {
		p_mads_engine->statistic_counter_increase(p_osm, attr_id);
		if (resp_expected)
			cl_atomic_inc(&p_mads_engine->outstanding_mads_on_wire);
		while (p_mads_engine->outstanding_mads_on_wire >
		       (int32_t)p_mads_engine->max_outstanding_mads_on_wire) {
			sts = CL_ERROR;
			while (sts != CL_SUCCESS) {
				sts = cl_event_wait_on(&p_mads_engine->sig_mads_on_wire_continue,
						       EVENT_NO_TIMEOUT, TRUE);
			}
		}
	} else {
		p_mads_engine->send_mad_failure_cb(&mad_context, &p_osm->log);
	}
}

static void mads_engine_poller(void *p_ptr)
{
	osm_mads_engine_t *p_mads_engine = p_ptr;
	osm_opensm_t *p_osm = p_mads_engine->p_osm;
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_mads_engine->thread_state == OSM_THREAD_STATE_NONE)
		p_mads_engine->thread_state = OSM_THREAD_STATE_RUN;

	while (p_mads_engine->thread_state == OSM_THREAD_STATE_RUN) {
		CL_SPINLOCK_ACQUIRE(&p_mads_engine->mad_queue_lock);
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_mads_engine->mad_queue);
		CL_SPINLOCK_RELEASE(&p_mads_engine->mad_queue_lock);

		if (p_madw != (osm_madw_t *) cl_qlist_end(&p_mads_engine->mad_queue)) {
			p_mads_engine->dump_mad(p_madw, &p_osm->log);
			mads_engine_poller_send(p_mads_engine, p_madw);
		} else {
			cl_event_wait_on(&p_mads_engine->poller_wakeup,
					 EVENT_NO_TIMEOUT, TRUE);
		}
	}

	OSM_LOG_EXIT(&p_osm->log);
}

cl_status_t osm_mads_engine_init(osm_mads_engine_t *p_mads_engine, IN osm_opensm_t *p_osm,
				 IN osm_mad_pool_t *p_mad_pool,
				 IN const char *const class_name)
{
	cl_status_t status = CL_SUCCESS;
	int i;

	p_mads_engine->p_osm = p_osm;
	p_mads_engine->mad_pool = p_mad_pool;

	cl_qlist_init(&p_mads_engine->mad_queue);

	status = cl_spinlock_init(&p_mads_engine->mad_queue_lock);
	if (status != CL_SUCCESS)
		goto Exit;

	cl_event_construct(&p_mads_engine->poller_wakeup);
	status = cl_event_init(&p_mads_engine->poller_wakeup, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	cl_event_construct(&p_mads_engine->outstanding_mads_done_event);
	status = cl_event_init(&p_mads_engine->outstanding_mads_done_event, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	cl_event_construct(&p_mads_engine->sig_mads_on_wire_continue);
	status = cl_event_init(&p_mads_engine->sig_mads_on_wire_continue, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	p_mads_engine->thread_state = OSM_THREAD_STATE_NONE;

	status = cl_thread_init(&p_mads_engine->poller, mads_engine_poller, p_mads_engine, class_name);
	if (status != CL_SUCCESS)
		goto Exit;

	for (i = 0; i < OSM_MAX_BINDING_PORTS; i++) {
		p_mads_engine->bind_handle[i] = OSM_BIND_INVALID_HANDLE;
	}

	p_mads_engine->class_name = strdup(class_name);
Exit:
	return status;
}

void osm_mads_engine_destroy(osm_mads_engine_t *mads_engine)
{ 
	osm_madw_t *p_madw;

	mads_engine->thread_state = OSM_THREAD_STATE_EXIT;

	cl_event_signal(&mads_engine->sig_mads_on_wire_continue);
	cl_event_signal(&mads_engine->poller_wakeup);

	cl_thread_destroy(&mads_engine->poller);

	CL_SPINLOCK_ACQUIRE(&mads_engine->mad_queue_lock);
	while (!cl_is_qlist_empty(&mads_engine->mad_queue)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&mads_engine->mad_queue);
		osm_mad_pool_put(mads_engine->mad_pool, p_madw);
	}
	CL_SPINLOCK_RELEASE(&mads_engine->mad_queue_lock);

	cl_spinlock_destroy(&mads_engine->mad_queue_lock);

	cl_event_destroy(&mads_engine->poller_wakeup);
	cl_event_destroy(&mads_engine->outstanding_mads_done_event);
	cl_event_destroy(&mads_engine->sig_mads_on_wire_continue);

	free(mads_engine->class_name);
}

int osm_mads_engine_wait_pending_transactions(osm_mads_engine_t *p_mads_engine)
{
	while (1) {
		unsigned count = p_mads_engine->outstanding_mads;
		if (!count || osm_exit_flag)
			break;
		cl_event_wait_on(&p_mads_engine->outstanding_mads_done_event,
				 EVENT_NO_TIMEOUT, TRUE);
	}

	return osm_exit_flag;
}

void osm_mads_engine_decrement_outstanding_mads(osm_mads_engine_t *p_mads_engine)
{
	uint32_t outstanding;

	outstanding = cl_atomic_dec(&p_mads_engine->outstanding_mads);
	if (!outstanding)
		cl_event_signal(&p_mads_engine->outstanding_mads_done_event);

	cl_atomic_dec(&p_mads_engine->outstanding_mads_on_wire);
	cl_event_signal(&p_mads_engine->sig_mads_on_wire_continue);
}

void osm_mads_engine_post_mad(osm_mads_engine_t *p_mads_engine, osm_madw_t *p_madw)
{
	CL_SPINLOCK_ACQUIRE(&p_mads_engine->mad_queue_lock);
	cl_atomic_inc(&p_mads_engine->outstanding_mads);
	cl_qlist_insert_tail(&p_mads_engine->mad_queue, &p_madw->list_item);
	CL_SPINLOCK_RELEASE(&p_mads_engine->mad_queue_lock);

	cl_event_signal(&p_mads_engine->poller_wakeup);
}

void osm_mads_engine_post_trap_repress(osm_mads_engine_t *p_mads_engine, osm_madw_t *resp_madw)
{
	ib_net16_t slid = resp_madw->mad_addr.dest_lid;

	/* post trap repress to n2n poller */
	CL_SPINLOCK_ACQUIRE(&p_mads_engine->mad_queue_lock);
	cl_qlist_insert_tail(&p_mads_engine->mad_queue, &resp_madw->list_item);
	CL_SPINLOCK_RELEASE(&p_mads_engine->mad_queue_lock);

	OSM_LOG(&p_mads_engine->p_osm->log, OSM_LOG_VERBOSE,
		"Responding with trap repress to trap from LID %u\n", cl_ntoh16(slid));

	cl_event_signal(&p_mads_engine->poller_wakeup);
}


ib_api_status_t osm_mads_engine_main_bind(osm_mads_engine_t *p_mads_engine,
					  IN osm_bind_info_t * const p_bind_info,
					  IN osm_vend_mad_recv_callback_t mad_recv_callback,
					  IN osm_vend_mad_send_err_callback_t mad_send_err_callback,
					  IN void *context)
{
	ib_api_status_t status = IB_SUCCESS;
	ib_net64_t port_guid = p_bind_info->port_guid;
	uint8_t port_index = p_mads_engine->p_osm->subn.sm_port_idx;
	p_mads_engine->bind_port_index = port_index;

	OSM_LOG(&p_mads_engine->p_osm->log, OSM_LOG_VERBOSE,
		"Binding to port GUID 0x%" PRIx64 ", port_index: %u\n",
		cl_ntoh64(port_guid), port_index);

	if (p_mads_engine->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE) {
		p_mads_engine->bind_handle[port_index] =
			osm_vendor_bind(p_mads_engine->vendor[port_index], p_bind_info,
					p_mads_engine->mad_pool,
					mad_recv_callback,
					mad_send_err_callback, context);
	} else {
		p_mads_engine->bind_handle[port_index] =
			osm_vendor_rebind(p_mads_engine->vendor[port_index], p_bind_info,
					  p_mads_engine->mad_pool,
					  mad_recv_callback,
					  mad_send_err_callback,
					  context, p_mads_engine->bind_handle[port_index]);
	}

	if (p_mads_engine->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE) {
		status = IB_ERROR;
		OSM_LOG(&p_mads_engine->p_osm->log, OSM_LOG_ERROR,
			"ERR B202: %s bind failed (%s)\n",
			p_mads_engine->class_name, ib_get_err_str(status));
	}

	return status;
}

void osm_mads_engine_shutdown(osm_mads_engine_t *p_mads_engine)
{
	uint8_t port_index;

	for (port_index = 0; port_index < OSM_MAX_BINDING_PORTS; port_index++) {
		if (p_mads_engine->bind_handle[port_index] != CL_DISP_INVALID_HANDLE) {
			osm_vendor_unbind(p_mads_engine->bind_handle[port_index]);
		}
	}
}
