/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for NVLink discovery
 *	Set planes and rails
 *
 * Author:
 *	Or Nechemia, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NVLINK_C
#include <opensm/osm_lid_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_gpu.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_db_pack.h>

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);
extern ib_api_status_t osm_ar_set_extended_node_info(osm_sm_t * sm, osm_node_t * p_node);
extern void osm_nvlink_mad_rcv_process(IN void *context, IN void *data);

static void nvlink_get_alid_info(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_gpu_t *p_gpu = (osm_gpu_t *)p_map_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_physp_t *p_physp;
	osm_nvlink_mad_context_t mad_context;
	ib_api_status_t status;

	p_physp = osm_node_get_valid_physp(p_gpu->p_node);
	if (!p_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A820: "
			"Could not find valid port for GPU 0x%"PRIx64"\n",
			cl_ntoh64(p_gpu->p_node->node_info.node_guid));
		return;
	}

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_gpu->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_GET;

	/* if we send GET for ALIDInfo, then our data is not updated */
	p_gpu->alid_info_updated = FALSE;

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
			     IB_MAD_ATTR_ALID_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A821: "
			"Getting ALIDInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid),	ib_get_err_str(status));

}

ib_api_status_t osm_nvlink_send_set_alid_info(osm_sm_t * sm, osm_gpu_t * p_gpu)
{
	osm_physp_t *p_physp;
	osm_nvlink_mad_context_t mad_context;
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_alid_info_t *p_alid_info_data = (ib_alid_info_t *)payload;
	ib_api_status_t status;
	uint8_t num_alids = p_gpu->p_node->new_ext_node_info.anycast_lid_table_top;

	/* No need for an update */
	if (p_gpu->alid_info_updated &&
	    !memcmp(p_gpu->calc_alid_info.alid_block, p_gpu->subn_alid_info.alid_block,
		    sizeof(ib_alid_entry_t) * num_alids))
		return IB_SUCCESS;

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_gpu->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_SET;

	/* Prepare ALIDInfo SET request */
	memcpy(p_alid_info_data, &p_gpu->calc_alid_info, sizeof(ib_alid_info_t));

	p_physp = osm_node_get_valid_physp(p_gpu->p_node);
	if (!p_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A822: "
			"Could not find valid port for GPU 0x%"PRIx64"\n",
			cl_ntoh64(p_gpu->p_node->node_info.node_guid));
		return IB_ERROR;
	}

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			     p_physp, payload, sizeof(ib_alid_info_t),
			     IB_MAD_ATTR_ALID_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A823: "
			"Setting ALIDInfo on 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_gpu_get_guid(p_gpu)), ib_get_err_str(status));
	return status;
}

static void nvlink_validate_sw_plane_db(IN osm_nvlink_mgr_t * p_mgr)
{
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint8_t plane;
	boolean_t valid_entry = TRUE;

	OSM_LOG_ENTER(p_log);

	cl_qlist_init(&guids);

	if (osm_db_guids(p_mgr->p_guid2planes, &guids)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A80F: "
			"could not get guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2plane_get(p_mgr->p_guid2planes, p_item->guid, &plane)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A810: "
				"could not get plane for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		} else {
			valid_entry = TRUE;

			/* Plane should be non zero */
			if (plane < OSM_NVLINK_MIN_PLANES || plane > OSM_NVLINK_MAX_PLANES) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A811: "
					"Plane %u of guid:0x%016" PRIx64 " is illegal. "
					"Valid range is [%u,%u].\n",
					plane, p_item->guid, OSM_NVLINK_MIN_PLANES,
					OSM_NVLINK_MAX_PLANES);

				valid_entry = FALSE;
			}

			if (!valid_entry &&
			    osm_db_guid2lid_delete(p_mgr->p_guid2planes, p_item->guid))
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A812: "
					"failed to delete plane entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(p_log);
}

static void nvlink_init_sw_plane_db(osm_nvlink_mgr_t * p_mgr)
{
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;

	OSM_LOG_ENTER(&p_osm->log);

	/* Initialize the given database */
	p_mgr->p_guid2planes = osm_db_domain_init(&p_osm->db, "guid2planes");
	if (!p_mgr->p_guid2planes) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A808: "
			"Error initializing switch GUID to plane persistent database\n");
		return;
	}

	if (osm_db_restore(p_mgr->p_guid2planes)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted guid2mkey file.
		 */
		if (p_osm->subn.opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring switch GUID to plane "
				"persistent database\n");
			return;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A809: Error restoring switch GUID to plane "
				"persistent database\n");
	}

	nvlink_validate_sw_plane_db(p_mgr);

	OSM_LOG_EXIT(&p_osm->log);
}

static void nvlink_validate_port_rail_db(IN osm_nvlink_mgr_t * p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_qlist_t guids;
	osm_db_guid_portnum_elem_t *p_item;
	uint8_t rail;
	boolean_t valid_entry = TRUE;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	cl_qlist_init(&guids);

	if (osm_db_guid_portnum_keys_list(p_mgr->p_port2rail, &guids)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A817: "
			"could not get guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_portnum_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_portnum_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_port2rail_get(p_mgr->p_port2rail, p_item->guid, p_item->portnum, &rail)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A818: "
				"could not get rail for guid:0x%016" PRIx64 " port:%u\n",
				p_item->guid, p_item->portnum);
		} else {
			valid_entry = TRUE;

			if (rail > OSM_NVLINK_MAX_RAIL) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A819: "
					"Rail %u of guid:0x%016" PRIx64 " port %u is illegal. "
					"Valid range is [%u,%u].\n",
					rail, p_item->guid, p_item->portnum,
					OSM_NVLINK_MIN_RAIL,
					OSM_NVLINK_MAX_RAIL);

				valid_entry = FALSE;
			}

			if (!valid_entry &&
			    osm_db_guid_portnum_delete(p_mgr->p_port2rail,
						       p_item->guid,
						       p_item->portnum)) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A81A: "
					"failed to delete rail entry for "
					"guid:0x%016" PRIx64 " port %u\n",
					p_item->guid, p_item->portnum);
			}
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static inline boolean_t nvlink_is_reassign_rails(osm_subn_t * p_subn)
{
	return p_subn->first_time_master_sweep && p_subn->opt.nvl_reassign_rails;
}

static void nvlink_init_port_rail_db(IN osm_nvlink_mgr_t * p_mgr)
{
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;

	OSM_LOG_ENTER(&p_osm->log);

	/* Initialize the given database */
	p_mgr->p_port2rail = osm_db_domain_init(&p_osm->db, "port2rail");
	if (!p_mgr->p_port2rail) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A81B: "
			"Error initializing GPU port GUID to rail persistent database\n");
		return;
	}

	if (nvlink_is_reassign_rails(p_mgr->p_subn))
		goto Exit;

	if (osm_db_restore(p_mgr->p_port2rail)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted guid2rail file.
		 */
		if (p_osm->subn.opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring GPU port GUID to rail "
				"persistent database\n");
			return;
		}
#endif
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A81C: "
			"Error restoring GPU port GUID to rail "
			"persistent database\n");
	}

	nvlink_validate_port_rail_db(p_mgr);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static ib_api_status_t nvlink_iterate_within_plane_validate_port(osm_subn_t * p_subn,
								 osm_switch_t * p_cur_sw,
								 uint8_t port_num,
								 osm_switch_t ** queue_plane,
								 uint32_t * queue_plane_in,
								 uint32_t * queue_plane_out)
{
	osm_physp_t *p_physp, *p_remote_physp;
	osm_node_t *p_remote_node;
	ib_api_status_t status = IB_ERROR;
	uint8_t remote_plane, cur_plane;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_physp = osm_node_get_physp_ptr(p_cur_sw->p_node, port_num);
	if (!p_physp)
		goto SkipPort;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		goto SkipPort;

	cur_plane = p_cur_sw->nvlink_sw_data.plane;
	p_remote_node = p_remote_physp->p_node;

	/*
	 * Mark for each GPU, which remote leaf switch is it connected to,
	 * according to plane (GPU is connected to no more than one leaf per plane)
	 * This array is used when creating ARgroup per ALID.
	 */
	if (osm_node_is_gpu(p_remote_node)) {
		p_remote_node->p_gpu->remote_leaf_sw_physp_arr[cur_plane] = p_physp;

		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Found GPU 0x%"PRIx64" is connected to leaf switch "
			"0x%"PRIx64" port %u as part of plane %u\n",
			osm_node_get_node_guid_ho(p_remote_node),
			osm_node_get_node_guid_ho(p_cur_sw->p_node),
			port_num, cur_plane);
	}

	/*
	 * Check port:
	 * 1. For service port, remote is of another plane, no need to set it to this plane.
	 * 2. If remote port is a switch port, add remote switch to this plane queue,
	 *    as it belongs to current plane and should keep iterating through.
	 */

	/*
	 * Planes are set to switches.
	 * GPU port plane is as the plane of its remote switch.
	 */
	if (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
		goto SkipPort;

	if ((osm_physp_is_fnm(p_subn, p_physp) &&
	     !osm_physp_is_fnm(p_subn, p_physp->p_remote_physp)) ||
	    (!osm_physp_is_fnm(p_subn, p_physp) &&
	     osm_physp_is_fnm(p_subn, p_physp->p_remote_physp))) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Found switch 0x%"PRIx64" port %u is %s port while remote "
			"is %s port. Fail NVLink planes configuration.\n",
			osm_node_get_node_guid_ho(p_cur_sw->p_node), port_num,
			osm_physp_is_fnm(p_subn, p_physp) ? "FNM" : "Mission",
			osm_physp_is_fnm(p_subn, p_physp->p_remote_physp) ?
			    "FNM" : "Mission");
		status = IB_INVALID_PORT;
		goto Exit;
	}

	remote_plane = p_remote_node->sw->nvlink_sw_data.plane;
	if (osm_physp_is_fnm(p_subn, p_physp)) {
		/*
		 * Planes are indexed from 1 to N and in preprocess step
		 * all to 0. Hence plane indicates if remote was visited.
		 */
		if (remote_plane) {
			/* Remote was visited and has different plane, as it should */
			if (cur_plane != remote_plane)
				goto SkipPort;

			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found switch 0x%"PRIx64" and remote switch "
				"0x%"PRIx64" are part of the same plane %u, "
				"although connected through FNM port. "
				"Reassign all planes.\n",
				osm_node_get_node_guid_ho(p_cur_sw->p_node),
				osm_node_get_node_guid_ho(p_remote_node),
				cur_plane);

			goto Exit;
		}
	} else {
		/*
		 * Planes are indexed from 1 to N and in preprocess step
		 * all to 0. Hence plane indicates if remote was visited.
		 */
		if (remote_plane) {
			/* Remote was visited and has the same plane, as it should */
			if (cur_plane == remote_plane)
				goto SkipPort;

			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found switch 0x%"PRIx64" plane %u is"
				" different than remote switch 0x%"PRIx64
				" plane %u. Reassign all planes.\n",
				osm_node_get_node_guid_ho(p_cur_sw->p_node),
				cur_plane,
				osm_node_get_node_guid_ho(p_remote_node),
				remote_plane);

			goto Exit;
		}

		/*
		 * For ports that are not service ports, if remote is a switch,
		 * add it to this plane queue, as it is a part of current plane.
		 */
		if (!p_remote_node->sw->nvlink_sw_data.plane) {
			queue_plane[(*queue_plane_in)++] = p_remote_node->sw;

			/* To avoid inserting the same switch multiple times */
			p_remote_node->sw->nvlink_sw_data.plane = cur_plane;
		}
	}

SkipPort:
	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static ib_api_status_t nvlink_iterate_within_plane(osm_nvlink_mgr_t * p_mgr,
						   osm_switch_t * p_start_sw,
						   uint8_t plane)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t **queue_plane = NULL;
	osm_switch_t *p_cur_sw;
	uint32_t queue_plane_out = 0, queue_plane_in = 0;
	uint8_t port_num, num_ports, db_plane;
	uint64_t cur_guid;
	ib_api_status_t status = IB_ERROR;

	OSM_LOG_ENTER(p_log);

	/* use for the internal iteration until discovers all port within the plane */
	queue_plane = (osm_switch_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					      sizeof(osm_switch_t *));
	if (!queue_plane) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A801: "
			"Cannot allocate memory for sw Queue\n");
		goto Exit;
	}

	/* Start iteration from input switch */
	queue_plane[queue_plane_in++] = p_start_sw;

	/*
	 * Starting BFS from input p_sw, iterate over the subnet in order to find
	 * all switches belong to the same plane as the input switch.
	 *
	 * Service port (FNM port) connects between planes (used for IB communication
	 * and avoided by NVLink communication)
	 * Hence, when gets to one, should not move forward with that switch.
	 */
	while (queue_plane_out < queue_plane_in) {
		p_cur_sw = queue_plane[queue_plane_out++];
		cur_guid = osm_node_get_node_guid_ho(p_cur_sw->p_node);

		/* Validate DB plane is identical to current iterated plane */
		if (!osm_db_guid2plane_get(p_mgr->p_guid2planes, cur_guid, &db_plane) &&
		    (plane != db_plane)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A80E: "
				"DB plane for switch 0x%" PRIx64 " is %u, "
				"while discovered through switch of plane %u. "
				"Invalidate NVLink planes cache and reassign planes.\n",
				cur_guid, db_plane, plane);

			status = IB_ERROR;
			goto Exit;
		} else if (!db_plane) {
			osm_db_guid2plane_set(p_mgr->p_guid2planes, cur_guid, plane);
		}

		/* Set current plane to current switch */
		p_cur_sw->nvlink_sw_data.plane = plane;

		/* Continue discovering through switch ports */
		num_ports = osm_node_get_num_physp(p_cur_sw->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			status = nvlink_iterate_within_plane_validate_port(
				    p_subn, p_cur_sw, port_num,
				    queue_plane, &queue_plane_in, &queue_plane_out);
			if (status)
				goto Exit;
		}
	}

	status = IB_SUCCESS;
Exit:
	free(queue_plane);

	OSM_LOG_EXIT(p_log);
	return status;
}

static void nvlink_init_sw_plane(cl_map_item_t * item, void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)item;
	p_sw->nvlink_sw_data.plane = OSM_NVLINK_NO_PLANE;
}

static void nvlink_init_gpu_remote_lids_arr(cl_map_item_t * item, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_gpu_t *p_gpu = (osm_gpu_t *)item;

	memset(p_gpu->remote_leaf_sw_physp_arr, 0, sizeof(p_gpu->remote_leaf_sw_physp_arr));
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Initialize GPU 0x%"PRIx64" plane to switch leaf mapping\n",
		osm_node_get_node_guid_ho(p_gpu->p_node));
}

static void nvlink_plane_discover_preprocess(osm_subn_t * p_subn)
{
	/* Zero all Switch previous planes */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_init_sw_plane, NULL);

	/* Zero remote switch lid array of each GPU (indexed by planes) */
	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, nvlink_init_gpu_remote_lids_arr, p_subn);

}

static void nvlink_invalidate_plane_cache(osm_nvlink_mgr_t * p_mgr,
					  osm_switch_t ** queue,
					  uint32_t *queue_in)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_switch_t *p_sw;
	cl_map_item_t *item;

	osm_db_clear(p_mgr->p_guid2planes);

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;

		p_sw->nvlink_sw_data.plane = OSM_NVLINK_NO_PLANE;
	
		queue[(*queue_in)++] = p_sw;
	}
}

static inline uint8_t nvlink_update_next_free_plane(boolean_t * is_used,
						    uint8_t next_free_plane)
{
	while (next_free_plane <= OSM_NVLINK_MAX_PLANES && is_used[next_free_plane])
		next_free_plane++;

	if (next_free_plane <= OSM_NVLINK_MAX_PLANES)
		is_used[next_free_plane] = TRUE;

	return next_free_plane;
}

static ib_api_status_t nvlink_plane_discover(osm_nvlink_mgr_t * p_mgr, boolean_t is_cache_valid)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t **queue;
	osm_switch_t *p_cur_sw;
	uint32_t queue_in = 0, queue_out = 0;
	uint8_t cur_plane = 1;
	uint64_t cur_guid;
	boolean_t is_used[OSM_NVLINK_MAX_PLANES + 1];
	uint8_t next_free_plane = 1;
	cl_map_item_t *item;
	ib_api_status_t status = IB_ERROR;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	queue = (osm_switch_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					sizeof(osm_switch_t *));
	if (!queue) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A802: "
			"Cannot allocate memory for sw Queue\n");
		goto Exit;
	}

	nvlink_plane_discover_preprocess(p_subn);
	memset(is_used, 0, sizeof(is_used));

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_cur_sw = (osm_switch_t *) item;

		/*
		 * If switch does not have an entry on switch plane cache, skip it.
		 *	It will get a plane in one of the following ways:
		 *	1. While iterating from another switch that exists on DB.
		 *	2. While iterating on the queue holding the switches that
		 *	   could not be discovered from the DB switches.
		 */
		cur_guid = osm_node_get_node_guid_ho(p_cur_sw->p_node);
		if (osm_db_guid2plane_get(p_mgr->p_guid2planes, cur_guid, &cur_plane)) {
			queue[queue_in++] = p_cur_sw;
			continue;
		}

		is_used[cur_plane] = TRUE;

		/*
		 * Start BFS from current switch.
		 * Any switch visited while iterating should have the same plane
		 * as current switch.
		 * If switch with cached plane different than current plane is found
		 * on this iteration, invalidate the cache and reassign all planes.
		 */
		is_cache_valid = !nvlink_iterate_within_plane(p_mgr, p_cur_sw, cur_plane);
	}

	if (!is_cache_valid) {
		/* This function also inserts all of the switches to the queue */
		nvlink_invalidate_plane_cache(p_mgr, queue, &queue_in);
		memset(is_used, 0, sizeof(is_used));
		next_free_plane = 1;
	}

	/* 
	 * At this point, queue holds all the switches that are missing from plane DB.
	 * Some may have gotten a plane while iterating by the BFS from switches that
	 * exist on DB.
	 * On this loop we set planes to those switches that are yet to be reached.
	 */
	while (queue_out < queue_in) {
		p_cur_sw = queue[queue_out++];

		/* Skip switch that was already set to a plane */
		if (p_cur_sw->nvlink_sw_data.plane)
			continue;

		next_free_plane = nvlink_update_next_free_plane(is_used, next_free_plane);
		if (next_free_plane > OSM_NVLINK_MAX_PLANES) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A803: "
				"Found more than maximal allowed planes (%u). "
				"Fail NVLink plane discovery.\n",
				OSM_NVLINK_MAX_PLANES);
			goto Exit;
		}

		/* Set current plane to current switch */
		p_cur_sw->nvlink_sw_data.plane = next_free_plane;
		osm_db_guid2plane_set(p_mgr->p_guid2planes,
				      osm_node_get_node_guid_ho(p_cur_sw->p_node),
				      next_free_plane);

		/* Start iterating on a new plane, from next unvisited switch */
		nvlink_iterate_within_plane(p_mgr, p_cur_sw, next_free_plane);
	}

	status = IB_SUCCESS;
Exit:
	free(queue);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static void nvlink_rail_filter_config_mark_port_mask(ib_rail_filter_config_t * p_rail_filter_config,
						     osm_nvlink_port_type_t port_type,
						     uint8_t port_num,
						     uint8_t mask_block_num)
{
	uint8_t *p_port_mask = NULL;
	uint8_t mask_index, mask_offset_bit;
	uint16_t mask_len, port_index, mask_len_in_bytes;

	CL_ASSERT(port_type < IB_PORT_TYPE_MAX);

	switch (port_type) {
	case IB_PORT_TYPE_INGRESS:
		mask_len = IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
		mask_len_in_bytes = IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN;
		p_port_mask = p_rail_filter_config->ingress_port_mask;
		break;
	case IB_PORT_TYPE_EGRESS:
		mask_len = IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;
		mask_len_in_bytes = IB_RAIL_FILTER_CFG_EGRESS_MASK_BYTES_LEN;
		p_port_mask = p_rail_filter_config->egress_port_mask;
		break;
	default:
		return;
	}

	/* Assert that input mask block num matches the relevant block for input port number */
	OSM_ASSERT(mask_block_num == port_num / mask_len);

	port_index = port_num % mask_len;

	mask_index = port_index / (uint8_t)IB_RAIL_FILTER_CFG_PORT_MASK_BLOCK_SIZE;
	mask_index = mask_len_in_bytes - 1 - mask_index;

	mask_offset_bit = port_index % IB_RAIL_FILTER_CFG_PORT_MASK_BLOCK_SIZE;

	p_port_mask[mask_index] |= ((uint8_t)1 << mask_offset_bit);
}

static void nvlink_rail_filter_config_send_block(osm_subn_t * p_subn,
						 osm_switch_t * p_sw,
						 uint8_t rail,
						 uint8_t ingress_block,
						 uint8_t egress_block)
{
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_rail_filter_cfg_context_t context;
	osm_physp_t *p_physp0;
	ib_rail_filter_config_t *p_calc_rail_filter_cfg =
	    osm_nvlink_get_rail_filter_mad(p_sw, rail, ingress_block, egress_block, OSM_RAIL_FILTER_CALC);
	ib_rail_filter_config_t *p_subn_rail_filter_cfg =
	    osm_nvlink_get_rail_filter_mad(p_sw, rail, ingress_block, egress_block, OSM_RAIL_FILTER_SUBN);
	uint32_t attr_mod;
	ib_api_status_t status;

	if (!p_subn->need_update && !p_sw->need_update &&
	    !memcmp(p_calc_rail_filter_cfg, p_subn_rail_filter_cfg,
		    sizeof(ib_rail_filter_config_t)))
		return;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A805: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	if (!memcmp(p_calc_rail_filter_cfg->ingress_port_mask, ZEROARRAY,
		    IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"RailFilterConfig with ingress port mask 0 is not allowed. "
			"Skip setting RailFilterConfig of rail %u for switch 0x%"PRIx64"\n",
			rail, cl_ntoh64(p_sw->p_node->node_info.node_guid));
		
		return;
	}

	/* Using ingress port mask, no need for port number on attr_mod */
	attr_mod = (ingress_block << IB_RAIL_FILTER_ATTR_MOD_INGRESS_BLOCK_SHIFT) |
	    (egress_block << IB_RAIL_FILTER_ATTR_MOD_EGRESS_BLOCK_SHIFT);

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.ingress_block = ingress_block;
	context.egress_block = egress_block;
	context.rail = rail;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)p_calc_rail_filter_cfg,
			     sizeof(ib_rail_filter_config_t), IB_MAD_ATTR_RAIL_FILTER_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A806: "
			"Setting RailFilterConfig of rail %u (ingress block %u, egress block %u) "
			"to 0x%" PRIx64 " failed (%s)\n",
			rail, ingress_block, egress_block,
			cl_ntoh64(p_physp0->port_guid), ib_get_err_str(status));
}

static void nvlink_rail_filter_preprocess(cl_map_item_t * item, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_switch_t *p_sw = (osm_switch_t *)item;
	ib_rail_filter_config_t *p_calc_rail_filter_cfg;
	uint32_t i, num_rail_filter_mads, num_ports, port_num;
	osm_physp_t *p_physp;

	num_rail_filter_mads = OSM_NVLINK_SN_NUM_RAILS *
			       p_sw->nvlink_sw_data.max_ingress_block *
			       p_sw->nvlink_sw_data.max_egress_block;

	for (i = 0; i < num_rail_filter_mads; i++) {
		p_calc_rail_filter_cfg = &p_sw->nvlink_sw_data.calc_rail_filter_cfg[i];

		memset(p_calc_rail_filter_cfg, 0, sizeof(ib_rail_filter_config_t));

		p_calc_rail_filter_cfg->vl_mask |= IB_RAIL_FILTER_CFG_NVLINK_VL_MASK;
		p_calc_rail_filter_cfg->mc_uc_enable |= (IB_RAIL_FILTER_CFG_MC_ENABLE_BIT |
						     IB_RAIL_FILTER_CFG_UC_ENABLE_BIT);
	}

	p_sw->nvlink_sw_data.rail_used_mask = 0;

	/* Initialize all port rails in order to reassign */
	if (nvlink_is_reassign_rails(p_subn)) {
		num_ports = osm_node_get_num_physp(p_sw->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp)
				continue;
			p_physp->rail = OSM_NVLINK_NO_RAIL;

			/*
			 * Rails are cached by switch ports.
			 * Initialize the remote as well, so values are synced
			*/
			if (p_physp->p_remote_physp)
				p_physp->p_remote_physp->rail = OSM_NVLINK_NO_RAIL;
		}
	}
}

static boolean_t nvlink_switch_is_rail_used(osm_switch_t * p_sw, uint8_t rail)
{
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	return (p_sw->nvlink_sw_data.rail_used_mask >> rail) & (uint16_t)0x1;
}

static void nvlink_switch_mark_rail_used(osm_switch_t * p_sw, uint8_t rail)
{
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	p_sw->nvlink_sw_data.rail_used_mask |= ((uint16_t)0x1) << rail;
}

static void nvlink_set_rail_filter_config(cl_map_item_t * item, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_switch_t *p_sw = (osm_switch_t *)item;
	uint8_t max_ingress_block, max_egress_block;
	uint8_t ingress_block, egress_block, rail;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_RAIL_FIILTER_SUPPORTED_BIT))
		return;

	max_ingress_block = p_sw->nvlink_sw_data.max_ingress_block;
	max_egress_block = p_sw->nvlink_sw_data.max_egress_block;

	for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
		/* No need to send RailFilterConfig for unused rail */
		if (!nvlink_switch_is_rail_used(p_sw, rail))
			continue;

		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Sending RailFilterConfig for switch 0x%"PRIx64" rail %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)), rail);

		for (ingress_block = 0; ingress_block < max_ingress_block; ingress_block++) {
			for (egress_block = 0; egress_block < max_egress_block; egress_block++) {
				nvlink_rail_filter_config_send_block(p_subn, p_sw, rail,
								     ingress_block, egress_block);
			}
		}
	}
}

static void nvlink_rail_filter_config_rail_add_port_num(osm_switch_t * p_sw,
							uint8_t rail,
							uint8_t port_num)
{
	ib_rail_filter_config_t *p_remote_rail_filter_cfg;
	uint8_t max_block, port_in_block, port_e_block;
	uint8_t ingress_block, egress_block;

	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	port_in_block = port_num / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
	port_e_block = port_num / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;

	/* Set rail egress port masks across all MADs with relevant egress mask block */
	max_block = p_sw->nvlink_sw_data.max_ingress_block;
	for (ingress_block = 0; ingress_block < max_block; ingress_block++) {
		p_remote_rail_filter_cfg = osm_nvlink_get_rail_filter_mad(
		    p_sw, rail, ingress_block, port_e_block, OSM_RAIL_FILTER_CALC);

		nvlink_rail_filter_config_mark_port_mask(
		    p_remote_rail_filter_cfg, IB_PORT_TYPE_EGRESS, port_num, port_e_block);
	}

	/* Set rail ingress port masks across all MADs with relevant ingress mask block */
	max_block = p_sw->nvlink_sw_data.max_egress_block;
	for (egress_block = 0; egress_block < max_block; egress_block++) {
		p_remote_rail_filter_cfg = osm_nvlink_get_rail_filter_mad(
		    p_sw, rail, port_in_block, egress_block, OSM_RAIL_FILTER_CALC);

		nvlink_rail_filter_config_mark_port_mask(
		    p_remote_rail_filter_cfg, IB_PORT_TYPE_INGRESS, port_num, port_in_block);
	}
}

static ib_api_status_t nvlink_get_remote_details(IN osm_physp_t * p_physp,
						 OUT uint64_t *p_guid,
						 OUT uint8_t *p_port_num)
{
	osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);

	if (!p_remote_physp)
		return IB_ERROR;

	*p_port_num = osm_physp_get_port_num(p_remote_physp);
	*p_guid = cl_ntoh64(p_remote_physp->port_guid);

	return IB_SUCCESS;
}

static uint8_t nvlink_get_new_rail(osm_gpu_t * p_gpu, osm_physp_t * p_physp)
{
	osm_switch_t *p_plane_sw;
	osm_physp_t *p_sw_physp;
	uint8_t plane_rail_count[OSM_NVLINK_SN_NUM_RAILS];
	uint8_t rail = OSM_NVLINK_NO_RAIL;
	uint8_t gpu_num_ports, port_num;

	OSM_ASSERT(p_gpu && p_physp);

	p_sw_physp = p_physp->p_remote_physp;
	if (!p_physp->p_remote_physp)
		return OSM_NVLINK_NO_RAIL;

	p_plane_sw = p_sw_physp->p_node->sw;
	if (!p_plane_sw)
		return OSM_NVLINK_NO_RAIL;

	/* Iterate over GPU ports for choosing a new rail */
	memset(plane_rail_count, 0, sizeof(plane_rail_count));

	gpu_num_ports = osm_node_get_num_physp(p_gpu->p_node);
	for (port_num = 1; port_num < gpu_num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
		if (!p_physp)
			continue;

		p_sw_physp = osm_physp_get_remote(p_physp);
		if (!p_sw_physp)
			continue;

		/* Count the rails that this GPU has for this plane */
		if (p_sw_physp->p_node != p_plane_sw->p_node)
			continue;

		if (p_sw_physp->rail >= OSM_NVLINK_SN_NUM_RAILS)
			continue;

		plane_rail_count[p_sw_physp->rail]++;
	}

	/*
	 *  Each GPU is connected to each plane with number of links
	 *  such that each rail on the plane gets a single link.
	 */
	rail = 0;
	while (plane_rail_count[rail] && rail < OSM_NVLINK_SN_NUM_RAILS)
		rail++;

	return rail < OSM_NVLINK_SN_NUM_RAILS ? rail : OSM_NVLINK_NO_RAIL;
}

/* NOTE: port2rail cache refers to sw port, therefore rail is marked on switch port side */
static uint8_t nvlink_get_gpu_port_rail(osm_nvlink_mgr_t * p_mgr, osm_gpu_t * p_gpu, osm_physp_t * p_physp)
{
	osm_physp_t *p_sw_physp;
	uint8_t remote_port_num = 0;
	uint64_t remote_guid = 0;
	uint8_t rail = OSM_NVLINK_NO_RAIL, cached_rail;

	/* Get plane from cache according to remote switch port */
	if (nvlink_get_remote_details(p_physp, &remote_guid, &remote_port_num))
		goto Exit;

	p_sw_physp = osm_physp_get_remote(p_physp);
	if (!p_sw_physp)
		goto Exit;

	if (nvlink_is_reassign_rails(p_mgr->p_subn))
		goto NewRail;

	if (p_physp->rail != OSM_NVLINK_NO_RAIL) {
		rail = p_physp->rail;
		goto Exit;
	}

	if (!osm_db_port2rail_get(p_mgr->p_port2rail, remote_guid, remote_port_num, &cached_rail)) {
		rail = cached_rail;
		goto Exit;
	}

	/* Avoid changing rails for active ports, unless nvl_reassign_rails is configured */
	if ((ib_port_info_get_port_state(&p_sw_physp->port_info) == IB_LINK_ACTIVE)) {
		rail = p_sw_physp->rail;
		goto Exit;
	}

NewRail:
	rail = nvlink_get_new_rail(p_gpu, p_physp);
	osm_db_port2rail_set(p_mgr->p_port2rail, remote_guid, remote_port_num, rail);

Exit:
	return rail;
}

static void nvlink_rail_configuration(osm_nvlink_mgr_t * p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_map_item_t *p_item;
	osm_gpu_t *p_gpu;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, remote_port_num, rail;
	osm_node_t *p_remote_node;
	osm_switch_t *p_remote_sw;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (nvlink_is_reassign_rails(p_subn))
		osm_db_clear(p_mgr->p_port2rail);

	/* Zero previously calculated RailFilterConfig MADs */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_rail_filter_preprocess, p_subn);

	/* Single Node with 2 switches (low priority topology) */
	for (p_item = cl_qmap_head(&p_subn->gpu_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->gpu_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_gpu = (osm_gpu_t *)p_item;

		num_ports = osm_node_get_num_physp(p_gpu->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
			if (!p_physp)
				continue;

			/* Initialize in order to sync with remote port rail */
			p_physp->rail = OSM_NVLINK_NO_RAIL;

			if (!p_physp->p_remote_physp)
				continue;

			p_remote_node = p_physp->p_remote_physp->p_node;
			if (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
				continue;

			/* Get rail and keep it according to switch port */
			rail = nvlink_get_gpu_port_rail(p_mgr, p_gpu, p_physp);
			p_physp->p_remote_physp->rail = rail;
			p_physp->rail = rail;
			if (rail == OSM_NVLINK_NO_RAIL)
				continue;

			/* On nvlink topologies, remote port of GPU is a switch port */
			p_remote_sw = p_remote_node->sw;

			remote_port_num = p_physp->p_remote_physp->port_num;
			nvlink_rail_filter_config_rail_add_port_num(p_remote_sw, rail,
								    remote_port_num);
			nvlink_switch_mark_rail_used(p_remote_sw, rail);
		}
	}

	/* Iterate over switches and send switch's req RailFilterConfig for each rail */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_set_rail_filter_config, p_subn);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_nvlink_sw_data_init(osm_switch_t * p_sw)
{
	uint8_t num_ports, max_ingress_block, max_egress_block;
	uint32_t num_mads;

	num_ports = p_sw->num_ports;

	/* Initialize nvlink MAD configuration structures */
	max_ingress_block = (num_ports / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN) + 1;
	max_egress_block = (num_ports / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN) + 1;
	p_sw->nvlink_sw_data.max_ingress_block = max_ingress_block;
	p_sw->nvlink_sw_data.max_egress_block = max_egress_block;
	p_sw->nvlink_sw_data.rail_used_mask = 0;

	num_mads = OSM_NVLINK_SN_NUM_RAILS * max_ingress_block * max_egress_block;
	p_sw->nvlink_sw_data.subn_rail_filter_cfg =
	    (ib_rail_filter_config_t *)calloc(num_mads, sizeof(ib_rail_filter_config_t));
	p_sw->nvlink_sw_data.calc_rail_filter_cfg =
	    (ib_rail_filter_config_t *)calloc(num_mads, sizeof(ib_rail_filter_config_t));
}

void osm_nvlink_sw_data_destroy(osm_nvlink_sw_data_t * nvlink_sw_data)
{
	free(nvlink_sw_data->subn_rail_filter_cfg);
	free(nvlink_sw_data->calc_rail_filter_cfg);
}

ib_rail_filter_config_t * osm_nvlink_get_rail_filter_mad(osm_switch_t * p_sw,
							 uint8_t rail,
							 uint8_t ingress_block,
							 uint8_t egress_block,
							 boolean_t is_calc)
{
	uint16_t rail_offset, in_offset;

	OSM_ASSERT((rail < OSM_NVLINK_SN_NUM_RAILS) &&
		   (ingress_block < p_sw->nvlink_sw_data.max_ingress_block) &&
		   (egress_block < p_sw->nvlink_sw_data.max_egress_block));

	rail_offset = rail *
	    p_sw->nvlink_sw_data.max_ingress_block *
	    p_sw->nvlink_sw_data.max_egress_block;
	in_offset = ingress_block * p_sw->nvlink_sw_data.max_egress_block;

	if (is_calc)
		return &p_sw->nvlink_sw_data.calc_rail_filter_cfg[rail_offset +
								  in_offset +
								  egress_block];

	return &p_sw->nvlink_sw_data.subn_rail_filter_cfg[rail_offset + in_offset + egress_block];
}

static void nvlink_lft_split_calc_reset_ranges_mad(osm_subn_t * p_subn,
						    osm_switch_t * p_sw,
						    ib_linear_forwarding_table_split_t * p_lft_split)
{
	ib_linear_forwarding_table_split_t * p_subn_lft_split = &p_sw->nvlink_sw_data.subn_lft_split;
	uint32_t range_start;

	memcpy(p_lft_split, p_subn_lft_split, sizeof(ib_linear_forwarding_table_split_t));

	/* Resetting LFTSplit LIDs ranges is done by setting each range top to (start - 1) */

	range_start = cl_ntoh32(p_subn_lft_split->global_lid_range_start) & IB_LFT_SPLIT_LID_MASK;
	p_lft_split->global_lid_range_top = cl_hton32((range_start - 1) & IB_LFT_SPLIT_LID_MASK);

	range_start = cl_ntoh32(p_subn_lft_split->alid_range_start) & IB_LFT_SPLIT_LID_MASK;
	p_lft_split->alid_range_top = cl_hton32((range_start - 1) & IB_LFT_SPLIT_LID_MASK);

	range_start = cl_ntoh32(p_subn_lft_split->local_plane_lid_range_start) & IB_LFT_SPLIT_LID_MASK;
	p_lft_split->local_plane_lid_range_top = cl_hton32((range_start - 1) & IB_LFT_SPLIT_LID_MASK);
}

static void nvlink_update_fwd_top(osm_sm_t *p_sm, osm_switch_t *p_sw, uint16_t fwd_top)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	osm_madw_context_t context;
	ib_api_status_t status;
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_switch_info_t *p_si = (ib_switch_info_t *)payload;

	OSM_LOG_ENTER(p_sm->p_log);

	OSM_ASSERT(p_sm);
	OSM_ASSERT(p_sw);

	p_node = p_sw->p_node;
	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	OSM_ASSERT(p_physp);

	p_path = osm_physp_get_dr_path_ptr(p_sm, p_physp);

	/*
	 * Set the top of the unicast forwarding table to 0 before
	 * setting new LFTSplit ranges for NVLink configuration.
	 */
	memcpy(p_si, &p_sw->switch_info, sizeof(ib_switch_info_t));

	/*
	 * Set port state change to 0 (ignored), so avoids override the field
	 * in case FW changed the value since latest discovery.
	 */
	ib_switch_info_clear_state_change(p_si);


	p_si->lin_top = cl_hton16(fwd_top);
	context.si_context.lft_top_change = TRUE;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
		"Setting switch LFT top to LID 0x%x before setting new LFTSplit values\n",
		cl_ntoh16(p_si->lin_top));

	context.si_context.light_sweep = FALSE;
	context.si_context.node_guid = osm_node_get_node_guid(p_node);
	context.si_context.set_method = TRUE;
	context.si_context.lft_split_state_change = TRUE;

	status = osm_req_set(p_sm, p_path, p_physp, payload,
			     sizeof(ib_switch_info_t), IB_MAD_ATTR_SWITCH_INFO,
			     0, FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A814: "
			"Sending SwitchInfo attribute failed (%s)\n",
			ib_get_err_str(status));

	OSM_LOG_EXIT(p_sm->p_log);
}

static void nvlink_set_zero_fwd_top(IN cl_map_item_t * p_map_item,
				    IN void * context)
{
	osm_sm_t *p_sm = context;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	/*
	 * If SM need to re-split LFT tables of the switch, SM first need to
	 * zero LFT top of the switch.
	 */
	if (p_sw->nvlink_sw_data.lft_split_state == IB_LFT_SPLIT_STATE_NEED_UPDATE)
		nvlink_update_fwd_top(p_sm, p_sw, 0);
}

static void nvlink_set_fwd_top(IN cl_map_item_t * p_map_item,
			       IN void * context)
{
	osm_sm_t *p_sm = context;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	/*
	 * Send SwitchInfo to update LinearFDBTop if needed.
	 */
	if (p_sw->nvlink_sw_data.lft_split_state == IB_LFT_SPLIT_STATE_LFT_SPLIT_RESET)
		nvlink_update_fwd_top(p_sm, p_sw, p_sw->max_lid_ho);
}

static void nvlink_lft_split_set_new_ranges_values(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_nvlink_mgr_t *p_mgr = (osm_nvlink_mgr_t *)context;
	ib_linear_forwarding_table_split_t *p_lft_split = &p_sw->nvlink_sw_data.calc_lft_split;

	memset(p_lft_split, 0, sizeof(ib_linear_forwarding_table_split_t));

	ib_lft_split_set_global_lid_range(p_lft_split, IB_LID_UCAST_START_HO,
					  OSM_NVLINK_GLOBAL_LID_RANGE_LEN,
					  p_mgr->global_lid_range_top);

	ib_lft_split_set_alid_range(p_lft_split, p_mgr->alid_range_start,
				    OSM_NVLINK_NVLID_RANGE_LEN,
				    p_mgr->alid_range_top);

	ib_lft_split_set_local_plane_range(
	    p_lft_split,
	    p_mgr->p_start_ranges_per_plane[p_sw->nvlink_sw_data.plane],
	    OSM_NVLINK_NVLID_RANGE_LEN,
	    p_mgr->lid_range_top_per_plane[p_sw->nvlink_sw_data.plane]);
}

static void nvlink_set_lft_split(cl_map_item_t * item, void *ctx)
{
	osm_sm_t *p_sm = (osm_sm_t *)ctx;
	osm_subn_t *p_subn = p_sm->p_subn;
	osm_switch_t *p_sw = (osm_switch_t *)item;
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_linear_forwarding_table_split_t * p_req_lft_split;
	ib_linear_forwarding_table_split_t * p_calc_lft_split = &p_sw->nvlink_sw_data.calc_lft_split;
	osm_lft_split_context_t context;
	osm_physp_t *p_physp0;
	uint32_t attr_mod = 0;
	ib_api_status_t status;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_LFT_SPLIT_SUPPORTED_BIT))
		return;

	p_req_lft_split = (ib_linear_forwarding_table_split_t *)payload;
	memcpy(p_req_lft_split, p_calc_lft_split, sizeof(ib_linear_forwarding_table_split_t));

	switch (p_sw->nvlink_sw_data.lft_split_state) {
	case IB_LFT_SPLIT_STATE_SWITCH_INFO_RESET:
		nvlink_lft_split_calc_reset_ranges_mad(p_subn, p_sw, p_req_lft_split);
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"Initializing LFTSplit ranges for switch 0x%"PRIx64" before"
			" setting new values\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		break;
	case IB_LFT_SPLIT_STATE_SWITCH_INFO_SET:
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"Setting new LFTSplit ranges for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		break;
	default:
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"LFTSplit state for switch 0x%"PRIx64" is %u, "
			"avoid sending LFTSplit set\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid),
			p_sw->nvlink_sw_data.lft_split_state);
		return;
	}

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A80C: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, payload,
			     sizeof(ib_linear_forwarding_table_split_t), IB_MAD_ATTR_LFT_SPLIT,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A80D: "
			"Setting LFTSplit to 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp0->port_guid), ib_get_err_str(status));

}

static void nvlink_get_lft_split(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_physp_t *p_physp;
	osm_nvlink_mad_context_t mad_context;
	ib_api_status_t status;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_LFT_SPLIT_SUPPORTED_BIT))
		return;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A815: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_sw->p_node->node_info.node_guid;
	mad_context.mad_method = IB_MAD_METHOD_GET;

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
			     IB_MAD_ATTR_LFT_SPLIT, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A816: "
			"Getting LFTSplit from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid),	ib_get_err_str(status));

	/*
	 * Beginning of LFTSplit flow. On receive response, determine whether
	 * need to update new LFTSplit values.
	 * See osm_nvlink_lftsplit_state_t for more details
	 */
	p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_NONE;
}

/* Note that after first subnet up, the ranges might be changed only if set new switch planes */
ib_api_status_t osm_nvlink_lft_split_process(osm_sm_t * sm)
{
	osm_opensm_t *p_osm = sm->p_subn->p_osm;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	/*
	 * Calculate new reguired ranges for LFTSplit so on Get Response compare to subnet
	 * values and determine whether need to get to the flow of setting new LFTSplit values.
	 */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_lft_split_set_new_ranges_values,
			   &sm->nvl_mgr);

	/* Get LFTSplit values configured on the subnet */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_get_lft_split, sm);
	if (wait_for_pending_transactions(p_osm))
		status = IB_ERROR;

	/*
	 * Set SwitchInfo.LinearFDBTop to 0.
	 * Wait for all SwitchInfo initialization MADs to be done before moving to next step
	 * of zero the LFTSplit ranges
	 */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_set_zero_fwd_top, sm);
	if (wait_for_pending_transactions(p_osm))
		status = IB_ERROR;

	/* Zero LFTSplit ranges */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_set_lft_split, sm);

	/* Wait for all value initialization MADs to be done, before setting the new values */
	if (wait_for_pending_transactions(p_osm))
		status = IB_ERROR;

	/*
	 * Need to Set new SwitchInfo.LinearFDBTop before setting
	 * new LFTSplit ranges due to firmware dependency
	 */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_set_fwd_top, sm);
	if (wait_for_pending_transactions(p_osm))
		status = IB_ERROR;

	/* Set new LFTSplit ranges */
	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, nvlink_set_lft_split, sm);
	if (wait_for_pending_transactions(p_osm))
		status = IB_ERROR;

	OSM_LOG_EXIT(sm->p_log);
	return status;
}

/* Set NVLINK parameters */
ib_api_status_t osm_nvlink_init(osm_sm_t * p_sm)
{
	osm_nvlink_mgr_t *p_mgr = &p_sm->nvl_mgr;
	uint8_t plane;
	uint16_t next_range_start;

	memset(p_mgr, 0, sizeof(osm_nvlink_mgr_t));

	p_mgr->p_subn = p_sm->p_subn;

	p_mgr->nvlink_mad_disp_h =
	    cl_disp_register(p_sm->p_disp, OSM_MSG_MAD_NVLINK_MAD,
			     osm_nvlink_mad_rcv_process, p_sm);
	if (p_mgr->nvlink_mad_disp_h == CL_DISP_INVALID_HANDLE) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A81D:"
			"could not register dispatcher handler\n");
		return IB_INSUFFICIENT_RESOURCES;
	}


	/* As global lids take range [1, OSM_NVLINK_GLOBAL_LID_RANGE_LEN] */
	p_mgr->alid_range_start = OSM_NVLINK_GLOBAL_LID_RANGE_LEN + 1;
	p_mgr->alid_range_top = p_mgr->alid_range_start - 1;
	p_mgr->num_alids_per_gpu = OSM_NVLINK_NUM_ALIDS_PER_GPU;

	next_range_start = p_mgr->alid_range_start + OSM_NVLINK_NVLID_RANGE_LEN;
	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++) {
		p_mgr->p_start_ranges_per_plane[plane] = next_range_start;
		p_mgr->lid_range_top_per_plane[plane] = next_range_start - 1;
		next_range_start += OSM_NVLINK_NVLID_RANGE_LEN;
	}

	nvlink_init_sw_plane_db(p_mgr);
	nvlink_init_port_rail_db(p_mgr);

	return IB_SUCCESS;
}


void osm_nvlink_destroy_disp(osm_nvlink_mgr_t * p_mgr)
{
	cl_disp_unregister(p_mgr->nvlink_mad_disp_h);
}

ib_api_status_t osm_nvlink_set_gpu_extended_node_info(osm_sm_t * sm, osm_gpu_t * p_gpu)
{
	osm_node_t *p_node = p_gpu->p_node;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	/* Set ENI for GPUs should be before setting alid_tbl_top */
	p_node->new_ext_node_info.anycast_lid_table_top = sm->nvl_mgr.num_alids_per_gpu;

	if ((p_node->ext_node_info.anycast_lid_table_top ==
	    p_node->new_ext_node_info.anycast_lid_table_top) &&
	    p_node->ext_node_info_updated && !sm->p_subn->need_update)
		goto Exit;

	status = osm_ar_set_extended_node_info(sm, p_node);

Exit:
	OSM_LOG_EXIT(sm->p_log);

	return status;
}

static void nvlink_clear_subn_tables(cl_map_item_t * const p_map_item, void * context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t i, num_rail_filter_mads;

	/*
	 * For switches that marked as need_update,
	 * initialize the NVLink related MADs that were retrieved from subnet,
	 * so in case of a timeout, next memcmp would not cause skipping MAD SET.
	 */

	if (p_sw->need_update) {
		/* Initialize RailFilterConfig */
		num_rail_filter_mads = OSM_NVLINK_SN_NUM_RAILS *
				       p_sw->nvlink_sw_data.max_ingress_block *
				       p_sw->nvlink_sw_data.max_egress_block;

		for (i = 0; i < num_rail_filter_mads; i++)
			memset(&p_sw->nvlink_sw_data.subn_rail_filter_cfg[i], 0,
			       sizeof(ib_rail_filter_config_t));

		/*
		 * Initialize LFTSplit:
		 * Initialize to first state, according to which SM
		 * first sends GET LFTSplit from the switch.
		 */
		p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_NONE;

		/* Initialize NVL HBF */
		memset(&p_sw->nvlink_sw_data.subn_nvl_hbf_cfg, 0,
		       sizeof(p_sw->nvlink_sw_data.subn_nvl_hbf_cfg));


		/* Initialize ProfilesConfig */
		osm_switch_initialize_profiles_config(p_sw, IB_PROFILES_FEATURE_NVL_HBF);
	}
}

/* Retrieves information used later by lid manager */
ib_api_status_t osm_nvlink_configuration(osm_subn_t * p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;
	osm_nvlink_mgr_t *p_mgr = &sm->nvl_mgr;
	ib_api_status_t status = IB_SUCCESS;

	OSM_ASSERT(p_mgr->p_guid2planes);
	OSM_ASSERT(p_mgr->p_port2rail);

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, nvlink_get_alid_info, sm);

	/*
	 * Clear subnet tables of devices that marked as need_update to enforce MADs
	 * to resend on next sweep in case some of them failed.
	 */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_clear_subn_tables, NULL);

	/* Attempt to set planes according to plane DB */
	status = nvlink_plane_discover(p_mgr, TRUE);

	/*
	 * Failed to set planes according to DB.
	 * IB_INVALID_PORT error means:
	 * 	a link was found in which one side is FNM port and the other is not.
	 * 	TODO: For NVLink multinode, do this validation where SM configures the FNM ports.
	 *
	 * Any other error, DB is invalid, SM should rediscover the planes.
	 */
	if (status == IB_INVALID_PORT) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A813: "
			"Failed to reassign NVLink planes\n");
		status = IB_ERROR;
		goto Exit;

	} else if (status) {
		status = nvlink_plane_discover(p_mgr, FALSE);
		if (status) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A807: "
				"Failed to reassign NVLink planes\n");
			status = IB_ERROR;
			goto Exit;
		}
	}

	/* store the guid to plane table in persistent db */
	osm_db_store(p_mgr->p_guid2planes, p_subn->opt.fsync_high_avail_files);

	nvlink_rail_configuration(p_mgr);
	osm_db_store(p_mgr->p_port2rail, p_subn->opt.fsync_high_avail_files);

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);

	return status;
}

void osm_nvlink_nvl_hbf_send_profiles_config(osm_sm_t * sm,
					     osm_switch_t * p_sw)
{
	uint8_t block_num, num_ports, num_profiles_blocks;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_NVL_HBF_CONFIG_SUPPORTED_BIT))
		return;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	/* num_ports count includes port 0 */
	num_profiles_blocks = num_ports < IB_PROFILE_CONFIG_NUM_PROFILES_PER_BLOCK ? 1 : 2;

	for (block_num = 0; block_num < num_profiles_blocks; block_num++) {
		osm_subn_send_profiles_config_block(sm, p_sw, IB_PROFILES_FEATURE_NVL_HBF,
						    block_num);
	}
}

void osm_nvlink_build_alids_ar_groups(osm_subn_t * p_subn)
{
	osm_ucast_mgr_t *p_ucast_mgr = &p_subn->p_osm->sm.ucast_mgr;
	cl_map_item_t *p_sw_item, *p_gpu_item;
	osm_switch_t *p_sw;
	osm_gpu_t *p_gpu;
	osm_physp_t *p_remote_physp_by_plane, *p_gpu_physp;
	uint16_t alid_group_id, leaf_group_id, group_top;
	uint16_t alid, leaf_sw_lid;
	osm_ar_subgroup_t *p_alid_subgroup, *p_leaf_subgroup;
	uint8_t source_plane, state, gpu_port_num, num_ports, exit_port_num;
	uint8_t port_num = 0;
	osm_advanced_routing_data_t *p_ar_data;

	/*
	 * Iterate over switches, update AR groups for each switch to hold for
	 * each ALID an AR group that is a union of the AR groups represent the
	 * remote leaf switches connected to the GPU of that ALID.
	 */
	for (p_sw_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw_item = cl_qmap_next(p_sw_item)) {
		p_sw = (osm_switch_t *)p_sw_item;

		if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
			continue;

		p_ar_data = &p_sw->ar_data;
		source_plane = p_sw->nvlink_sw_data.plane;
		group_top = 0;

		/* Iterate over all GPUs of the subnet and create AR Group for each ALID (GPU) */
		for (p_gpu_item = cl_qmap_head(&p_subn->gpu_guid_tbl);
		     p_gpu_item != cl_qmap_end(&p_subn->gpu_guid_tbl);
		     p_gpu_item = cl_qmap_next(p_gpu_item)) {
			p_gpu = (osm_gpu_t *)p_gpu_item;

			/* Currently supports only a single ALID per GPU */
			alid = osm_gpu_get_alid_by_index(p_gpu, 0);

			/* If exists, returns the group id. Otherwise, creates a new one. */
			alid_group_id = osm_ucast_mgr_make_groupid(p_ucast_mgr, alid, 0);
			if (alid_group_id == UCAST_MGR_INVALID_GROUP_ID) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"Failed to find AR group id for ALID 0x%x\n", alid);
				continue;
			}

			p_alid_subgroup = osm_ar_subgroup_get(p_ar_data, alid_group_id, 0);

			p_remote_physp_by_plane = p_gpu->remote_leaf_sw_physp_arr[source_plane];
			/* GPU has no remote switch of this plane, hence no route to it */
			if (!p_remote_physp_by_plane || !p_remote_physp_by_plane->p_node->sw)
				continue;

			leaf_sw_lid = osm_node_get_base_lid_ho(p_remote_physp_by_plane->p_node, 0);

			/* NVLink configuration supports only plft 0 */
			osm_ar_lft_entry_get(p_sw, 0, leaf_sw_lid, &port_num, &state, &leaf_group_id);

			/* no routing for this lid */
			if (port_num == OSM_NO_PATH || leaf_group_id == AR_INVALID_GROUP_ID)
				continue;

			/*
			 * Remote leaf switch of the GPU is not this source switch.
			 * Route toward the ALID must be only through devices of the same
			 * plane as source switch plane.
			 * Assuming ALID is connected to a single leaf switch (at most) of each plane,
			 * then for remote switch, the ALID AR group is the same as for that leaf.
			 * (This is for NVLink multinode case)
			 */
			if (p_remote_physp_by_plane->p_node->sw != p_sw) {
				/* NVLink configuration supports a single subgroup */
				p_leaf_subgroup = osm_ar_subgroup_get(p_ar_data, leaf_group_id, 0);
				osm_ar_subgroup_add_subgroup(p_alid_subgroup, p_leaf_subgroup);
			}

			/*
			 * Source switch is the leaf that is directly connected to the GPU.
			 * Meaning ALID AR group should hold all the exit ports of this
			 * switch that are linked to the GPU.
			 * (For NVLink Umbriel case, should use only this section)
			 */
			if (p_remote_physp_by_plane->p_node->sw == p_sw) {
				num_ports = osm_node_get_num_physp(p_gpu->p_node);
				for (gpu_port_num = 1; gpu_port_num < num_ports; gpu_port_num++) {
					p_gpu_physp = osm_node_get_physp_ptr(p_gpu->p_node, gpu_port_num);
					if (!p_gpu_physp || !p_gpu_physp->p_remote_physp ||
					    p_gpu_physp->p_remote_physp->p_node != p_sw->p_node)
						continue;

					exit_port_num = p_gpu_physp->p_remote_physp->port_num;
					osm_ar_subgroup_add_port(p_alid_subgroup, exit_port_num);
					port_num = exit_port_num;
				}
			}

			/*
			 * Probably not used, as ALIDs used only for NVLink traffic (which takes
			 * AR LFTs for HBF routes).
			 * Add anyway so all devices appear on LFTs dump.
			 */
			p_sw->new_lft[alid] = port_num;
			/* NVLink configuration supports only plft 0 */
			osm_ar_lft_set(p_ar_data, alid, 0, port_num, OSM_AR_IB_LID_STATE_HBF, alid_group_id);

			if (group_top < alid_group_id)
				group_top = alid_group_id;
		}

		if (p_ar_data->group_top < group_top)
			osm_ar_set_required_group_top(p_ar_data, group_top);
	}
}
