/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2010 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009-2011 ZIH, TU Dresden, Federal Republic of Germany. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_opensm_t.
 * This object represents the opensm super object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complib/cl_dispatcher.h>
#include <complib/cl_passivelock.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_OPENSM_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_version.h>
#include <opensm/osm_base.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_vl15intf.h>
#include <opensm/osm_event_plugin.h>
#include <opensm/osm_plugin.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_router_policy.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_subscriber.h>
#include <opensm/osm_tenant_mgr.h>


static void update_start_time_str(IN osm_opensm_t * p_osm)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	sprint_time_str(&tv, p_osm->start_time_str, sizeof(p_osm->start_time_str), 1);
}

void osm_opensm_construct(IN osm_opensm_t * p_osm)
{
	memset(p_osm, 0, sizeof(*p_osm));

	update_start_time_str(p_osm);

	p_osm->osm_version = OSM_VERSION;
	osm_subn_construct(&p_osm->subn);
	osm_db_construct(&p_osm->db);
	osm_log_construct(&p_osm->log);
}

void osm_opensm_construct_finish(IN osm_opensm_t * p_osm)
{
	osm_sm_construct(&p_osm->sm, &p_osm->subn.opt);
	osm_sa_construct(&p_osm->sa);
	osm_mad_pool_construct(&p_osm->mad_pool);
	p_osm->mad_pool_constructed = TRUE;
	/* p_subn is already initialized with p_opt */
	osm_super_vl15_construct(&p_osm->super_vl15, &p_osm->subn.opt);
	p_osm->vl15_constructed = TRUE;
}


static void destroy_plugins(osm_opensm_t *p_osm)
{
	osm_plugin_t *p_plugin;

	/*
	 * Remove the plugins.  Call the unload function.  (The plugins
	 * themselves are responsible for freeing their resources.)
	 */
	while (p_plugin = osm_plugin_first(), p_plugin != NULL) {
		osm_plugin_unload(p_osm, osm_plugin_name(p_plugin), osm_plugin_data(p_plugin));
	}
}

void osm_opensm_destroy(IN osm_opensm_t * p_osm)
{
	int port_idx;

	/* in case of shutdown through exit proc - no ^C */
	osm_exit_flag = TRUE;

	/*
	 * First of all, clear the is_sm bit.
	 */
	if (p_osm->sm.mad_ctrl.h_bind[p_osm->subn.sm_port_idx])
		osm_vendor_set_sm(p_osm->sm.mad_ctrl.h_bind[p_osm->subn.sm_port_idx], FALSE);

#ifdef ENABLE_OSM_PERF_MGR
	/* Shutdown the PerfMgr */
	osm_perfmgr_shutdown(&p_osm->perfmgr);
#endif				/* ENABLE_OSM_PERF_MGR */

	osm_congestion_control_shutdown(&p_osm->cc);
	osm_vendor_specific_shutdown(&p_osm->vs);
	osm_n2n_shutdown(&p_osm->n2n);

	/* shut down the SM
	 * - make sure the SM sweeper thread exited
	 * - unbind from QP0 messages
	 */
	osm_sm_shutdown(&p_osm->sm);

	/* shut down the SA
	 * - unbind from QP1 messages
	 */
	osm_sa_shutdown(&p_osm->sa);

	/* cleanup all messages on VL15 fifo that were not sent yet */
	osm_super_vl15_shutdown(&p_osm->super_vl15, &p_osm->mad_pool);

	/* shut down the dispatcher - so no new messages cross */
	cl_disp_shutdown(&p_osm->disp);
	cl_disp_shutdown(&p_osm->sminfo_get_disp);
	cl_disp_shutdown(&p_osm->gmp_disp);
	cl_disp_shutdown(&p_osm->gmp_trap_disp);
	if (p_osm->sa_set_disp_initialized)
		cl_disp_shutdown(&p_osm->sa_set_disp);
	cl_disp_shutdown(&p_osm->sa_security_disp);

	/* dump SA DB */
	if ((p_osm->sm.p_subn->sm_state_info.sm_state == IB_SMINFO_STATE_MASTER) &&
	     p_osm->subn.opt.sa_db_dump)
		osm_sa_db_file_dump(p_osm);

	/* do the destruction in reverse order as init */
	destroy_plugins(p_osm);
	osm_sa_destroy(&p_osm->sa);
	osm_sm_destroy(&p_osm->sm);
#ifdef ENABLE_OSM_PERF_MGR
	osm_perfmgr_destroy(&p_osm->perfmgr);
#endif				/* ENABLE_OSM_PERF_MGR */
	osm_congestion_control_destroy(&p_osm->cc);
	osm_vendor_specific_destroy(&p_osm->vs);
	osm_n2n_destroy(&p_osm->n2n);

	for (port_idx = 0; port_idx < OSM_MAX_BINDING_PORTS + 1; port_idx++) {
		if (p_osm->statistic_dump_file[port_idx]) {
			fclose(p_osm->statistic_dump_file[port_idx]);
			p_osm->statistic_dump_file[port_idx] = NULL;
		}
	}

	if (p_osm->subn.opt.enable_performance_logging) {
		if (p_osm->perflog_mgr.perflog_dump_file) {
			fclose(p_osm->perflog_mgr.perflog_dump_file);
			p_osm->perflog_mgr.perflog_dump_file = NULL;
		}

		free(p_osm->perflog_mgr.current_str);
	}
}

void osm_opensm_destroy_finish(IN osm_opensm_t * p_osm)
{
	uint8_t port_index;

	if (p_osm->vl15_constructed && p_osm->mad_pool_constructed)
		osm_super_vl15_destroy(&p_osm->super_vl15, &p_osm->mad_pool);
	if (p_osm->mad_pool_constructed)
		osm_mad_pool_destroy(&p_osm->mad_pool);
	p_osm->vl15_constructed = FALSE;
	p_osm->mad_pool_constructed = FALSE;
	for (port_index = 0; port_index < p_osm->subn.opt.num_guids;
	     port_index++) {
		osm_vendor_delete(&p_osm->p_vendor[port_index]);
		if (p_osm->p_server_vendor[port_index])
			osm_vendor_delete(&p_osm->p_server_vendor[port_index]);
		if (p_osm->p_gmp_vendor[port_index])
			osm_vendor_delete(&p_osm->p_gmp_vendor[port_index]);
	}
	if (p_osm->sm.thread_state != OSM_THREAD_STATE_NONE)
		osm_subn_destroy(&p_osm->subn);
	cl_disp_destroy(&p_osm->disp);
	cl_disp_destroy(&p_osm->sminfo_get_disp);
	cl_disp_destroy(&p_osm->gmp_disp);
	cl_disp_destroy(&p_osm->gmp_trap_disp);
	osm_db_destroy(&p_osm->db);

	if (p_osm->sa_set_disp_initialized)
		cl_disp_destroy(&p_osm->sa_set_disp);

	for (port_index = 0; port_index < p_osm->subn.opt.num_guids;
	     port_index ++)
	{
#ifdef HAVE_LIBPTHREAD
		pthread_cond_destroy(&p_osm->stats[port_index].cond);
		pthread_mutex_destroy(&p_osm->stats[port_index].mutex);
#else
		cl_event_destroy(&p_osm->stats[port_index].event);
#endif
	}
	if (p_osm->node_name_map)
		close_node_name_map(p_osm->node_name_map);
	cl_plock_destroy(&p_osm->lock);
	cl_plock_destroy(&p_osm->routing_lock);
	osm_log_destroy(&p_osm->log);
}

static void load_plugins(IN osm_opensm_t *osm, IN const char *names)
{
	void *p_data;
	char *p_names, *name, *p;
	const char *delimiters = " \t\n";

	p_names = strdup(names);
	name = strtok_r(p_names, delimiters, &p);
	while (name && *name) {
		p_data = osm_plugin_load(osm, name, osm);
		if (p_data == NULL) {
			osm_log_v2(&osm->log, OSM_LOG_ERROR, FILE_ID,
				   "ERR 1000: cannot load plugin \'%s\'\n",
				   name);
		}
		name = strtok_r(NULL, delimiters, &p);
	}
	free(p_names);
}

static void load_event_plugins(osm_opensm_t *osm, const char *plugin_names)
{
	load_plugins(osm, plugin_names);
}

ib_api_status_t osm_opensm_init(IN osm_opensm_t * p_osm,
				IN const osm_subn_opt_t * p_opt)
{
	cl_status_t cl_status;
	ib_api_status_t status;
	uint8_t port_index;

	/* Can't use log macros here, since we're initializing the log */
	osm_opensm_construct(p_osm);

	if (p_opt->daemon)
		p_osm->log.daemon = 1;

	status = osm_log_init_v3(&p_osm->log, p_opt->force_log_flush,
				 p_opt->log_flags, p_opt->syslog_log_flags,
				 p_opt->log_file, p_opt->log_max_size,
				 p_opt->accum_log_file);
	if (status != IB_SUCCESS)
		return status;
	p_osm->log.log_prefix = p_opt->log_prefix;

	/* If there is a log level defined - add the OSM_VERSION to it */
	osm_log_v2(&p_osm->log,
		   osm_log_get_level(&p_osm->log) & (OSM_LOG_SYS ^ 0xFF),
		   FILE_ID, "%s\n", p_osm->osm_version);
	/* Write the OSM_VERSION to the SYS_LOG */
	osm_log_v2(&p_osm->log, OSM_LOG_SYS, FILE_ID, "%s\n", p_osm->osm_version);	/* Format Waived */

	OSM_LOG(&p_osm->log, OSM_LOG_FUNCS, "[\n");	/* Format Waived */

	cl_status = cl_plock_init(&p_osm->lock);
	if (cl_status != CL_SUCCESS)
		goto Exit;

	cl_status = cl_plock_init(&p_osm->routing_lock);
	if (cl_status != CL_SUCCESS)
		goto Exit;

	for (port_index = 0; port_index < p_opt->num_guids; port_index ++)
	{
#ifdef HAVE_LIBPTHREAD
		pthread_mutex_init(&p_osm->stats[port_index].mutex, NULL);
		pthread_cond_init(&p_osm->stats[port_index].cond, NULL);
#else
		cl_status = cl_event_init(&p_osm->stats[port_index].event, FALSE);
		if (cl_status != CL_SUCCESS)
			goto Exit;
#endif
	}

	if (p_opt->single_thread) {
		OSM_LOG(&p_osm->log, OSM_LOG_INFO,
			"Forcing single threaded dispatcher\n");
		cl_status = cl_disp_init(&p_osm->disp, 1, "opensm");
	} else {
		/*
		 * Normal behavior is to initialize the dispatcher with
		 * one thread per CPU, as specified by a thread count of '0'.
		 */
		cl_status = cl_disp_init(&p_osm->disp, p_opt->smp_threads, "opensm");
	}
	if (cl_status != CL_SUCCESS)
		goto Exit;

	/* Unless OpenSM runs in single threaded mode, we create new single
	 * threaded dispatcher for SA Set and Delete request.
	 */
	p_osm->sa_set_disp_initialized = FALSE;
	if (!p_opt->single_thread) {
		cl_status = cl_disp_init(&p_osm->sa_set_disp, 1, "subnadmin_set");
		if (cl_status != CL_SUCCESS)
			goto Exit;
		p_osm->sa_set_disp_initialized = TRUE;
	}

	osm_plugin_init(p_osm);
	osm_routing_engine_init();
	osm_adv_routing_engine_init();

	/* the DB is in use by subn so init before */
	status = osm_db_init(&p_osm->db, &p_osm->log);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_status = cl_disp_init(&p_osm->sminfo_get_disp, 1, "sminfo_get");
	if (cl_status != CL_SUCCESS)
		goto Exit;

	cl_status = cl_disp_init(&p_osm->gmp_disp, p_opt->gmp_threads, "gmp");
	if (cl_status != CL_SUCCESS)
		goto Exit;

	cl_status = cl_disp_init(&p_osm->sa_security_disp, 1, "sa_security");
	if (cl_status != CL_SUCCESS)
		goto Exit;

	cl_status = cl_disp_init(&p_osm->gmp_trap_disp, p_opt->gmp_traps_threads_num, "gmp_trap");
	if (cl_status != CL_SUCCESS)
		goto Exit;

	status = osm_subn_init(&p_osm->subn, p_osm, p_opt);
	if (status != IB_SUCCESS)
		goto Exit;

	/* smp vendors for OpenSM auxiliary ports */
	for (port_index = 0; port_index < p_opt->num_guids; port_index++) {
		p_osm->p_vendor[port_index] =
		    osm_vendor_on_port_new(&p_osm->log,
					   p_opt->transaction_timeout,
					   port_index);
		if (p_osm->p_vendor[port_index] == NULL)
			status = IB_INSUFFICIENT_RESOURCES;

		p_osm->p_server_vendor[port_index] =
		    osm_vendor_on_port_new(&p_osm->log,
					   p_opt->transaction_timeout,
					   port_index);
		if (p_osm->p_server_vendor[port_index] == NULL)
			status = IB_INSUFFICIENT_RESOURCES;
	}

	for (port_index = 0; port_index < p_opt->num_guids; port_index++) {
		p_osm->p_gmp_vendor[port_index] =
		    osm_vendor_on_port_new(&p_osm->log,
					   p_opt->transaction_timeout,
					   port_index);
		if (p_osm->p_gmp_vendor[port_index] == NULL)
			status = IB_INSUFFICIENT_RESOURCES;
	}

	for (port_index = 0; port_index < OSM_MAX_BINDING_PORTS + 1; port_index++)
		p_osm->statistic_dump_file[port_index] = NULL;

	p_osm->perflog_mgr.perflog_max_size = p_opt->osm_perflog_dump_limit << 20;
	if (p_opt->enable_performance_logging) {
		p_osm->perflog_mgr.perflog_dump_file = NULL;
		p_osm->perflog_mgr.current_str = (char *) calloc(1, p_osm->perflog_mgr.perflog_max_size);
		if (!p_osm->perflog_mgr.current_str) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 1001: current_str memory allocation failed\n");
			status = IB_INSUFFICIENT_RESOURCES;
		}
	}

	p_osm->statistic_max_size = p_opt->osm_stats_dump_limit << 20;

Exit:
	if (cl_status != CL_SUCCESS) {
		status = IB_UNKNOWN_ERROR;
	}

	OSM_LOG(&p_osm->log, OSM_LOG_FUNCS, "]\n");	/* Format Waived */
	return status;
}

ib_api_status_t osm_opensm_init_finish(IN osm_opensm_t * p_osm,
				       IN const osm_subn_opt_t * p_opt)
{
	ib_api_status_t status;

	osm_opensm_construct_finish(p_osm);

	p_osm->subn.sm_port_guid = p_opt->guid[p_osm->subn.sm_port_idx];

	status = osm_mad_pool_init(&p_osm->mad_pool);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_super_vl15_init(&p_osm->super_vl15, p_osm->p_vendor,
				     &p_osm->log, p_osm->stats, &p_osm->subn,
				     p_opt);
	if (status != IB_SUCCESS)
		goto Exit;

	/* Router policy initialization is needed in order to determine FLID ranges
	 * prior to lid manager initialization in osm_sm_init. */
	if (p_opt->router_policy_enabled)
		osm_router_policy_init(&p_osm->subn);

	status = osm_sm_init(&p_osm->sm, &p_osm->subn, &p_osm->db,
			     p_osm->p_vendor, p_osm->p_server_vendor, &p_osm->mad_pool, &p_osm->super_vl15,
			     &p_osm->log, p_osm->stats, &p_osm->disp, &p_osm->sminfo_get_disp,
			     &p_osm->lock);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_sa_init(&p_osm->sm, &p_osm->sa, &p_osm->subn,
			     p_osm->p_gmp_vendor, &p_osm->mad_pool, &p_osm->log,
			     &p_osm->stats[0], &p_osm->gmp_disp,
			     p_opt->single_thread ? NULL : &p_osm->sa_set_disp,
			     &p_osm->sa_security_disp,
			     &p_osm->lock, &p_osm->routing_lock);

	if (status != IB_SUCCESS)
		goto Exit;

	if (p_opt->plugin_name)
		load_plugins(p_osm, p_opt->plugin_name);

	if (p_opt->event_plugin_name)
		load_event_plugins(p_osm, p_opt->event_plugin_name);

#ifdef ENABLE_OSM_PERF_MGR
	status = osm_perfmgr_init(&p_osm->perfmgr, p_osm, p_opt);
	if (status != IB_SUCCESS)
		goto Exit;
#endif				/* ENABLE_OSM_PERF_MGR */

	status = osm_n2n_init(&p_osm->n2n, p_osm, p_opt);
        if (status != IB_SUCCESS)
                goto Exit;

	status = osm_vendor_specific_init(&p_osm->vs, p_osm, p_opt);
        if (status != IB_SUCCESS)
                goto Exit;

	status = osm_congestion_control_init(&p_osm->cc,
					     p_osm, p_opt);
	if (status != IB_SUCCESS)
		goto Exit;

	p_osm->subn.no_fallback_routing_engine = FALSE;
	p_osm->subn.is_auto_engine_used = FALSE;

	osm_setup_routing_engines(&p_osm->subn, p_opt->routing_engine_names);
	osm_setup_advanced_routing_engine(&p_osm->subn,
					  p_opt->adv_routing_engine_name);

	p_osm->subn.routing_engine_used = NULL;

	p_osm->node_name_map = NULL;

	osm_update_rtr_selection_algo_parameters(&p_osm->subn,
						 p_opt->rtr_selection_algo_parameters_s);
	osm_update_rtr_selection_function(&p_osm->subn, p_opt->rtr_selection_function_s);

	osm_update_allowed_sm_guids(&p_osm->subn, p_opt->allowed_sm_guids);

	osm_setup_hbf_weights(&p_osm->subn, p_opt->hbf_weights);

Exit:
	OSM_LOG(&p_osm->log, OSM_LOG_FUNCS, "]\n");	/* Format Waived */
	return status;
}

ib_api_status_t osm_opensm_bind(IN osm_opensm_t * p_osm, IN ib_net64_t * guids)
{
	ib_api_status_t status;

	OSM_LOG_ENTER(&p_osm->log);

	status = osm_sm_bind(&p_osm->sm, guids);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_sa_bind(&p_osm->sa, guids[p_osm->subn.sm_port_idx]);
	if (status != IB_SUCCESS)
		goto Exit;

#ifdef ENABLE_OSM_PERF_MGR
	status = osm_perfmgr_bind(&p_osm->perfmgr, guids[p_osm->subn.sm_port_idx]);
	if (status != IB_SUCCESS)
		goto Exit;
#endif				/* ENABLE_OSM_PERF_MGR */

	status = osm_congestion_control_bind(&p_osm->cc, guids[p_osm->subn.sm_port_idx]);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_vendor_specific_bind(&p_osm->vs, guids[p_osm->subn.sm_port_idx]);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_n2n_bind(&p_osm->n2n, guids[p_osm->subn.sm_port_idx]);
	if (status != IB_SUCCESS)
		goto Exit;

	/* setting IS_SM in capability mask */
	if (p_osm->subn.opt.stand_alone != OSM_STANDALONE_SA) {
		OSM_LOG(&p_osm->log, OSM_LOG_INFO, "Setting IS_SM on port 0x%016" PRIx64 "\n",
			cl_ntoh64(guids[p_osm->subn.sm_port_idx]));
		osm_vendor_set_sm(p_osm->sm.mad_ctrl.h_bind[p_osm->subn.sm_port_idx], TRUE);
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

void osm_opensm_report_event(IN osm_opensm_t *p_osm,
			     IN osm_epi_event_id_t event_id,
			     IN void *event_data)
{
	osm_subn_t		*p_subn = &p_osm->subn;
	osm_subscriber_t	*p_subscriber;
	osm_event_callbacks_t	*p_callbacks;

	if (osm_exit_flag == 0) {
		foreach_event_subscriber(p_subn, p_subscriber) {
			p_callbacks = (osm_event_callbacks_t *)p_subscriber->callbacks;
			p_callbacks->report(p_subscriber->context, event_id, event_data);
		}
	}
}

#include <opensm/osm_modnames.h>

int find_module_name(const char *name, uint8_t *file_id)
{
	int		status = 1;
	size_t		i;
	size_t		len = sizeof(module_names) / sizeof(module_names[0]);

	if (file_id != NULL) {
		for (i = 0; i < len; i++) {
			if (strcmp(name, module_names[i]) == 0) {
				*file_id = (uint8_t)i;
				status = 0;
				break;
			}
		}
	}

	return status;
}
