/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for planarization feature
 *
 * Author:
 *	Gull Sharon, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PLANARIZED_C
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_port.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>

boolean_t osm_aport_need_update(IN osm_aport_t * p_aport)
{
	osm_physp_t *p_physp;
	uint8_t plane;

	if (!p_aport || !osm_aport_is_valid(p_aport))
		return FALSE;

	for (plane = p_aport->min_found_plane_number; plane <= p_aport->max_found_plane_number; plane++) {
		p_physp = p_aport->plane_ports[plane];
		if (p_physp && p_physp->need_update)
			return TRUE;
	}
	return FALSE;
}

static const char *aports_system_type_strings[OSM_APORTS_SYSTEM_TYPE_MAX] = {
	"switch",
	"ca"
};

const char * osm_aports_system_get_type_str(IN const osm_aports_system_t * p_aports_system)
{
	return aports_system_type_strings[p_aports_system->type];
}

boolean_t osm_planar_link_is_internal_system_fnm_link(IN struct osm_subn * p_subn,
						      IN struct osm_physp * p_physp)
{
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_remote_physp;

	if (!p_physp)
		return FALSE;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		return FALSE;

	p_node = osm_physp_get_node_ptr(p_physp);
	p_remote_node = osm_physp_get_node_ptr(p_remote_physp);

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
	    osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
		return FALSE;

	if (!osm_physp_is_fnm(p_subn, p_physp) || !osm_physp_is_fnm(p_subn, p_remote_physp))
		return FALSE;

	if (p_node->node_info.sys_guid != p_remote_node->node_info.sys_guid)
		return FALSE;

	return TRUE;
}

/*
 * Given a physp of a host connected to a switch, this returns the AN physp connected to this
 * switch, and NULL if it can't be found.
 * i.e. given the following links: host -- switch -- AN,
 * an_physp = host_physp_get_neighbor_an_physp(host_physp)
 */
static osm_physp_t * host_physp_get_neighbor_an_physp(osm_physp_t *p_physp) {
	osm_physp_t *p_remote_physp, *p_sw_an_physp, *p_an_physp;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp || !osm_physp_is_valid(p_remote_physp) ||
	    !p_remote_physp->p_node || !p_remote_physp->p_node->sw ||
	    p_remote_physp->p_node->sw->an_port_num == OSM_NO_PATH)
		return NULL;

	p_sw_an_physp = osm_node_get_physp_ptr(p_remote_physp->p_node,
					       p_remote_physp->p_node->sw->an_port_num);
	if (!p_sw_an_physp || !osm_physp_is_valid(p_sw_an_physp))
		return NULL;

	p_an_physp = osm_physp_get_remote(p_sw_an_physp);
	if (!p_an_physp || !osm_physp_is_valid(p_an_physp))
		return NULL;

	return p_an_physp;
}

static ib_api_status_t
calc_end_port_plane_filter_config(osm_sm_t *p_sm, osm_port_t *p_port, boolean_t is_prism,
				  ib_end_port_plane_filter_config_t *p_endport_plane_filter)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_port->p_physp);
	osm_physp_t *p_physp, *p_an_physp;
	uint16_t base_lid_ho, max_lid_ho, lid_ho;
	uint8_t plane, entry;

	memset(p_endport_plane_filter, 0, sizeof(ib_end_port_plane_filter_config_t));

	for (plane = p_aport->min_found_plane_number; plane <= p_aport->max_found_plane_number; plane++) {
		p_physp = p_aport->plane_ports[plane];
		if (!p_physp || !osm_physp_is_valid(p_physp)) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B402: "
				"Port 0x%" PRIx64 " plane %u has no physp\n",
				osm_port_get_guid_ho(p_port), plane);
			return IB_ERROR;
		}

		entry = osm_physp_get_port_num(p_physp) - 1;
		OSM_ASSERT(entry < IB_END_PORT_PLANE_FILTER_CONFIG_NUM_ENTRIES);

		p_an_physp = host_physp_get_neighbor_an_physp(p_physp);
		if (!p_an_physp) {
			/*
			 * If there's no AN, the MAD should still be sent in order to clear the
			 * corresponding entry in the filter that's already configured in the host.
			 */
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
				"Can't get port 0x%" PRIx64 " plane %u remote switch's AN\n",
				osm_port_get_guid_ho(p_port), plane);
			continue;
		}

		base_lid_ho = cl_ntoh16(osm_physp_get_base_lid(p_an_physp));

		if (is_prism) {
			lid_ho = base_lid_ho + plane - 1;
			max_lid_ho = base_lid_ho + (1 << osm_physp_get_lmc(p_an_physp)) - 1;

			if (lid_ho > max_lid_ho) {
				OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B405: "
					"AN 0x%" PRIx64 " has LID range %u-%u, "
					"but LID %u is needed for plane %u\n",
					cl_ntoh64(osm_physp_get_port_guid(p_an_physp)), base_lid_ho,
					max_lid_ho, lid_ho, plane);
				return IB_ERROR;
			}
		} else {
			/* For AN on non-prism switch, just take base LID */
			lid_ho = base_lid_ho;
		}

		ib_end_port_plane_filter_config_set_lid_ho(p_endport_plane_filter, entry, lid_ho);
	} /* Planes loop */

	return IB_SUCCESS;
}

static void send_end_port_plane_filter_config(osm_sm_t *p_sm, osm_port_t *p_port,
					      ib_end_port_plane_filter_config_t *p_endport_plane_filter)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_port->p_physp);
	osm_end_port_plane_filter_config_context_t mad_context;
	ib_api_status_t status;

	if (!p_sm->p_subn->need_update && !osm_aport_need_update(p_aport) &&
	    !memcmp(p_endport_plane_filter,
		    &p_port->planarized_data.subn_end_port_plane_filter_config,
		    sizeof(ib_end_port_plane_filter_config_t)))
		return;

	/* Clear subn data, to make sure MAD is resent if current MAD has a timeout. */
	memset(&p_port->planarized_data.subn_end_port_plane_filter_config, 0,
	       sizeof(ib_end_port_plane_filter_config_t));

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.port_guid = osm_port_get_guid(p_port);

	/* Can send it to a single physp, since this will set it for the whole aport */
	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_port->p_physp),
			     p_port->p_physp, (const uint8_t *)p_endport_plane_filter,
			     sizeof(ib_end_port_plane_filter_config_t),
			     IB_MAD_ATTR_END_PORT_PLANE_FILTER_CONFIG, 0, FALSE,
			     ib_port_info_get_m_key(&p_port->p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B406: "
			"Setting EndPortPlaneFilterConfig of port 0x%" PRIx64 " failed (%s)\n",
			osm_port_get_guid_ho(p_port), ib_get_err_str(status));
}

void osm_planar_set_end_port_plane_filter_config(IN cl_map_item_t * p_item, IN void * context)
{
	osm_sm_t *p_sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)p_item;
	osm_aport_t *p_aport;
	osm_physp_t *p_remote_physp;
	ib_end_port_plane_filter_config_t endport_plane_filter;
	boolean_t is_prism;

	/* Relevant only for HCAs. */
	if (!p_port->p_node || osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_CA)
		return;

	if (!ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
					    IB_GENERAL_INFO_IS_END_PORT_PLANE_FILTER_CONFIG_SUPPORTED_BIT))
		return;

	/* Relevant only for planarized ports. */
	if (!p_port->p_physp || !osm_hierarchy_is_planarized_physp(p_port->p_physp))
		return;

	p_aport = osm_physp_get_aport(p_port->p_physp);
	if (!p_aport || !osm_aport_is_valid(p_aport)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B407: "
			"Planarized port 0x%" PRIx64 " has no aport. "
			"Cannot set EndPortPlaneFilterConfig\n",
			osm_port_get_guid_ho(p_port));
		return;
	}

	/* Relevant only if remote is a switch. */
	p_remote_physp = osm_physp_get_remote(p_port->p_physp);
	if (!p_remote_physp || !p_remote_physp->p_node || !p_remote_physp->p_node->sw)
		return;

	is_prism = osm_switch_is_prism_switch(p_remote_physp->p_node->sw, NULL);
	if (calc_end_port_plane_filter_config(p_sm, p_port, is_prism, &endport_plane_filter) !=
	    IB_SUCCESS)
		return;

	send_end_port_plane_filter_config(p_sm, p_port, &endport_plane_filter);
}

static void destroy_aports_systems_table(IN cl_qmap_t * p_aports_system_id_tbl)
{
	cl_map_item_t * p_map_item;
	osm_aports_system_t *p_aports_system;

	while (cl_qmap_count(p_aports_system_id_tbl)) {
		p_map_item = cl_qmap_head(p_aports_system_id_tbl);
		cl_qmap_remove_item(p_aports_system_id_tbl, p_map_item);
		p_aports_system = (osm_aports_system_t *)p_map_item;
		osm_aports_system_delete(&p_aports_system);
	}
}

void osm_aports_systems_tables_init(IN osm_aports_systems_tables_t * p_aports_systems_tables)
{
	cl_qmap_init(&p_aports_systems_tables->switch_aports_systems);
	cl_qmap_init(&p_aports_systems_tables->ca_aports_systems);
}

void osm_aports_systems_tables_destroy(IN osm_aports_systems_tables_t * p_aports_systems_tables)
{
	destroy_aports_systems_table(&p_aports_systems_tables->ca_aports_systems);
	destroy_aports_systems_table(&p_aports_systems_tables->switch_aports_systems);
}

osm_switch_system_t * osm_switch_get_planarized_system(IN struct osm_subn * p_subn,
						       IN struct osm_switch * p_sw)
{
	cl_qmap_t *p_sw_aports_systems_tbl;
	osm_switch_system_t *p_sw_system;
	ib_net64_t sys_guid;

	p_sw_aports_systems_tbl = &p_subn->aports_systems_tables.switch_aports_systems;
	sys_guid = p_sw->p_node->node_info.sys_guid;

	p_sw_system = (osm_switch_system_t *) cl_qmap_get(p_sw_aports_systems_tbl, sys_guid);
	if (p_sw_system == (osm_switch_system_t *)cl_qmap_end(p_sw_aports_systems_tbl))
		return NULL;

	return p_sw_system;
}

osm_physp_t* osm_planar_get_physp_by_plane(IN osm_physp_t *p_physp, IN uint8_t plane)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_physp);

	if (!p_aport || !plane || plane > IB_MAX_NUM_OF_PLANES_XDR)
		return NULL;
	return p_aport->plane_ports[plane];
}

osm_node_t* osm_planar_get_remote_node_by_plane(IN osm_subn_t * p_subn,
						IN osm_port_t * p_port,
						IN uint8_t plane,
						OUT uint8_t * p_remote_port_num)
{
	osm_node_t *p_node = osm_port_get_node(p_port);
	uint8_t port_num, plane_port_num;
	osm_physp_t *p_plane_physp = NULL;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	OSM_ASSERT(p_node);
	port_num = osm_physp_get_port_num(osm_port_get_physp(p_port));

	p_plane_physp = osm_planar_get_physp_by_plane(p_port->p_physp, plane);

	if (!plane || !p_plane_physp) {
		return osm_node_get_remote_node(p_node, port_num, p_remote_port_num);
	}
	plane_port_num = osm_physp_get_port_num(p_plane_physp);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return osm_node_get_remote_node(p_node, plane_port_num, p_remote_port_num);
}

uint8_t osm_planar_get_plane_mask(IN osm_physp_t *p_physp)
{
	if (!osm_hierarchy_is_planarized_physp(p_physp))
		return 0;

	uint8_t member_num_of_planes = 0, plane_mask = 0;
	member_num_of_planes = osm_hierarchy_get_num_of_planes(p_physp);

	plane_mask = (1 << (member_num_of_planes + 1)) - 2;
	return plane_mask;
}

boolean_t osm_planar_are_all_switches_same_system(IN osm_subn_t * p_subn)
{
	osm_switch_t *p_sw;
	cl_map_item_t *sw_item;
	ib_net64_t sys_guid = 0;
	boolean_t is_single_system = TRUE;

	if (p_subn->is_planarized &&
	    cl_qmap_count(&p_subn->sw_guid_tbl) <= IB_MAX_NUM_OF_PLANES_XDR) {
		for (sw_item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     sw_item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     sw_item = cl_qmap_next(sw_item)) {
			p_sw = (osm_switch_t *)sw_item;
			if (!sys_guid) {
				sys_guid = p_sw->p_node->node_info.sys_guid;
				continue;
			}
			/* switches with different system guid, don't belong to the same system */
			if (p_sw->p_node->node_info.sys_guid != sys_guid) {
				is_single_system = FALSE;
			}
		}
	} else {
		is_single_system = FALSE;
	}

	return is_single_system;
}
