/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation for port Profiles Configuration MAD receive function
 *
 * Author:
 *    Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PROFILES_CONFIG_RCV_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_fast_recovery.h>

void osm_profiles_config_qos_config_vl_rcv_process(osm_switch_t * p_sw,
						   void * p_payload,
						   osm_profiles_context_t * p_context)
{
	/*
	 * For each feature supported by port profiles, SM shall continue to
	 * configure only after all profilesConfig blocks were set.
	 * Hence, mark block on bitmask when block is set successfully
	 */
	if (p_context->mad_method == IB_MAD_METHOD_SET)
		osm_switch_set_profiles_configured(p_sw, p_context->block, TRUE,
						   IB_PROFILES_FEATURE_QOS_VL_CONFIG);

	memcpy(&p_sw->qos_config_vl_sw_data.ports_profiles[p_context->block],
	       p_payload, sizeof(ib_profiles_config_t));

	p_sw->qos_config_vl_sw_data.qos_config_vl_need_update = TRUE;
}

void osm_profiles_config_fast_recovery_rcv_process(osm_switch_t * p_sw,
						   void * p_payload,
						   osm_profiles_context_t * p_context)
{

	/*
	 * For each feature supported by port profiles, SM shall continue to
	 * configure only after all profilesConfig blocks were set.
	 * Hence, mark block on bitmask when block is set successfully
	 */
	if (p_context->mad_method == IB_MAD_METHOD_SET)
		osm_switch_set_profiles_configured(p_sw, p_context->block, TRUE,
						   IB_PROFILES_FEATURE_FAST_RECOVERY);

	memcpy(&p_sw->fast_recovery_data.ports_profiles[p_context->block],
	       p_payload, sizeof(ib_profiles_config_t));
}

void osm_profiles_config_nvl_hbf_rcv_process(osm_switch_t * p_sw,
					     void * p_payload,
					     osm_profiles_context_t * p_context)
{
	/*
	 * FW flow requirement is that in order to set new values, must:
	 * 	first, set the profile number of each port,
	 * 	then, set the MAD with the new values for that profile.
	 * Hence, use bitmask to mark which ProfilesConfig blockes were set
	 * successfully.
	 */

	if (p_context->mad_method == IB_MAD_METHOD_SET)
		osm_switch_set_profiles_configured(p_sw, p_context->block, TRUE,
						   IB_PROFILES_FEATURE_NVL_HBF);

	/* Mark that the switch received the successful response */
	p_sw->ar_data.pending_mads--;

	memcpy(&p_sw->nvlink_sw_data.ports_profiles[p_context->block],
	       p_payload, sizeof(ib_profiles_config_t));
}

void osm_profiles_config_rcv_process(IN void * context, IN void * data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	void *p_payload;
	osm_profiles_context_t *p_context;
	boolean_t verbose_bypass = FALSE;

	osm_node_t *p_node;
	ib_net64_t node_guid;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = (osm_profiles_context_t *)&p_madw->context;
	p_payload = ib_smp_get_payload_ptr(p_smp);

	node_guid = p_context->node_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || !p_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A201: "
			"Received %s for nonexistent switch node "
			"0x%" PRIx64 "\n", ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
		goto Exit;
	}
	verbose_bypass = osm_verbose_bypass_handle_profiles_config(sm, p_node);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	switch (p_context->feature) {
	case IB_PROFILES_FEATURE_FAST_RECOVERY:
		osm_profiles_config_fast_recovery_rcv_process(p_node->sw, p_payload, p_context);
		break;
	case IB_PROFILES_FEATURE_QOS_VL_CONFIG:
		osm_profiles_config_qos_config_vl_rcv_process(p_node->sw, p_payload, p_context);
		break;
	case IB_PROFILES_FEATURE_NVL_HBF:
		osm_profiles_config_nvl_hbf_rcv_process(p_node->sw, p_payload, p_context);
		break;
	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A202: "
			"Received %s from switch node 0x%" PRIx64
			" for non supported feature %u\n",
			ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid), p_context->feature);
	}

Exit:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
