/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2006-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2010-2013 Mellanox, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of OpenSM QoS infrastructure primitives
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_QOS_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_qos_policy.h>
#include <opensm/osm_unhealthy_ports.h>

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);

struct qos_config {
	uint8_t max_vls;
	uint8_t vl_high_limit;
	ib_vl_arb_table_t vlarb_high[2];
	ib_vl_arb_table_t vlarb_low[2];
	ib_slvl_table_t sl2vl;
};

typedef struct sl2vl_port_mask {
	ib_slvl_table_t sl2vl_configuration;
	uint8_t input_port_mask[IB_PORT_MASK_BYTES_SIZE];
} sl2vl_port_mask_t;

typedef struct switch_sl2vl_map_item {
	cl_fmap_item_t map_item;
	struct sl2vl_port_mask in_sl2vl_port_mask;
	uint8_t output_port_mask[IB_PORT_MASK_BYTES_SIZE];
} switch_sl2vl_map_item_t;

typedef struct end_port_qos_task {
	osm_port_t *p_port;
	struct qos_config *ep_qos_config;
	struct qos_config *port0_qos_config;
	int ep_vlarb_only;
	int port0_vlarb_only;
} end_port_qos_task_t;

static void qos_build_config(struct qos_config *cfg,
			     osm_qos_options_t * opt,
			     osm_qos_options_t * dflt);

/*
 * QoS primitives
 */

static osm_madw_t *osm_qos_mad_create(IN osm_sm_t * sm,
					  IN osm_physp_t * p,
					  IN uint32_t data_size,
					  IN uint8_t * p_data,
					  IN ib_net16_t attr_id,
					  IN uint32_t attr_mod)

{
	osm_madw_context_t context;
	osm_madw_t *p_madw;
	osm_node_t *p_node;
	osm_physp_t *physp0;
	ib_net64_t m_key;
	uint32_t timeout = 0;

	p_node = osm_physp_get_node_ptr(p);
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH &&
	    osm_physp_get_port_num(p) != 0) {
		physp0 = osm_node_get_physp_ptr(p_node, 0);
		m_key = ib_port_info_get_m_key(&physp0->port_info);
	} else
		m_key = ib_port_info_get_m_key(&p->port_info);

	switch (attr_id){
	case IB_MAD_ATTR_SLVL_TABLE:
		context.slvl_context.node_guid = osm_node_get_node_guid(p_node);
		context.slvl_context.port_guid = osm_physp_get_port_guid(p);
		context.slvl_context.set_method = TRUE;
		if ((attr_mod & 0x130000) != 0)	/* optimized ? */
			timeout = sm->p_subn->opt.long_transaction_timeout;
		break;
	case IB_MAD_ATTR_VL_ARBITRATION:
		context.vla_context.node_guid = osm_node_get_node_guid(p_node);
		context.vla_context.port_guid = osm_physp_get_port_guid(p);
		context.vla_context.set_method = TRUE;
		break;
	default:
		return NULL;
	}

	p_madw = osm_prepare_req_set(sm, osm_physp_get_dr_path_ptr(sm, p),
				     p_data, data_size,
				     attr_id, cl_hton32(attr_mod),
				     FALSE, m_key,
				     timeout, CL_DISP_MSGID_NONE,
				     osm_node_get_node_guid(p_node), &context);

	return p_madw;
}

static ib_api_status_t vlarb_update_table_block(osm_sm_t * sm,
						osm_physp_t * p,
						uint8_t port_num,
						unsigned force_update,
						const ib_vl_arb_table_t *
						table_block,
						unsigned block_length,
						unsigned block_num,
						cl_qlist_t *mad_list)
{
	struct osm_routing_engine *re = sm->p_subn->routing_engine_used;
	ib_vl_arb_table_t block;
	uint32_t attr_mod;
	unsigned vl_mask, i;
	osm_madw_t *p_mad;
	vl_mask = (1 << (ib_port_info_get_op_vls(&p->port_info) - 1)) - 1;

	memset(&block, 0, sizeof(block));
	memcpy(&block, table_block, block_length * sizeof(block.vl_entry[0]));

	if (re && re->update_vlarb)
		re->update_vlarb(re->context, p, port_num, &block,
				 block_length, block_num);

	for (i = 0; i < block_length; i++)
		block.vl_entry[i].vl &= vl_mask;

	if (!force_update &&
	    !memcmp(&p->vl_arb[block_num], &block,
		    block_length * sizeof(block.vl_entry[0])))
		return IB_SUCCESS;

	attr_mod = ((block_num + 1) << 16) | port_num;

	p_mad = osm_qos_mad_create(sm,p,sizeof(block),(uint8_t *) & block,
				   IB_MAD_ATTR_VL_ARBITRATION, attr_mod);

	if (!p_mad)
		return IB_INSUFFICIENT_MEMORY;

	/*
	 * Zero the stored VL Arbitration block, so in case the MAD will
	 * end up with error, we will resend it in the next sweep.
	 */
	memset(&p->vl_arb[block_num], 0,
	       block_length * sizeof(block.vl_entry[0]));

	cl_qlist_insert_tail(mad_list, &p_mad->list_item);

	return IB_SUCCESS;
}

static ib_api_status_t vlarb_update(osm_sm_t * sm, osm_physp_t * p,
				    uint8_t port_num, unsigned force_update,
				    const struct qos_config *qcfg,
				    cl_qlist_t *mad_list)
{
	ib_api_status_t status = IB_SUCCESS;
	ib_port_info_t *p_pi = &p->port_info;
	unsigned len;

	if (ib_port_info_get_vl_cap(p_pi) == 1) {	/* no VLArb if 1 data VL */
		p->vl_high_limit = 0;
		return status;
	}
	if (p_pi->vl_arb_low_cap > 0) {
		len = p_pi->vl_arb_low_cap < IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK ?
		    p_pi->vl_arb_low_cap : IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK;
		if ((status = vlarb_update_table_block(sm, p, port_num,
						       force_update,
						       &qcfg->vlarb_low[0],
						       len, 0,
						       mad_list)) != IB_SUCCESS)
			return status;
	}
	if (p_pi->vl_arb_low_cap > IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK) {
		len = p_pi->vl_arb_low_cap % IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK;
		if ((status = vlarb_update_table_block(sm, p, port_num,
						       force_update,
						       &qcfg->vlarb_low[1],
						       len, 1,
						       mad_list)) != IB_SUCCESS)
			return status;
	}
	if (p_pi->vl_arb_high_cap > 0) {
		len = p_pi->vl_arb_high_cap < IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK ?
		    p_pi->vl_arb_high_cap : IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK;
		if ((status = vlarb_update_table_block(sm, p, port_num,
						       force_update,
						       &qcfg->vlarb_high[0],
						       len, 2,
						       mad_list)) != IB_SUCCESS)
			return status;
	}
	if (p_pi->vl_arb_high_cap > IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK) {
		len = p_pi->vl_arb_high_cap % IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK;
		if ((status = vlarb_update_table_block(sm, p, port_num,
						       force_update,
						       &qcfg->vlarb_high[1],
						       len, 3,
						       mad_list)) != IB_SUCCESS)
			return status;
	}

	return status;
}


static ib_slvl_table_t sl2vl_get_table(osm_physp_t * p,
				       const ib_slvl_table_t * sl2vl_table)
{
	ib_slvl_table_t tbl;
	unsigned vl_mask;
	uint8_t vl1, vl2;
	int i;

	vl_mask = (1 << (ib_port_info_get_op_vls(&p->port_info) - 1)) - 1;

	for (i = 0; i < IB_MAX_NUM_VLS / 2; i++) {
		vl1 = sl2vl_table->raw_vl_by_sl[i] >> 4;
		vl2 = sl2vl_table->raw_vl_by_sl[i] & 0xf;
		if (vl1 != 15)
			vl1 &= vl_mask;
		if (vl2 != 15)
			vl2 &= vl_mask;
		tbl.raw_vl_by_sl[i] = (vl1 << 4) | vl2;
	}

	return tbl;
}

static ib_api_status_t sl2vl_update_table(osm_sm_t * sm, osm_physp_t * p,
					  uint8_t in_port, uint32_t attr_mod,
					  unsigned force_update,
					  const ib_slvl_table_t * sl2vl_table,
					  cl_qlist_t *mad_list)
{
	ib_slvl_table_t tbl, *p_tbl;
	osm_madw_t *p_mad;

	tbl = sl2vl_get_table(p, sl2vl_table);
	p_tbl = osm_physp_get_slvl_tbl(p, in_port);

	if (!force_update && !memcmp(p_tbl, &tbl, sizeof(tbl)))
		return IB_SUCCESS;

	p_mad = osm_qos_mad_create(sm, p, sizeof(tbl), (uint8_t *) & tbl,
				   IB_MAD_ATTR_SLVL_TABLE, attr_mod);
	if (!p_mad)
		return IB_INSUFFICIENT_MEMORY;

	/*
	 * Zero the stored SL2VL block, so in case the MAD will
	 * end up with error, we will resend it in the next sweep.
	 */
	memset(p_tbl, 0, sizeof(tbl));

	cl_qlist_insert_tail(mad_list, &p_mad->list_item);
	return IB_SUCCESS;
}

static boolean_t is_port_no_discover(osm_sm_t * sm, osm_node_t *p_node,
				     uint32_t port_num)
{
	return ((p_node->physp_ignored[port_num] != OSM_PHYSP_NORMAL) &&
		lookup_no_discover_port(sm, p_node, port_num));
}

static int sl2vl_port_mask_compare(IN const void *const p1,
				   IN const void *const p2)
{
	return memcmp(p1, p2, sizeof(sl2vl_port_mask_t));
}

static void set_port_mask(uint8_t port_mask[], unsigned port_num, unsigned size)
{
	port_mask[size - 1 - (port_num / 8)] |= 1 << (port_num % 8);
}

static int is_port_mask_zero(uint8_t * port_mask, uint32_t size)
{
	int port_idx = 0;
	for (; port_idx < size; port_idx++) {
		if (port_mask[port_idx] != 0)
			return 0;
	}
	return 1;
}

static ib_api_status_t update_optimazed_sl2vl_mad_list(osm_sm_t * sm, osm_physp_t * p,
						       cl_fmap_t * switch_sl2vl_map,
						       cl_qlist_t *mad_list)
{
	cl_fmap_item_t *p_fitem;
	struct switch_sl2vl_map_item * switch_sl2vl_item;
	ib_sl2vl_mapping_table_t sl2vl_mapping_table;
	uint32_t attr_mod;
	int upper_in_ports = 0, upper_out_ports = 0;
	osm_madw_t *p_mad;

	p_fitem = cl_fmap_head(switch_sl2vl_map);
	while (p_fitem != cl_fmap_end(switch_sl2vl_map)) {
		switch_sl2vl_item = (switch_sl2vl_map_item_t *) p_fitem;
		p_fitem = cl_fmap_next(p_fitem);
		memset(&sl2vl_mapping_table, 0, sizeof(sl2vl_mapping_table));
		memcpy(sl2vl_mapping_table.raw_vl_by_sl,
		       switch_sl2vl_item->in_sl2vl_port_mask.sl2vl_configuration.raw_vl_by_sl,
		       sizeof(sl2vl_mapping_table.raw_vl_by_sl));

		for (upper_in_ports = 0; upper_in_ports < 2; upper_in_ports++) {
			for (upper_out_ports = 0; upper_out_ports < 2; upper_out_ports++) {
				memset(sl2vl_mapping_table.input_port_mask, 0,
				       sizeof(sl2vl_mapping_table.input_port_mask));
				memset(sl2vl_mapping_table.output_port_mask, 0,
				       sizeof(sl2vl_mapping_table.output_port_mask));
				attr_mod = 0x100000;
				attr_mod |= upper_in_ports << 19;
				attr_mod |= upper_out_ports << 18;

				memcpy(sl2vl_mapping_table.input_port_mask,
				       &switch_sl2vl_item->in_sl2vl_port_mask.input_port_mask[(1 - upper_in_ports) * IB_SL2VL_PORT_MASK_BYTES_SIZE],
				       IB_SL2VL_PORT_MASK_BYTES_SIZE);
				memcpy(sl2vl_mapping_table.output_port_mask,
				       &switch_sl2vl_item->output_port_mask[(1 - upper_out_ports) * IB_SL2VL_PORT_MASK_BYTES_SIZE],
				       IB_SL2VL_PORT_MASK_BYTES_SIZE);

				if (is_port_mask_zero(sl2vl_mapping_table.input_port_mask, IB_SL2VL_PORT_MASK_BYTES_SIZE) ||
				    is_port_mask_zero(sl2vl_mapping_table.output_port_mask, IB_SL2VL_PORT_MASK_BYTES_SIZE))
					continue;

				p_mad = osm_qos_mad_create(sm, p, sizeof(sl2vl_mapping_table),
							   (uint8_t *) &sl2vl_mapping_table,
							   IB_MAD_ATTR_SLVL_TABLE, attr_mod);
				if (!p_mad)
					return IB_INSUFFICIENT_MEMORY;

				cl_qlist_insert_tail(mad_list, &p_mad->list_item);
			}
		}
	}

	return IB_SUCCESS;
}

static int qos_extports_setup(osm_sm_t * sm, osm_node_t *node,
			      const struct qos_config *qcfg)

{
	osm_physp_t *p0, *p;
	osm_port_t *p_port;
	unsigned force_update;
	unsigned num_ports = osm_node_get_num_physp(node);
	struct osm_routing_engine *re = sm->p_subn->routing_engine_used;
	int ret = 0;
	unsigned in, out;
	uint8_t op_vl, common_op_vl;
	ib_slvl_table_t tbl, *p_tbl;
	const ib_slvl_table_t *port_sl2vl;
	ib_slvl_table_t routing_sl2vl;
	uint8_t in_port_mask[IB_PORT_MASK_BYTES_SIZE];
	struct sl2vl_port_mask in_sl2vl_port_mask_table[IB_ENDP_MAX_NUMBER_PORTS];
	uint32_t in_sl2vl_port_mask_table_valid[IB_ENDP_MAX_NUMBER_PORTS/32];
	uint8_t out_port_mask_table[IB_PORT_MASK_BYTES_SIZE][IB_ENDP_MAX_NUMBER_PORTS];
	cl_fmap_t switch_sl2vl_map; /* map of (sl2vl,in_port_mask) -> out_port_mask */
	cl_fmap_item_t *p_fitem;
	struct switch_sl2vl_map_item * switch_sl2vl_item;
	unsigned  port_idx, in_valid_idx, vld_mask = 0;

	/*
	 * Do nothing unless the most recent routing attempt was successful.
	 */
	if (!re)
		return ret;

	for (out = 1; out < num_ports; out++) {
		p = osm_node_get_physp_ptr(node, out);
		if (PF(!p))
			continue;
		if (ib_port_info_get_port_state(&p->port_info) == IB_LINK_DOWN)
			continue;
		if (is_port_no_discover(sm, node, out))
			continue;
		force_update = p->need_update || sm->p_subn->need_update;
		p->vl_high_limit = qcfg->vl_high_limit;
		p_port = osm_get_port_by_guid(sm->p_subn, p->port_guid);
		if (vlarb_update(sm, p, p->port_num, force_update, qcfg,
				 &p_port->mads_queue))
			ret = -1;
	}

	p0 = osm_node_get_physp_ptr(node, 0);
	if (!(p0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP))
		return ret;

	/*
	 * First optional optimization - (wild-card based setting):
	 * In case the device supports OptimizedSLtoVLMappingProgramming bit 0,
	 * and routing configuration is uniform (all ports share same configurations,
	 * we will send only one MAD per outport
	 */
	if ((ib_switch_info_get_opt_sl2vlmapping_programming(&node->sw->switch_info) & 0x01) &&
	    sm->p_subn->opt.use_optimized_slvl && !re->update_sl2vl) {
		/* we should find the op_vl that is used by all ports (when exist) */
		p0 = NULL;
		common_op_vl = 0; /* 0 is invalid value, meaning this variable was not set */

		out = ib_switch_info_is_enhanced_port0(&node->sw->switch_info) ? 0 : 1;
		for (; out < num_ports; out++) {
			p = osm_node_get_physp_ptr(node, out);
			if (PF(!p))
				continue;
			if (ib_port_info_get_port_state(&p->port_info) ==
			    IB_LINK_DOWN)
				continue;
			if (is_port_no_discover(sm, node, out))
				continue;
			op_vl = ib_port_info_get_op_vls(&p->port_info);
			if (common_op_vl == 0)
				common_op_vl = op_vl;
			else if (op_vl != common_op_vl)
				goto port_mask_optimization;
			p0 = p;
		}
		if (!p0)
			return -1;
		force_update = node->sw->need_update || sm->p_subn->need_update;
		p_port = osm_get_port_by_guid(sm->p_subn, p0->port_guid);
		return (sl2vl_update_table(sm, p0, p0->port_num, 0x30000, force_update,
					   &qcfg->sl2vl, &p_port->mads_queue));
	}

port_mask_optimization:
	/*
	 * Second optional optimization - (port-mask based setting):
	 * In case the device supports OptimizedSLtoVLMappingProgramming bit 1,
	 * we will send maximum 4 MADs per sl2vl configuration
	 */
	if ((ib_switch_info_get_opt_sl2vlmapping_programming(&node->sw->switch_info) & 0x02) &&
	     sm->p_subn->opt.use_optimized_port_mask_slvl) {
		cl_fmap_init(&switch_sl2vl_map, sl2vl_port_mask_compare);
		memset(out_port_mask_table, 0, sizeof(out_port_mask_table));

		out = ib_switch_info_is_enhanced_port0(&node->sw->switch_info) ? 0 : 1;
		for (; out < num_ports; out++) {
			p = osm_node_get_physp_ptr(node, out);
			if (PF(!p))
				continue;
			if (ib_port_info_get_port_state(&p->port_info) == IB_LINK_DOWN)
				continue;
			if (is_port_no_discover(sm, node, out))
				continue;

			force_update = p->need_update || sm->p_subn->need_update;
			/* hold array of sl2vl_port_mask per port, to avoid malloc (harm parallel threads performance) */
			memset(in_sl2vl_port_mask_table, 0, sizeof(in_sl2vl_port_mask_table));
			memset(in_sl2vl_port_mask_table_valid, 0, sizeof(in_sl2vl_port_mask_table_valid));

			/*
			 * go over all in ports, 
 			 * the index in in_sl2vl_port_mask_table is in port value
 			 * sl2vl configuration is the out-in sl2vl configuration
 			 * port mask is bit mask of in port (later will be collapsed with other masks)
 			*/
			for (in = 0; in < num_ports; in++) {
				port_sl2vl = &qcfg->sl2vl;
				if (re->update_sl2vl) {
					routing_sl2vl = *port_sl2vl;
					re->update_sl2vl(re->context,
							 p, in, out, &routing_sl2vl);
					port_sl2vl = &routing_sl2vl;
				}
				tbl = sl2vl_get_table(p, port_sl2vl);
				p_tbl = osm_physp_get_slvl_tbl(p, in);

				if (!force_update && !memcmp(p_tbl, &tbl, sizeof(tbl))) {
					/* no change is needed */
					continue;
				}

				/*
				 * Zero the stored SL2VL block, so in case the MAD will
				 * end up with error, we will resend it in the next sweep.
				 */
				memset(p_tbl, 0, sizeof(tbl));

				memset(in_port_mask, 0, sizeof(in_port_mask));
				set_port_mask(in_port_mask, in, IB_PORT_MASK_BYTES_SIZE);
				in_sl2vl_port_mask_table_valid[in / 32] |= 1 << (in % 32);
				in_sl2vl_port_mask_table[in].sl2vl_configuration = tbl;
				memcpy(in_sl2vl_port_mask_table[in].input_port_mask, in_port_mask,
				       IB_PORT_MASK_BYTES_SIZE*sizeof(uint8_t));
			}

			/*
			 * go over in_sl2vl_port_mask_table, and check for valid entry
 			 * in case of valid entry, go over all in ports and build mask that holds all ports that holds same configuration
 			 * if port was added to mask, mark its entry as invalid (to prevent duplicating configurations)
			 */
			for (in_valid_idx = 0; in_valid_idx < IB_ENDP_MAX_NUMBER_PORTS; in_valid_idx++) {
				vld_mask = 1 << (in_valid_idx % 32);
				if (!(in_sl2vl_port_mask_table_valid[in_valid_idx / 32] & vld_mask))
					continue;

				tbl = in_sl2vl_port_mask_table[in_valid_idx].sl2vl_configuration;
				memcpy(in_port_mask, in_sl2vl_port_mask_table[in_valid_idx].input_port_mask, sizeof(in_port_mask));

				/* amount of different configurations is expected to be very low (2-4 configurations)*/
				for (port_idx = 0; port_idx < IB_ENDP_MAX_NUMBER_PORTS; port_idx++) {
					vld_mask = 1 << (port_idx % 32);
					if ((in_sl2vl_port_mask_table_valid[port_idx / 32] & vld_mask) &&
					     !memcmp(&in_sl2vl_port_mask_table[port_idx].sl2vl_configuration, &tbl, sizeof(tbl))) {
						set_port_mask(in_sl2vl_port_mask_table[in_valid_idx].input_port_mask,
							      port_idx, IB_PORT_MASK_BYTES_SIZE);
						if (in_valid_idx != port_idx) {
							vld_mask = 1 << (port_idx % 32);
							in_sl2vl_port_mask_table_valid[port_idx / 32] &= ~vld_mask;
						}
					}
				}

				p_fitem = cl_fmap_get(&switch_sl2vl_map, &in_sl2vl_port_mask_table[in_valid_idx]);
				if (p_fitem != cl_fmap_end(&switch_sl2vl_map)) {
					/* sl2vl configuration already exist, update the entry */
					switch_sl2vl_item = (switch_sl2vl_map_item_t *) p_fitem;
					set_port_mask(switch_sl2vl_item->output_port_mask, out, IB_PORT_MASK_BYTES_SIZE);
				} else {
					/* sl2vl configuration doesn't exist, add the entry */
					switch_sl2vl_item = (switch_sl2vl_map_item_t *) malloc(sizeof(*switch_sl2vl_item));
					if (!switch_sl2vl_item) {
						ret = -1;
						goto Exit;
					}
					memset(switch_sl2vl_item, 0, sizeof(*switch_sl2vl_item));
					set_port_mask(switch_sl2vl_item->output_port_mask, out, IB_PORT_MASK_BYTES_SIZE);
					switch_sl2vl_item->in_sl2vl_port_mask = in_sl2vl_port_mask_table[in_valid_idx];
					cl_fmap_insert(&switch_sl2vl_map,
						       &switch_sl2vl_item->in_sl2vl_port_mask,
						       &switch_sl2vl_item->map_item);
				}
			}
		}

		p_port = osm_get_port_by_guid(sm->p_subn, p0->port_guid);
		if (update_optimazed_sl2vl_mad_list(sm, p0, &switch_sl2vl_map,
						    &p_port->mads_queue))
			ret = -1;

Exit:
		p_fitem = cl_fmap_head(&switch_sl2vl_map);
		while (p_fitem != cl_fmap_end(&switch_sl2vl_map)) {
			switch_sl2vl_item = (switch_sl2vl_map_item_t *) p_fitem;
			p_fitem = cl_fmap_next(p_fitem);
			cl_fmap_remove_item(&switch_sl2vl_map,
					    &switch_sl2vl_item->map_item);
			free(switch_sl2vl_item);
		}

		return ret;
	}

	/* non optimized sl2vl configuration */
	out = ib_switch_info_is_enhanced_port0(&node->sw->switch_info) ? 0 : 1;
	for (; out < num_ports; out++) {
		p = osm_node_get_physp_ptr(node, out);
		if (PF(!p))
			continue;
		if (ib_port_info_get_port_state(&p->port_info) == IB_LINK_DOWN)
			continue;
		if (is_port_no_discover(sm, node, out))
			continue;

		force_update = p->need_update || sm->p_subn->need_update;
		p_port = osm_get_port_by_guid(sm->p_subn, p->port_guid);
		/* go over all in ports */
		for (in = 0; in < num_ports; in++) {
			port_sl2vl = &qcfg->sl2vl;

			if (re->update_sl2vl) {
				routing_sl2vl = *port_sl2vl;
				re->update_sl2vl(re->context,
						 p, in, out, &routing_sl2vl);
				port_sl2vl = &routing_sl2vl;
			}
			if (sl2vl_update_table(sm, p, in, in << 8 | out,
					       force_update, port_sl2vl,
					       &p_port->mads_queue))
				ret = -1;
		}
	}

	return ret;
}

static int qos_endport_setup(osm_sm_t * sm, osm_port_t * p_port,
			     const struct qos_config *qcfg, int vlarb_only)
{
	unsigned force_update;
	struct osm_routing_engine *re = sm->p_subn->routing_engine_used;
	const ib_slvl_table_t *port_sl2vl = &qcfg->sl2vl;
	ib_slvl_table_t routing_sl2vl;
	osm_physp_t *p;

	OSM_ASSERT(p_port);
	p = p_port->p_physp;
	force_update = p->need_update || sm->p_subn->need_update;

	p->vl_high_limit = qcfg->vl_high_limit;
	if (vlarb_update(sm, p, 0, force_update, qcfg, &p_port->mads_queue))
		return -1;
	if (vlarb_only)
		return 0;

	if (!(p->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP))
		return 0;

	if (re && re->update_sl2vl) {
		routing_sl2vl = *port_sl2vl;
		re->update_sl2vl(re->context, p, 0, 0, &routing_sl2vl);
		port_sl2vl = &routing_sl2vl;
	}
	if (sl2vl_update_table(sm, p, 0, 0, force_update, port_sl2vl,
			       &p_port->mads_queue))
		return -1;

	return 0;
}

void osm_qos_init(osm_opensm_t * p_osm)
{
	if (!p_osm->subn.opt.qos)
		return;

	OSM_LOG_ENTER(&p_osm->log);

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);

	/* read QoS policy config file */
	osm_qos_parse_policy_file(&p_osm->subn);

	CL_PLOCK_RELEASE(&p_osm->lock);
	OSM_LOG_EXIT(&p_osm->log);
}

static int node_qos_setup(void **task, unsigned task_len,
			  void *context, unsigned thread_id) {
	osm_opensm_t *p_osm = (osm_opensm_t *)context;
	end_port_qos_task_t *node_task;
	//end_port_qos_task_t *node_task = (end_port_qos_task_t *)(*task);
	int result = 0;
	int i;

	/* Each task represents an endport */
	for (i = 0; i < task_len; i++) {
		node_task = (end_port_qos_task_t *)(task[i]);
		if (node_task->p_port->p_node->sw) {
			if (qos_extports_setup(&p_osm->sm, node_task->p_port->p_node,
					       node_task->ep_qos_config))
				return -1;
			if (ib_switch_info_is_enhanced_port0(&node_task->p_port->p_node->sw->switch_info)) {
				result = result | qos_endport_setup(&p_osm->sm, node_task->p_port,
								    node_task->port0_qos_config,
								    node_task->port0_vlarb_only);
			}
		} else
			result = result | qos_endport_setup(&p_osm->sm, node_task->p_port,
							    node_task->ep_qos_config,
							    node_task->ep_vlarb_only);
	}
	return result;
}

static void osm_qos_set_initial_mads_batch(IN struct osm_sm * p_sm,
					   OUT cl_qlist_t * batch_mad_list)
{
	osm_subn_t *p_subn = p_sm->p_subn;
	cl_qmap_t *p_port_guid_tbl = &p_subn->port_guid_tbl;
	osm_port_t *p_port;
	osm_madw_t *p_port_mad;
	int i;

	for (i = 0; i < p_subn->opt.max_wire_smps_per_device; i++) {
		for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
		     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
		     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
			if (!cl_qlist_count(&p_port->mads_queue))
				continue;
			p_port_mad = (osm_madw_t *)
				     cl_qlist_remove_head(&p_port->mads_queue);
			cl_qlist_insert_tail(batch_mad_list, &p_port_mad->list_item);
		}
	}
}

int osm_qos_setup(osm_opensm_t * p_osm, boolean_t enhanced_qos_sweep_only)
{
	struct qos_config ca_config, sw0_config, swe_config, rtr_config;
	cl_qmap_t *p_tbl;
	cl_map_item_t *p_next;
	osm_port_t *p_port;
	osm_node_t *p_node;
	int ret = 0;
	cl_qlist_t batch_mad_list;
	unsigned num_of_ports, port_idx;
	struct end_port_qos_task *qos_tasks = NULL;

	if (!p_osm->subn.opt.qos)
		return 0;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Starting QoS configuration\n");

	cl_qlist_init(&batch_mad_list);

	if (enhanced_qos_sweep_only) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
                        "performing enhanced qos sweep only.\n");
		/* acquire lock */
		CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);
		goto enhanced_qos;
	}

	qos_build_config(&ca_config, &p_osm->subn.opt.qos_ca_options,
			 &p_osm->subn.opt.qos_options);
	qos_build_config(&sw0_config, &p_osm->subn.opt.qos_sw0_options,
			 &p_osm->subn.opt.qos_options);
	qos_build_config(&swe_config, &p_osm->subn.opt.qos_swe_options,
			 &p_osm->subn.opt.qos_options);
	qos_build_config(&rtr_config, &p_osm->subn.opt.qos_rtr_options,
			 &p_osm->subn.opt.qos_options);

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);

	p_tbl = &p_osm->subn.port_guid_tbl;
	num_of_ports = cl_qmap_count(p_tbl);
	
	qos_tasks = (struct end_port_qos_task *) malloc(sizeof(struct end_port_qos_task) * num_of_ports);
	if (!qos_tasks) {
		CL_PLOCK_RELEASE(&p_osm->lock);
		return -1;
	}
	memset(qos_tasks, 0, sizeof(struct end_port_qos_task) * num_of_ports);

	port_idx = 0;
	p_next = cl_qmap_head(p_tbl);
	while (p_next != cl_qmap_end(p_tbl)) {
		p_port = (osm_port_t *) p_next;
		p_next = cl_qmap_next(p_next);

		qos_tasks[port_idx].p_port = p_port;
		qos_tasks[port_idx].ep_vlarb_only = 0;
		/* Port MAD queue should be empty so no leftover MADs shall be sent here */
		OSM_ASSERT(!cl_qlist_count(&p_port->mads_queue));

		p_node = p_port->p_node;
		if (p_node->sw) {
			qos_tasks[port_idx].ep_qos_config = &swe_config;
			
			/* skip base port 0 */
			if (!ib_switch_info_is_enhanced_port0
			    (&p_node->sw->switch_info))
				goto Continue;

			/* port 0 sl2vl configurations */ 
			qos_tasks[port_idx].port0_vlarb_only = 0;
			qos_tasks[port_idx].port0_qos_config = &sw0_config;

			if (ib_switch_info_get_opt_sl2vlmapping(&p_node->sw->switch_info) &&
			    p_osm->sm.p_subn->opt.use_optimized_slvl &&
			    !memcmp(&swe_config.sl2vl, &sw0_config.sl2vl,
				    sizeof(swe_config.sl2vl)))
				qos_tasks[port_idx].port0_vlarb_only = 1;

		} else if (osm_node_get_type(p_node) == IB_NODE_TYPE_ROUTER) {
			qos_tasks[port_idx].ep_qos_config = &rtr_config;

		} else {
			qos_tasks[port_idx].ep_qos_config = &ca_config;

		}

Continue:
		port_idx ++;

	}

	if (cl_mp_thread_pool_xarray_apply(&p_osm->sm.ucast_mgr.thread_pool,
					   qos_tasks, num_of_ports,
					   sizeof(struct end_port_qos_task),
					   CL_MP_THREAD_POOL_BATCH_NUM(4), /* only 4 threads for now */
					   node_qos_setup,
					   p_osm)) {
		CL_PLOCK_RELEASE(&p_osm->lock);
		ret = -1;
		goto Exit;
	 }

	/*
	 * Insert max_wire_smps_per_device MADs from each port MAD queue, to batch_mad_list.
	 * On each MAD response, the next MAD on port MAD queue would be sent.
	 */
	osm_qos_set_initial_mads_batch(&p_osm->sm, &batch_mad_list);

	CL_PLOCK_RELEASE(&p_osm->lock);
	osm_send_req_mad_batch(&p_osm->sm, &batch_mad_list);
	if (wait_for_pending_transactions(p_osm)) {
		ret = -1;
		goto Exit;
	}

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);
enhanced_qos:
	/* enhanced qos part */
	osm_enhanced_qos_mgr_process(&p_osm->sm, &batch_mad_list);

	CL_PLOCK_RELEASE(&p_osm->lock);

	osm_send_req_mad_batch(&p_osm->sm, &batch_mad_list);
Exit:
	free(qos_tasks);
	osm_subn_clear_ports_mads_queue(&p_osm->subn);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"QoS configuration completed\n");

	OSM_LOG_EXIT(&p_osm->log);

	return ret;
}

/*
 *  QoS config stuff
 */
static int parse_one_unsigned(const char *str, char delim, unsigned *val)
{
	char *end;
	*val = strtoul(str, &end, 0);
	if (*end)
		end++;
	return (int)(end - str);
}

static int parse_vlarb_entry(const char *str, ib_vl_arb_element_t * e)
{
	unsigned val;
	const char *p = str;
	p += parse_one_unsigned(p, ':', &val);
	e->vl = val % 15;
	p += parse_one_unsigned(p, ',', &val);
	e->weight = (uint8_t) val;
	return (int)(p - str);
}

static int parse_sl2vl_entry(const char *str, uint8_t * raw)
{
	unsigned val1, val2;
	const char *p = str;
	p += parse_one_unsigned(p, ',', &val1);
	p += parse_one_unsigned(p, ',', &val2);
	*raw = (val1 << 4) | (val2 & 0xf);
	return (int)(p - str);
}

static void qos_build_config(struct qos_config *cfg, osm_qos_options_t * opt,
			     osm_qos_options_t * dflt)
{
	int i;
	const char *p;

	memset(cfg, 0, sizeof(*cfg));

	if (opt->max_vls > 0)
		cfg->max_vls = opt->max_vls;
	else {
		if (dflt->max_vls > 0)
			cfg->max_vls = dflt->max_vls;
		else
			cfg->max_vls = OSM_DEFAULT_QOS_MAX_VLS;
	}

	if (opt->high_limit >= 0)
		cfg->vl_high_limit = (uint8_t) opt->high_limit;
	else {
		if (dflt->high_limit >= 0)
			cfg->vl_high_limit = (uint8_t) dflt->high_limit;
		else
			cfg->vl_high_limit = (uint8_t) OSM_DEFAULT_QOS_HIGH_LIMIT;
	}

	if (opt->vlarb_high)
		p = opt->vlarb_high;
	else {
		if (dflt->vlarb_high)
			p = dflt->vlarb_high;
		else
			p = OSM_DEFAULT_QOS_VLARB_HIGH;
	}
	for (i = 0; i < 2 * IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK; i++) {
		p += parse_vlarb_entry(p,
				       &cfg->vlarb_high[i /
							IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK].
				       vl_entry[i %
						IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK]);
	}

	if (opt->vlarb_low)
		p = opt->vlarb_low;
	else {
		if (dflt->vlarb_low)
			p = dflt->vlarb_low;
		else
			p = OSM_DEFAULT_QOS_VLARB_LOW;
	}
	for (i = 0; i < 2 * IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK; i++) {
		p += parse_vlarb_entry(p,
				       &cfg->vlarb_low[i /
						       IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK].
				       vl_entry[i %
						IB_NUM_VL_ARB_ELEMENTS_IN_BLOCK]);
	}

	p = opt->sl2vl ? opt->sl2vl : dflt->sl2vl;
	if (opt->sl2vl)
		p = opt->sl2vl;
	else {
		if (dflt->sl2vl)
			p = dflt->sl2vl;
		else
			p = OSM_DEFAULT_QOS_SL2VL;
	}
	for (i = 0; i < IB_MAX_NUM_VLS / 2; i++)
		p += parse_sl2vl_entry(p, &cfg->sl2vl.raw_vl_by_sl[i]);
}

static void qos_config_vl_send_profiles_config(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *) cxt;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;
	uint8_t block_num, num_ports, num_profiles_blocks;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_QOS_CONFIG_VL_SUPPORTED_BIT) ||
	    !ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_VL_BUFFER_WEIGHT_SUPPORTED_BIT))
		return;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	/* num_ports count includes port 0 */
	num_profiles_blocks = num_ports < IB_PROFILE_CONFIG_NUM_PROFILES_PER_BLOCK ? 1 : 2;

	for (block_num = 0; block_num < num_profiles_blocks; block_num++) {
		osm_subn_send_profiles_config_block(p_sm, p_sw, IB_PROFILES_FEATURE_QOS_VL_CONFIG,
						    block_num);
	}
}

static void send_qos_config_vl_per_profile(osm_sm_t * p_sm, osm_switch_t * p_sw, uint8_t profile)
{
	ib_qos_config_vl_t *p_qos_qonfig_vl_req = &p_sw->qos_config_vl_sw_data.calc_qos_config_vl[profile];
	ib_qos_config_vl_t *p_qos_config_vl = &p_sw->qos_config_vl_sw_data.subn_qos_config_vl[profile];
	osm_physp_t *p_physp0;
	osm_qos_config_vl_context_t context;
	ib_api_status_t status;
	uint32_t attr_mod;

	if (!p_sm->p_subn->need_update && !p_sw->need_update &&
	    !p_sw->qos_config_vl_sw_data.qos_config_vl_need_update &&
	    !memcmp(p_qos_qonfig_vl_req, p_qos_config_vl, sizeof(ib_qos_config_vl_t)))
		return;

	/* Send the MAD only if the profile is set for this switch */
	if (!osm_switch_is_profile_set_on_switch(p_sw, profile, IB_PROFILES_FEATURE_QOS_VL_CONFIG))
		return;

	memset(p_qos_config_vl, 0, sizeof(ib_qos_config_vl_t));

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2601: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	attr_mod = profile | IB_QOS_CONFIG_VL_ATTR_MOD_PROFILE_EN_BIT;

	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_sw->p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp0);
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = profile;
	context.port = p_physp0->port_num;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0), p_physp0,
			     (const uint8_t *) p_qos_qonfig_vl_req,
			     sizeof(ib_qos_config_vl_t),
			     IB_MAD_ATTR_QOS_CONFIG_VL,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2602: "
			"Setting QosConfigVl (profile %u) to "
			"0x%" PRIx64 " failed (%s)\n",
			profile, cl_ntoh64(p_physp0->port_guid),
			ib_get_err_str(status));
	}
}

static void send_qos_config_vl_per_port(osm_sm_t * p_sm, osm_physp_t *p_physp, ib_qos_config_vl_t *p_qos_qonfig_vl_req)
{
	static const uint8_t ZEROARRAY[sizeof(ib_qos_config_vl_t)] = {0};
	osm_qos_config_vl_context_t context;
	ib_api_status_t status;
	uint8_t attr_mod;

	if (!p_sm->p_subn->need_update && !p_physp->need_update &&
	    !memcmp(p_qos_qonfig_vl_req, &p_physp->qos_config_vl, sizeof(ib_qos_config_vl_t)))
		return;

	/* No need to send the MAD, if set to all zeroes */
	if (!memcmp(p_qos_qonfig_vl_req, ZEROARRAY, sizeof(ib_qos_config_vl_t)))
		return;

	memset(&p_physp->qos_config_vl, 0, sizeof(ib_qos_config_vl_t));
	attr_mod = p_physp->port_num;

	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_physp->p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp);
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = 0;
	context.port = p_physp->port_num;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp), p_physp,
			     (const uint8_t *) p_qos_qonfig_vl_req,
			     sizeof(ib_qos_config_vl_t),
			     IB_MAD_ATTR_QOS_CONFIG_VL,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2603: "
			"Setting QosConfigVl to port "
			"0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid), ib_get_err_str(status));
	}
}

static void send_sw_qos_config_vl(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t profile;

	/* Skip switches of which profiles weren't configured yet */
	if (!osm_switch_are_profiles_configured(p_sw, IB_PROFILES_FEATURE_QOS_VL_CONFIG))
		return;

	/* Switch has to support vl_buffer_weights to use QoSConfigVL MAD */
	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info, IB_GENERAL_INFO_IS_QOS_CONFIG_VL_SUPPORTED_BIT) ||
	    !ib_general_info_is_cap_enabled(&p_sw->p_node->general_info, IB_GENERAL_INFO_IS_VL_BUFFER_WEIGHT_SUPPORTED_BIT))
		return;

	/*
	 * Send QoSConfigVL configuration MAD per profile.
	 */
	for (profile = IB_QOS_CONFIG_VL_START_PROFILE; profile <= p_sw->qos_config_vl_sw_data.max_profile; profile++) {
		send_qos_config_vl_per_profile(p_sm, p_sw, profile);
	}

	p_sw->qos_config_vl_sw_data.qos_config_vl_need_update = FALSE;
}

static void send_gpu_qos_config_vl(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_gpu_t *p_gpu = (osm_gpu_t *)p_map_item;
	osm_node_t *p_node = p_gpu->p_node;
	osm_physp_t *p_physp;
	int num_of_ports = osm_node_get_num_physp(p_gpu->p_node);
	int port_num, vl;

	if (!ib_general_info_is_cap_enabled(&p_gpu->p_node->general_info, IB_GENERAL_INFO_IS_QOS_CONFIG_VL_SUPPORTED_BIT))
		return;

	if (!ib_general_info_is_cap_enabled(&p_gpu->p_node->general_info, IB_GENERAL_INFO_IS_VL_BUFFER_WEIGHT_SUPPORTED_BIT)) {
		for (vl = 0; vl < IB_DROP_VL; vl++) {
			if (p_gpu->calc_qos_config_vl.vl_config_block[vl].vl_buffer_weight != 0) {
				OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2605: "
					"Setting QoSConfigVl for GPU: 0x%" PRIx64 " failed, "
					"setting VL buffer weights not supported for GPU's\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)));
				return;
			}
		}
	}

	/*
	 * Send QosConfigVL configuration MAD per port.
	 */
	for (port_num = 1; port_num < num_of_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);

		if (!p_physp)
			continue;

		send_qos_config_vl_per_port(p_sm, p_physp, &p_gpu->calc_qos_config_vl);
	}
}

static void qos_config_vl_update_profiles(osm_subn_t * p_subn)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	cl_map_item_t *p_item;
	int i;

	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;

		for (i = 1; i < p_sw->p_node->physp_tbl_size; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (!p_physp)
				continue;

			p_physp->qos_config_vl_prev_profile = p_physp->qos_config_vl_profile;
			p_physp->qos_config_vl_profile = IB_QOS_CONFIG_VL_INVALID_PROFILE;
		}
	}
}

static void qos_config_vl_build_config(osm_sm_t *sm)
{
	ib_qos_config_vl_t qos_config_vl[IB_PROFILE_CONFIG_NUM_PROFILES];
	ib_qos_config_vl_t *p_qos_config_vl;
	osm_switch_t *p_sw;
	osm_gpu_t *p_gpu;
	osm_physp_t *p_physp;
	cl_map_item_t *p_item;
	boolean_t is_one_non_down_port_exists = FALSE;
	boolean_t is_qos_config_vl_parsed;
	int i;

	memset(qos_config_vl, 0, sizeof(qos_config_vl));

	for (p_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;
		is_one_non_down_port_exists = FALSE;
		is_qos_config_vl_parsed = p_sw->p_node->conf_params.is_qos_config_vl_parsed;

		for (i = 1; i < p_sw->p_node->physp_tbl_size; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (!p_physp)
				continue;

			p_physp->qos_config_vl_prev_profile = p_physp->qos_config_vl_profile;
			if (ib_port_info_get_port_state(&p_physp->port_info) != IB_LINK_DOWN &&
			    is_qos_config_vl_parsed && p_physp->port_reset_required == OSM_PORT_RESET_TYPE_NONE) {
				p_physp->qos_config_vl_profile = p_physp->conf_params.qos_config_vl_profile;
				is_one_non_down_port_exists = TRUE;
			} else {
				p_physp->qos_config_vl_profile = IB_QOS_CONFIG_VL_INVALID_PROFILE;
			}
		}

		if (is_qos_config_vl_parsed && is_one_non_down_port_exists) {
			memcpy(&p_sw->qos_config_vl_sw_data.calc_qos_config_vl, &p_sw->conf_params.qos_config_vl,
			       sizeof(p_sw->qos_config_vl_sw_data.calc_qos_config_vl));

			p_sw->qos_config_vl_sw_data.max_profile = p_sw->conf_params.qos_config_vl_max_profile;
		} else {
			p_sw->qos_config_vl_sw_data.max_profile = IB_QOS_CONFIG_VL_INVALID_PROFILE;
		}
	}

	for (p_item = cl_qmap_head(&sm->p_subn->gpu_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->gpu_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_gpu = (osm_gpu_t *)p_item;

		is_qos_config_vl_parsed = p_gpu->p_node->conf_params.is_qos_config_vl_parsed;
		p_qos_config_vl = is_qos_config_vl_parsed ?
		    &p_gpu->p_node->conf_params.qos_config_vl : &qos_config_vl[IB_QOS_CONFIG_VL_INVALID_PROFILE];

		memcpy(&p_gpu->calc_qos_config_vl, p_qos_config_vl, sizeof(p_gpu->calc_qos_config_vl));
	}
}

void osm_qos_config_vl_calculate(osm_subn_t *p_subn)
{
	if (p_subn->opt.qos_config_vl_enabled)
		qos_config_vl_build_config(&p_subn->p_osm->sm);
	else
		qos_config_vl_update_profiles(p_subn);
}

void osm_qos_config_vl_configure(osm_subn_t *p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (!p_subn->opt.qos_config_vl_enabled)
		goto Exit;

	/* Set port profiles of QosConfigVL feature by ProfileConfig MAD */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, qos_config_vl_send_profiles_config, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send QosConfigVL to switches according to parsed cfg parameters */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, send_sw_qos_config_vl, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send QosConfigVL to gpus according to parsed cfg parameters */
	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, send_gpu_qos_config_vl, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_qos_config_vl_rcv_process(IN void *context, IN void *data)
{
	osm_qos_config_vl_context_t *p_context;
	ib_qos_config_vl_t *p_qos_config_vl;
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_node_t *p_node;
	uint32_t attr_mod;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	attr_mod = cl_ntoh32(p_smp->attr_mod);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_QOS_CONFIG_VL);

	p_context = (osm_qos_config_vl_context_t *)&p_madw->context;
	p_qos_config_vl = ib_smp_get_payload_ptr(p_smp);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_port = osm_get_port_by_guid(sm->p_subn, p_context->port_guid);
	if (PF(!p_port)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 2604: "
			"No port object for port with GUID 0x%" PRIx64
			" for parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(p_context->port_guid),
			cl_ntoh64(p_context->node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	/* process only valid QosConfigVL responses */
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	if (attr_mod & IB_QOS_CONFIG_VL_ATTR_MOD_PROFILE_EN_BIT) {
		/* Currently only switches support PortProfiles MAD */
		OSM_ASSERT(p_node->sw);
		memcpy(&p_node->sw->qos_config_vl_sw_data.subn_qos_config_vl[p_context->profile],
		       p_qos_config_vl, sizeof(ib_qos_config_vl_t));
	} else {
		memcpy(&p_port->p_physp->qos_config_vl, p_qos_config_vl, sizeof(ib_qos_config_vl_t));
	}

Exit:

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
