/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of "auto routing engine".
 *      Routing engine encapsulate other routing engine, and invoke them
 *      according to topology type.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_AUTO_C
#include <opensm/osm_sm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_multicast.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>

extern int osm_ucast_dfp2_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_updn_setup(struct osm_routing_engine *, osm_opensm_t *);

/* auto engine structure */
typedef struct auto_engine {
	struct osm_routing_engine *engines[OSM_TOPOLOGY_TYPE_MAX];
	osm_subn_t *p_subn;
} auto_engine_t;

static
void auto_engine_delete(void *context)
{
	auto_engine_t *auto_engine = (auto_engine_t *)context;
	struct osm_routing_engine *r;
	unsigned int i;

	for (i = 0; i < OSM_TOPOLOGY_TYPE_MAX; i++) {
		r = auto_engine->engines[i];
		if (!r)
			continue;

		if (r->destroy)
			r->destroy(r->context);

		osm_routing_engine_free(r);
		auto_engine->engines[i] = NULL;
	}

	free(auto_engine);
}

static
struct osm_routing_engine_attributes ucast_auto_engine_get_attributes(struct osm_routing_engine *r)
{
	auto_engine_t *auto_engine = (auto_engine_t *)r->context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re) {
		struct osm_routing_engine_attributes result = {};
		return result;
	}

	return re->get_attributes(re);
}

static
const char *ucast_auto_engine_get_name(struct osm_routing_engine *r)
{
	auto_engine_t *auto_engine = (auto_engine_t *)r->context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re)
		return NULL;

	return re->get_name(re);
}

static
int ucast_auto_engine_build_lid_matrices(void *context)
{
	auto_engine_t *auto_engine = (auto_engine_t *)context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	osm_opensm_t *p_osm = p_subn->p_osm;

	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re)
		return -1;

	if (re->build_lid_matrices)
		return re->build_lid_matrices(re->context);


	return osm_ucast_mgr_build_lid_matrices(&p_osm->sm.ucast_mgr);
}

static
int ucast_auto_engine_build_fwd_tables(void *context)
{
	auto_engine_t *auto_engine = (auto_engine_t *)context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	osm_opensm_t *p_osm = p_subn->p_osm;
	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re)
		return -1;

	if (re->ucast_build_fwd_tables)
		return re->ucast_build_fwd_tables(re->context);

	return osm_ucast_mgr_build_lft_tables(&p_osm->sm.ucast_mgr);
}

static
ib_api_status_t ucast_auto_build_stree(void *context, osm_mgrp_box_t * mbox)
{
	auto_engine_t *auto_engine = (auto_engine_t *)context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re)
		return IB_UNKNOWN_ERROR;

	if (re->mcast_build_stree)
		return re->mcast_build_stree(re->context, mbox);
	else if (p_subn->opt.const_multicast)
		return osm_mcast_mgr_build_spanning_tree_const(p_subn, mbox);
	else
		return osm_mcast_mgr_build_spanning_tree(p_subn, mbox);
}

static
void ucast_auto_update_sl2vl(void *context, osm_physp_t *p_physp,
			     uint8_t iport_num, uint8_t oport_num,
			     ib_slvl_table_t *osm_oport_sl2vl)
{
	auto_engine_t *auto_engine = (auto_engine_t *)context;
	osm_subn_t *p_subn = auto_engine->p_subn;
	struct osm_routing_engine *re;

	re = auto_engine->engines[p_subn->detected_topo];
	if (!re)
		return;

	if (re->update_sl2vl)
		re->update_sl2vl(re->context, p_physp, iport_num, oport_num, osm_oport_sl2vl);
}


int osm_ucast_auto_engine_setup(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	auto_engine_t *auto_engine = (auto_engine_t *)malloc(sizeof(auto_engine_t));
	if (!auto_engine)
		return -1;

	memset(auto_engine, 0, sizeof(auto_engine_t));

	auto_engine->engines[OSM_TOPOLOGY_TYPE_TREE] = setup_routing_engine(osm->sm.ucast_mgr.p_subn, "ar_updn");
	if (!auto_engine->engines[OSM_TOPOLOGY_TYPE_TREE])
		goto error;

	auto_engine->engines[OSM_TOPOLOGY_TYPE_DFP] = setup_routing_engine(osm->sm.ucast_mgr.p_subn, "dfp2");
	if (!auto_engine->engines[OSM_TOPOLOGY_TYPE_DFP])
		goto error;

	auto_engine->p_subn = osm->sm.ucast_mgr.p_subn;

	r->context = auto_engine;
	r->destroy = auto_engine_delete;

	r->get_name = ucast_auto_engine_get_name;
	r->get_attributes = ucast_auto_engine_get_attributes;
	r->build_lid_matrices = ucast_auto_engine_build_lid_matrices;
	r->ucast_build_fwd_tables = ucast_auto_engine_build_fwd_tables;
	r->mcast_build_stree = ucast_auto_build_stree;
	r->update_sl2vl = ucast_auto_update_sl2vl;

	return 0;

error:
	auto_engine_delete(auto_engine);

	return -1;
}
