/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Minhop Algorithm.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_list.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_MINHOP_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>

/* ---------------------------------------------------------------------------------------------- */

typedef struct {
	uint16_t	type;
	uint16_t	min_lid;
	uint16_t	max_lid;
} minhop_port_t;

typedef struct {
	uint16_t	group_id;
	boolean_t	is_leaf_switch;
	uint16_t	ca_count;
	uint16_t	max_group_id;
	uint16_t	lid;
	uint8_t		*distances;
	osm_switch_t	*p_sw;
	minhop_port_t	ports[256];
} minhop_node_t;

typedef struct {
	osm_opensm_t	*p_osm;
	osm_subn_t	*p_subn;
	minhop_node_t	*nodes;
	uint8_t		*distances;
	int		num_switches;
	flow_graph_t	*p_flow;
	int		ar_enable;
	int		ar_mode;
} minhop_t;

/* ---------------------------------------------------------------------------------------------- */

#define index_to_node(INDEX)		&p_minhop->nodes[INDEX]
#define index_to_distance(INDEX)	&p_minhop->distances[INDEX * p_minhop->num_switches]

#define foreach_minhop_node(P_MINHOP, P_NODE)				\
	for (P_NODE = P_MINHOP->nodes;					\
	     P_NODE < &P_MINHOP->nodes[P_MINHOP->num_switches];		\
	     P_NODE++)

#define foreach_minhop_link(P_NODE, P_LINK)				\
	for (P_LINK = P_NODE->links;					\
	     P_LINK < &P_NODE->links[P_NODE->num_links];		\
	     P_LINK++)

/* ---------------------------------------------------------------------------------------------- */

static inline void minhop_free(IN minhop_t *p_minhop)
{
	free(p_minhop->nodes);
	free(p_minhop->distances);
	p_minhop->nodes = NULL;
	p_minhop->distances = NULL;
}

/* ---------------------------------------------------------------------------------------------- */

static int minhop_init(IN osm_subn_t *p_subn,
		       IN minhop_t *p_minhop)
{
	int		status;
	flow_vertex_t	*p_vertex;
	minhop_node_t	*p_minhop_node;
	osm_switch_t	*p_sw;
	flow_graph_t	*p_flow;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/*
	 * If the old data structs are around, free them.
	 */
	minhop_free(p_minhop);

	/*
	 * Allocate space for all of the MINHOP BFS passes.
	 */
	p_minhop->num_switches = cl_qmap_count(&p_subn->sw_guid_tbl);

	p_minhop->nodes = (minhop_node_t *)calloc(p_minhop->num_switches, sizeof(minhop_node_t));
	p_minhop->distances = (uint8_t *)calloc(p_minhop->num_switches, p_minhop->num_switches * sizeof(uint8_t));
	if ((p_minhop->nodes == NULL) || (p_minhop->distances == NULL)) {
		OSM_LOG(&p_minhop->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA01: allocation failed for nodes and distances\n");
		status = -1;
		goto error_exit;
	}

	/*
	 * Create the graph for the algorithm to work on.
	 */
	p_flow = osm_flow_build_graph(p_subn);

	/*
	 * Label the switches with their indexes.
	 */
	foreach_vertex(p_flow, p_vertex) {
		p_sw = p_vertex->p_sw;
		p_sw->re_index = p_vertex->index;
		p_sw->re_priv = p_minhop;

		p_minhop_node = index_to_node(p_vertex->index);
		p_minhop_node->p_sw = p_sw;
		p_minhop_node->distances = index_to_distance(p_vertex->index);
		p_minhop_node->lid = osm_node_get_base_lid_ho(p_sw->p_node, 0);
		p_minhop_node->is_leaf_switch = FALSE;
		p_minhop_node->ca_count = 0;
	}

	/*
	 * Destroy the graph.
	 */
	osm_flow_destroy_graph(&p_flow);

	status = 0;

error_exit:
	if (status < 0) {
		minhop_free(p_minhop);
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static void minhop_destroy(IN void *context)
{
	minhop_t	*p_minhop = (minhop_t *)context;
	osm_subn_t	*p_subn = p_minhop->p_subn;
	osm_switch_t	*p_sw = NULL;

	/*
	 * Remove the minhop pointer from the switches.
	 */
	foreach_switch(p_subn, p_sw) {
		p_sw->re_index = 0;
		p_sw->re_priv = NULL;
	}

	minhop_free(p_minhop);
	free(p_minhop);
}

static void minhop_execute_threaded(IN void *function, IN void *context)
{
	minhop_t	*p_minhop;
	osm_subn_t	*p_subn;
	osm_switch_t	*p_sw;
	osm_ucast_mgr_t	*p_mgr;
	void		**tasks;

	p_minhop = (minhop_t *)context;
	p_subn = p_minhop->p_subn;
	p_mgr = &p_subn->p_osm->sm.ucast_mgr;

	/*
	 * Allocate space for task pointers.
	 */
	if ((tasks = (void **)calloc(p_minhop->num_switches, sizeof(void *))) == NULL) {
		OSM_LOG(&p_minhop->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA03: allocation failed for tasks\n");
		return;
	}

	/*
	 * Put the switches into the task list.
	 */
	foreach_switch(p_subn, p_sw) {
		tasks[p_sw->re_index] = (void *)p_sw;
	}

	/*
	 * Execute the function.
	 */
	cl_mp_thread_pool_array_apply(&p_mgr->thread_pool,
				      tasks,
				      p_minhop->num_switches,
				      CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				      (cl_pfn_mp_thread_pool_apply_t)function,
				      context);

	/*
	 * Cleanup.
	 */
	free(tasks);
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_bfs(IN minhop_t *p_minhop,
		       IN int index,
		       IN uint8_t *visited,
		       IN uint16_t *queue)
{
	int		i;
	int		v;
	int		l_adj;
	int		q_max;
	int		hop_count;
	uint8_t		*distances;
	flow_vertex_t	*p_vertex;
	flow_edge_t	*p_edge;

	distances = p_minhop->nodes[index].distances;

	distances[index] = 0;
	visited[index] = 1;

	q_max = 0;
	queue[q_max++] = index;

	for (i = 0; i < q_max; i++) {
		v = queue[i];
		hop_count = distances[v] + 1;
		p_vertex = &p_minhop->p_flow->vertices[v];

		foreach_vertex_edge(p_minhop->p_flow, p_vertex, p_edge) {
			l_adj = p_edge->adj_index;

			if (visited[l_adj] == 0) {
				visited[l_adj] = 1;
				distances[l_adj] = hop_count;
				queue[q_max++] = l_adj;
			}
		}
	}
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_resolve_node(IN minhop_t *p_minhop,
				IN osm_switch_t *p_sw,
				IN uint8_t *visited,
				IN uint16_t *queue)
{
	memset((void *)visited, 0, p_minhop->num_switches * sizeof(uint8_t));
	memset((void *)queue, 0, p_minhop->num_switches * sizeof(uint16_t));

	/*
	 * Resolve the node.
	 */
	minhop_bfs(p_minhop, p_sw->re_index, visited, queue);
}

static int minhop_resolve_nodes(IN void **tasks,
				IN unsigned task_len,
				IN void *context,
				IN unsigned int thread_id)
{
	int		i;
	int		status = 0;
	uint8_t		*visited;
	uint16_t	*queue;
	minhop_t	*p_minhop;

	p_minhop = (minhop_t *)context;
	OSM_LOG_ENTER(&p_minhop->p_subn->p_osm->log);

	visited = malloc(p_minhop->num_switches * sizeof(uint8_t));
	queue   = malloc(p_minhop->num_switches * sizeof(uint16_t));

	if ((visited != NULL) && (queue != NULL)) {
		for (i = 0; i < task_len; i++) {
			minhop_resolve_node(p_minhop, tasks[i], visited, queue);
		}
	} else {
		OSM_LOG(&p_minhop->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA02: allocation failed for visited and queue\n");
		status = 1;
	}

	free(visited);
	free(queue);

	OSM_LOG_EXIT(&p_minhop->p_subn->p_osm->log);
	return status;
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_set_hop_counts(IN minhop_t *p_minhop)
{
	int		v;
	uint8_t		port;
	uint16_t	lid;
	osm_switch_t	*p_sw;
	uint8_t		*p_distance;
	minhop_node_t	*p_minhop_node;
	minhop_node_t	*p_minhop_remote_node;
	minhop_node_t	*p_minhop_neighbor_node;
	flow_edge_t	*p_edge;
	flow_vertex_t	*p_vertex;

	/*
	 * Clear the hop counts.
	 */
	foreach_switch(p_minhop->p_subn, p_sw) {
		osm_switch_clear_hops(p_sw);
	}

	foreach_minhop_node(p_minhop, p_minhop_node) {
		p_sw = p_minhop_node->p_sw;

		p_vertex = &p_minhop->p_flow->vertices[p_sw->re_index];
		foreach_vertex_edge(p_minhop->p_flow, p_vertex, p_edge) {
			port = p_edge->port_num;
			p_minhop_neighbor_node = index_to_node(p_edge->adj_index);

			p_minhop_remote_node = p_minhop->nodes;
			p_distance = p_minhop_neighbor_node->distances;
			for (v = 0; v < p_minhop->num_switches; v++) {
				osm_switch_set_hops(p_sw, p_minhop_remote_node->lid, port, *p_distance + 1);
				p_distance++;
				p_minhop_remote_node++;
			}
		}
	}

	/*
	 * Clear the switch hop count to itself.
	 */
	foreach_switch(p_minhop->p_subn, p_sw) {
		lid = osm_node_get_base_lid(p_sw->p_node, 0);
		lid = cl_ntoh16(lid);
		osm_switch_set_hops(p_sw, lid, 0, 0);
	}
}

static int minhop_lid_matrices(IN void *context)
{
	minhop_t	*p_minhop;
	osm_subn_t	*p_subn;

	p_minhop = (minhop_t *)context;
	p_subn = p_minhop->p_subn;
	p_minhop->p_flow = osm_flow_build_graph(p_subn);

	/*
	 * Run BFS on the switches.
	 */
	minhop_execute_threaded(minhop_resolve_nodes, context);

	/*
	 * Set all of the port->hop_counts.
	 */
	minhop_set_hop_counts(p_minhop);

	osm_flow_destroy_graph(&p_minhop->p_flow);

	return 0;
}

static int minhop_ar_lid_matrices(IN void *context)
{
	minhop_t	*p_minhop = (minhop_t *)context;
	osm_subn_t	*p_subn = p_minhop->p_subn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (minhop_init(p_subn, p_minhop) < 0) {
		return -1;
	}

	/*
	 * Run the BFS for the graph.
	 */
	minhop_lid_matrices(context);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return 0;
}

static int minhop_ar_fwd_tables(IN void *context)
{
	minhop_t	*p_minhop = (minhop_t *)context;
	osm_subn_t	*p_subn = p_minhop->p_subn;

	return osm_ucast_mgr_build_lft_tables(&p_subn->p_osm->sm.ucast_mgr);
}

static ib_api_status_t minhop_mcast_stree(IN void *context,
					  IN OUT struct osm_mgrp_box *mgb)
{
	minhop_t	*p_minhop = (minhop_t *)context;

	return osm_mcast_mgr_build_spanning_tree(p_minhop->p_subn, mgb);
}

/* ---------------------------------------------------------------------------------------------- */
int osm_ucast_ar_minhop_setup(IN struct osm_routing_engine *r, IN osm_opensm_t *p_osm)
{
	minhop_t	*p_minhop;

	OSM_LOG_ENTER(&p_osm->log);

	if ((p_minhop = (minhop_t *)calloc(1, sizeof(minhop_t))) == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR AA04: allocation failed for minhop\n");
		return -1;
	}

	p_minhop->p_osm = p_osm;
	p_minhop->p_subn = p_osm->sm.ucast_mgr.p_subn;

	r->context = p_minhop;
	r->destroy = minhop_destroy;
	r->build_lid_matrices = minhop_ar_lid_matrices;
	r->ucast_build_fwd_tables = minhop_ar_fwd_tables;
	r->mcast_build_stree = minhop_mcast_stree;

	r->attributes.ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->attributes.shield_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->attributes.pfrn_mode = TRUE;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_AR_MINHOP;
	r->calculate_flids = NULL;

	OSM_LOG_EXIT(&p_osm->log);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static uint16_t minhop_set_group_id(IN osm_subn_t *p_subn, IN uint16_t lid)
{
	return osm_ucast_mgr_make_groupid(&p_subn->p_osm->sm.ucast_mgr, lid, 0);
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_set_new_lft_entry(IN osm_subn_t *p_subn,
				     IN osm_switch_t *p_sw,
				     IN uint16_t lid,
				     IN int state,
				     IN int id)
{
	uint8_t		lft_port;

	lft_port = osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT);
	if (lft_port != OSM_NO_PATH) {
		osm_ar_lft_set(&p_sw->ar_data, lid, 0, lft_port, state, id);
	} else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE, "%s - no port found for lid %d\n", p_sw->p_node->print_desc, lid);
	}
}

static void minhop_set_new_static_lft_entry(IN osm_subn_t *p_subn,
					    IN osm_switch_t *p_sw,
					    IN uint16_t lid)
{
	minhop_set_new_lft_entry(p_subn, p_sw, lid, OSM_AR_IB_LID_STATE_STATIC, 0);
}

static void minhop_set_new_free_lft_entry(IN osm_subn_t *p_subn,
					  IN osm_switch_t *p_sw,
					  IN uint16_t lid,
					  IN uint16_t group_id,
					  IN osm_ar_subgroup_t *ar_group)
{
	osm_ar_subgroup_t		*p_subgroup;
	osm_advanced_routing_data_t	*p_ar_data;

	/*
	 * Set the subgroup with the remote switch's AR group.
	 */
	p_ar_data = &p_sw->ar_data;
	p_subgroup = osm_ar_subgroup_get(p_ar_data, group_id, 0);
	memcpy(p_subgroup, ar_group, sizeof(osm_ar_subgroup_t));
	minhop_set_new_lft_entry(p_subn, p_sw, lid, OSM_AR_IB_LID_STATE_FREE, group_id);
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_ar_switch(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{
	int				i;
	int				port_type;
	uint8_t				ar_state;
	uint16_t			lid;
	uint16_t			min_lid;
	uint16_t			max_lid;
	uint16_t			remote_lid;
	uint16_t			group_id;
	uint16_t			group_cap;
	uint16_t			max_group_id;
	boolean_t			group_is_static;
	minhop_t			*p_minhop;
	minhop_node_t			*p_minhop_node;
	minhop_node_t			*p_minhop_remote_node;

	osm_node_t			*p_remote_node;
	osm_switch_t			*p_remote_sw;
	osm_ar_subgroup_t		remote_ar_group;
	osm_advanced_routing_data_t	*p_ar_data;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_minhop = (minhop_t *)p_sw->re_priv;
	p_minhop_node = index_to_node(p_sw->re_index);
	ar_state = p_minhop->ar_enable ? p_minhop->ar_mode : OSM_AR_IB_LID_STATE_STATIC;

	p_ar_data = &p_sw->ar_data;

	group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);

	/*
	 * Is this switch AR capable?
	 */
	if (!osm_switch_is_ar_supported(p_sw)) {
		return;
	}

	/*
	 * Set a static route to port 0.
	 */
	lid = osm_node_get_base_lid(p_sw->p_node, 0);
	lid = cl_ntoh16(lid);
	minhop_set_new_static_lft_entry(p_subn, p_sw, lid);

	/*
	 * Set a static route to attached CAs.
	 */
	for (i = 0; i < p_minhop_node->ca_count; i++) {
		min_lid = p_minhop_node->ports[i].min_lid;
		max_lid = p_minhop_node->ports[i].max_lid;

		for (lid = min_lid; lid < max_lid; lid++) {
			minhop_set_new_static_lft_entry(p_subn, p_sw, lid);
		}
	}

	/*
	 * For every switch in the fabric, we determine which of the neighbors are the
	 * start of the shortest path between the source and the destination.  It could
	 * be that several neighbors have the shortest path.  In this case, we remember
	 * all of the neighbors for the destination.
	 */
	max_group_id = 0;

	foreach_switch(p_subn, p_remote_sw) {
		if (p_remote_sw == p_sw) {	// loopback - ignore
			continue;
		}

		p_remote_node = p_remote_sw->p_node;
		p_minhop_remote_node = index_to_node(p_remote_sw->re_index);

		/*
		 * Set static route to remote switch port 0.
		 */
		remote_lid = osm_node_get_base_lid(p_remote_node, 0);
		remote_lid = cl_ntoh16(remote_lid);

		minhop_set_new_static_lft_entry(p_subn, p_sw, remote_lid);

		/*
		 * Get the AR group of ALL of the ports with minimum number of hops to this switch.
		 */
		memset((void *)remote_ar_group.mask, 0, sizeof(remote_ar_group.mask));
		if (osm_switch_get_minhop_ports(p_sw, remote_lid, &remote_ar_group, NULL) == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE, "%s - no route to switch %s\n",
				p_sw->p_node->print_desc, p_remote_sw->p_node->print_desc);
			continue;
		}

		/*
		 * We previously processed the ports_list and cached the lids which were attached to
		 * each switch.  We now take that list and use it to create the AR entries.
		 */
		group_id = p_minhop_remote_node->group_id;
		group_is_static = (ar_state == OSM_AR_IB_LID_STATE_STATIC) ||
				  (group_id == UCAST_MGR_INVALID_GROUP_ID) ||
				  (group_id >= group_cap);

		if (!group_is_static && (group_id > max_group_id)) {
			max_group_id = group_id;
		}

		for (i = 0; i < p_minhop_remote_node->ca_count; i++) {
			port_type = p_minhop_remote_node->ports[i].type;
			min_lid = p_minhop_remote_node->ports[i].min_lid;
			max_lid = p_minhop_remote_node->ports[i].max_lid;

			/*
			 * Set the AR group for these lids if needed.  (ROUTERS and AN ports get static group.)
			 */
			for (lid = min_lid; lid < max_lid; lid++) {
				if (group_is_static || (port_type != IB_SPECIAL_PORT_TYPE_RESERVED)) {
					minhop_set_new_static_lft_entry(p_subn, p_sw, lid);
				} else {
					minhop_set_new_free_lft_entry(p_subn, p_sw, lid, group_id, &remote_ar_group);
				}
			}
		}
	}

	osm_ar_set_required_group_top(p_ar_data, max_group_id);
	p_minhop_node->max_group_id = max_group_id;
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return;
}

static int minhop_ar_switches(IN void **tasks,
			      IN unsigned task_len,
			      IN void *context,
			      IN unsigned int thread_id)
{
	int		i;
	minhop_t	*p_minhop;

	p_minhop = (minhop_t *)context;
	OSM_LOG_ENTER(&p_minhop->p_subn->p_osm->log);

	for (i = 0; i < task_len; i++) {
		minhop_ar_switch(p_minhop->p_subn, tasks[i]);
	}

	OSM_LOG_EXIT(&p_minhop->p_subn->p_osm->log);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

int osm_minhop_ar_process_switch(IN osm_subn_t *p_subn,
				 IN osm_switch_t *p_sw) {
	minhop_t	*p_minhop;
	minhop_node_t	*p_minhop_node;

	p_minhop = (minhop_t *)p_sw->re_priv;
	p_minhop_node = index_to_node(p_sw->re_index);
	return p_minhop_node->max_group_id + 1;
}

void osm_minhop_ar_process_subnet(IN osm_subn_t *p_subn,
				  IN void *context,
				  IN int ar_enable,
				  IN int ar_mode)
{
	int		port_type;
	uint16_t	min_lid;
	uint16_t	max_lid;
	minhop_t	*p_minhop;
	minhop_node_t	*p_minhop_node;
	minhop_node_t	*p_minhop_remote_node;
	osm_port_t	*p_port;
	osm_node_t	*p_remote_node;
	osm_physp_t	*p_physp;
	osm_switch_t	*p_sw;
	osm_switch_t	*p_remote_sw;

	p_minhop = (minhop_t *)context;
	p_minhop->ar_enable = ar_enable;
	p_minhop->ar_mode = ar_mode;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/*
	 * Find the leaf switches and the lids of the attached HCAs.
	 *
	 * Set the group ids.  This is done here because locking isn't allowed in the routing code.
	 * If we tried to do it later, it might be in the parallel threading model.
	 */
	foreach_minhop_node(p_minhop, p_minhop_node) {
		p_minhop_node->ca_count = 0;
	}

	foreach_port(p_minhop->p_subn, p_port) {
		p_physp = p_port->p_physp;

		if (p_port->p_node->sw != NULL) {
			continue;
		} else if (p_physp == NULL) {
			continue;
		} else if (p_physp->p_remote_physp == NULL) {
			continue;
		}

		port_type = osm_physp_get_special_port_type(p_physp);

		min_lid = cl_hton16(p_port->lid);
		max_lid = min_lid + (1 << osm_physp_get_lmc(p_physp));

		p_remote_node = p_port->p_remote_sw_node;
		p_remote_sw = p_remote_node->sw;

		p_minhop_remote_node = index_to_node(p_remote_sw->re_index);
		p_minhop_remote_node->ports[p_minhop_remote_node->ca_count].type = port_type;
		p_minhop_remote_node->ports[p_minhop_remote_node->ca_count].min_lid = min_lid;
		p_minhop_remote_node->ports[p_minhop_remote_node->ca_count].max_lid = max_lid;
		p_minhop_remote_node->ca_count++;

		if (port_type == IB_SPECIAL_PORT_TYPE_RESERVED) {
			p_minhop_remote_node->is_leaf_switch = TRUE;
		}
	}

	/*
	 * Set the group ids.  This is done here because locking isn't allowed in the routing code.
	 * If we tried to do it later, it might be in the parallel threading model.
	 */
	foreach_switch(p_minhop->p_subn, p_sw) {
		p_minhop_node = index_to_node(p_sw->re_index);

		if (p_minhop_node->is_leaf_switch) {
			p_minhop_node->group_id = minhop_set_group_id(p_subn, p_minhop_node->lid);
		} else {
			p_minhop_node->group_id = AR_INVALID_GROUP_ID;
		}
	}

	/*
	 * Set the AR groups on the switches.
	 */
	minhop_execute_threaded(minhop_ar_switches, context);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return;
}
